Package: vismeteor
Type: Package
Title: Analysis of Visual Meteor Data
Version: 3.0.1
Authors@R: c(person("Janko", "Richter", role = c("aut", "cre"), email = "janko@richtej.de"))
Author: Janko Richter [aut, cre]
Maintainer: Janko Richter <janko@richtej.de>
Description: Provides a suite of analytical functionalities to process and analyze
    visual meteor observations from the Visual Meteor Database
    of the International Meteor Organization <https://www.imo.net/>.
License: MIT + file LICENSE
URL: https://github.com/jankorichter/vismeteor
BugReports: https://github.com/jankorichter/vismeteor/issues
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.1.0)
Imports: methods, stats, httr2 (>= 1.0.0)
Suggests: testthat (>= 3.2.0), httptest2, knitr, rmarkdown
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-19 18:05:10 UTC; j.richter
Repository: CRAN
Date/Publication: 2026-05-19 21:30:02 UTC
Built: R 4.5.2; ; 2026-05-19 21:41:53 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cvmideal)
export(dmideal)
export(dvmgeom)
export(dvmideal)
export(freq_quantile)
export(load_vmdb_magnitudes)
export(load_vmdb_rates)
export(pmideal)
export(pvmgeom)
export(pvmideal)
export(qmideal)
export(qvmgeom)
export(qvmideal)
export(rmideal)
export(rvmgeom)
export(rvmideal)
export(select_knots)
export(vmgeom_vst_from_magn)
export(vmgeom_vst_to_r)
export(vmideal_vst_from_magn)
export(vmideal_vst_to_psi)
export(vmperception)
export(vmtable)
import(httr2)

# R package `vismeteor`

## About The Project

This R package provides tools for the analysis of visual meteor observations from the Visual Meteor Database (VMDB) of the [International Meteor Organization](https://www.imo.net/).

The data used in this package are provided by [imo-vmdb](https://pypi.org/project/imo-vmdb/). But it is also possible to perform analyses without using this data.

## Installation

Install the released version from CRAN:

```r
install.packages("vismeteor")
```

## Getting Started

Explore the vignettes for full workflows and methodology:

- `vignette("vismeteor")` – package overview and concepts
- `vignette("vmgeom")` – parameter estimation for the geometric model
- `vignette("vmideal")` – parameter estimation for the ideal distribution

## Contributing

Contributions are what make the open source community such an amazing place to learn, inspire, and create. Any contributions you make are **greatly appreciated**.

If you have a suggestion that would make this better, please fork the repo and create a pull request. You can also simply open an issue with the tag "enhancement".

1.  Fork the Project
2.  Create your Feature Branch (`git checkout -b feature/AmazingFeature`)
3.  Commit your Changes (`git commit -m 'Add some AmazingFeature'`)
4.  Push to the Branch (`git push origin feature/AmazingFeature`)
5.  Open a Pull Request

## License

Distributed under the MIT License. See `LICENSE` for more information.

## Contact

Janko Richter - [janko\@richtej.de](mailto:janko@richtej.de)

Project Link: <https://github.com/jankorichter/vismeteor>

##
## Script for deriving/validating the approximation of perception probabilities
## g(m) used in vmgeom/vmideal likelihoods.
##
## Purpose
## - Provide a smooth parametric approximation g(m) to empirical perception
##   probabilities as a function of magnitude m.
## - Use a simple polynomial inside an exponential form so that
##   g(m) = 1 - exp(-f(m+0.5)), which is numerically stable and monotone.
## - Calibrate coefficients by matching moments/MLE behavior under vmgeom and
##   vmideal to those obtained from an empirical table (p.fun).
## - Visualize g(m), its derivative g'(m), and q(m) = g'(m)/g(m).
##
## Contents
## - perception.fun(): evaluates g(m) and optionally its derivative using a
##   polynomial f on m+0.5; coefficients are passed explicitly.
## - p.fun: piecewise‑linear interpolation of empirical probabilities.
## - coef.model: pre‑tuned polynomial coefficients (log‑scale tuning optional).
## - model.geom.fun()/model.ideal.fun(): compare moment/MLE behavior using
##   the fitted g against the empirical p.fun for vmgeom/vmideal.
## - Plots: g(m), g'(m), and q(m) on log‑scaled y‑axes for diagnostics.
##
## Usage / workflow
## 1. Run the script once end-to-end to reproduce the diagnostics. Toggle the
##    optimisation block (`if (FALSE)`) only when recalibrating the polynomial
##    coefficients; optimisation can take several minutes.
## 2. After updating `coef.model`, copy the rounded coefficients (including the
##    leading zero term) into `R/perception.R` where `vmperception()` is defined.
##    The function expects coefficients in ascending power order.
## 3. Rebuild the package (`devtools::load_all()`) and rerun perception-related
##    tests or models that rely on `vmperception` to ensure behaviour is
##    unchanged or improved.
##
## Output locations in the package
## - `coef.model` → polynomial coefficients in `R/perception.R`.
## - The helper functions `model.geom.fun()` and `model.ideal.fun()` provide
##   diagnostic tables only; they are not stored automatically.
## - Plot sections print ggplot objects for manual inspection; no files are
##   written by default.
##
## Dependencies
## - Requires 'vismeteor' (dvmgeom, dvmideal, f.polynomial, f.polynomial.coef)
##   and ggplot2 for plots.
##
## Notes on execution
## - The if (FALSE) block holds an optimization that can be used to refine
##   coef.model by minimizing squared errors in mean magnitudes for both
##   vmgeom and vmideal models; it is disabled by default.
## - Coefficients provided in coef.model are used as‑is for reproducibility.
##
# Approximation of perception probabilities
library(vismeteor)
library(ggplot2)

#' Perception probability model g(m) with optional derivative
#'
#' Evaluates g(m) = 1 - exp(-f(m+0.5)), where f is a polynomial with
#' coefficients 'poly.coef'. If deriv.degree = 1, returns g'(m).
#'
#' Arguments
#' - poly.coef: numeric coefficients of f in ascending power order.
#' - m: numeric vector of magnitudes.
#' - deriv.degree: 0 for g(m), 1 for g'(m); higher orders not implemented.
#'
#' Returns
#' - Numeric vector of g(m) or g'(m).
perception.fun <- function(poly.coef, m, deriv.degree = 0L) {
    names(poly.coef) <- seq(along = poly.coef) # exponents

    m <- m + 0.5
    p <- rep(0.0, length(m))
    idx <- m > .Machine$double.eps
    if (any(idx)) {
        f0 <- vismeteor:::f.polynomial(m[idx], poly.coef)
        if (0L == deriv.degree) {
            p[idx] <- 1.0 - exp(-f0)
        } else if (1L == deriv.degree) {
            poly.coef1 <- vismeteor:::f.polynomial.coef(poly.coef, deriv.degree = 1L)
            f1 <- vismeteor:::f.polynomial(m[idx], poly.coef1)
            p[idx] <- exp(-f0) * f1
        } else {
            stop('Not implemented')
        }
    }

    p
}

data <- data.frame(
    m = seq(-0.4, 7.6, 0.2),
    p = c(
        0.00046, 0.0011, 0.0023, 0.0046, 0.0081, 0.0122, 0.0182,
        0.0257, 0.03444, 0.04365, 0.05495, 0.06918, 0.08511, 0.10351,
        0.13, 0.16, 0.2, 0.24, 0.29, 0.34674, 0.4, 0.46, 0.52, 0.57544,
        0.63096, 0.67764, 0.71, 0.74, 0.77, 0.79, 0.81, 0.83, 0.85,
        0.87, 0.89, 0.91, 0.93, 0.94, 0.96, 0.98, 1.0
    )
)
data <- rbind(data.frame(m=-0.5, p=0.0), data)
p.fun <- approxfun(data$m, data$p, yleft = 0.0, yright = 1.0)
limmag <- seq(5.6, 6.4, 0.2)

# coef.model <- c(0.00373590946848783, 0.00189710356299022, 0.00271083620131325, 0.000899903791081514)
# After tuning, copy the rounded coefficients (plus the implicit zero intercept)
# into `R/perception.R::vmperception`.
# Set the working polynomial coefficients for g(m).
coef.model <- c(0.0037, 0.0019, 0.00271, 0.0009)

# Estimate coefficients (optional calibration)
if (FALSE) {
    coef.model <- with(new.env(), {
        r <- seq(1.4, 3.5, 0.1)
        psi <- seq(-4, 9, 0.5)
        m <- seq(-200, 6, 1)

        model.geom.fun <- function(poly.coef, r) {
            vmperception.local <- function(m, deriv.degree = 0L) {
                perception.fun(poly.coef, m, deriv.degree)
            }

            model.geom <- expand.grid(r=r, limmag=limmag)
            model.geom <- do.call(
                rbind.data.frame,
                mapply(function(r, limmag) {
                    p.org <- dvmgeom(m, limmag, r, perception.fun = p.fun)
                    p.est <- dvmgeom(m, limmag, r, perception.fun = vmperception.local)
                    m.mean.org <- sum(p.org * m)
                    m.mean.est <- sum(p.est * m)

                    list(
                        m.mean.org = m.mean.org,
                        m.mean.est = m.mean.est
                    )
                }, model.geom$r, model.geom$limmag, SIMPLIFY = FALSE)
            )
        }

        model.ideal.fun <- function(poly.coef, psi) {
            vmperception.local <- function(m, deriv.degree = 0L) {
                perception.fun(poly.coef, m, deriv.degree)
            }

            model.ideal <- expand.grid(psi=psi, limmag=limmag)
            model.ideal <- do.call(
                rbind.data.frame,
                mapply(function(psi, limmag) {
                    p.org <- dvmideal(m, limmag, psi, perception.fun = p.fun)
                    p.est <- dvmideal(m, limmag, psi, perception.fun = vmperception.local)
                    m.mean.org <- sum(p.org * m)
                    m.mean.est <- sum(p.est * m)

                    list(
                        m.mean.org = m.mean.org,
                        m.mean.est = m.mean.est
                    )
                }, model.ideal$psi, model.ideal$limmag, SIMPLIFY = FALSE)
            )
        }

        # Objective: match mean magnitudes derived with empirical p.fun and
        # with perception.fun using exp(params) to keep coefficients positive.
        f <- function(params) {
            model.geom <- model.geom.fun(exp(params), r = r)
            model.ideal <- model.ideal.fun(exp(params), psi = psi)
            print(
                sum((model.geom$m.mean.org - model.geom$m.mean.est)^2) +
                    sum((model.ideal$m.mean.org - model.ideal$m.mean.est)^2)
            )
        }

        optim(log(coef.model), f, control = list('reltol' = 1e-06))$par
    })
    coef.model <- exp(coef.model)
}
names(coef.model) <- seq(along = coef.model) # exponents
print(paste(c('coef model:', paste(coef.model, collapse = ', '))))

#' Compare vmgeom behavior using empirical vs. modeled perception
#'
#' Returns a data.frame with original vs. modeled means and the MLE of r under
#' the modeled perception, including Hessian‑based variance proxy.
model.geom.fun <- function(poly.coef, r) {
    m <- seq(-200, 6, 1)
    vmperception.local <- function(m, deriv.degree = 0L) {
        perception.fun(poly.coef, m, deriv.degree)
    }

    model.geom <- expand.grid(r=r, limmag=limmag)
    model.geom <- model.geom[order(model.geom$r, model.geom$limmag),]
    do.call(
        rbind.data.frame,
        mapply(function(r, limmag) {
            p.org <- dvmgeom(m, limmag, r, perception.fun = p.fun)
            p.est <- dvmgeom(m, limmag, r, perception.fun = vmperception.local)

            # Maximum likelihood estimation (MLE) of r
            llr <- function(r) {
                -sum(p.org * dvmgeom(m, limmag, r, log=TRUE, perception.fun = vmperception.local))
            }
            opt.result <- optim(r, llr, method='Brent', lower=1.1, upper=5, hessian=TRUE)
            r.est <- opt.result$par
            r.var <- opt.result$hessian[1][1]
            m.mean.org <- sum(p.org * (limmag - m))
            m.mean.est <- sum(p.est * (limmag - m))

            list(
                limmag = limmag,
                r = r,
                r.est = r.est,
                r.var = r.var,
                m.mean.org = m.mean.org,
                m.mean.est = m.mean.est
            )
        }, model.geom$r, model.geom$limmag, SIMPLIFY = FALSE)
    )
}

#' Compare vmideal behavior using empirical vs. modeled perception
#'
#' Returns a data.frame with original vs. modeled means and the MLE of psi under
#' the modeled perception, including Hessian‑based variance proxy.
model.ideal.fun <- function(poly.coef, psi) {
    m <- seq(-200, 6, 1)
    vmperception.local <- function(m, deriv.degree = 0L) {
        perception.fun(poly.coef, m, deriv.degree)
    }

    model.ideal <- expand.grid(psi=psi, limmag=limmag)
    model.ideal <- model.ideal[order(model.ideal$psi, model.ideal$limmag),]
    do.call(
        rbind.data.frame,
        mapply(function(psi, limmag) {
            p.org <- dvmideal(m, limmag, psi, perception.fun = p.fun)
            p.est <- dvmideal(m, limmag, psi, perception.fun = vmperception.local)

            # Maximum likelihood estimation (MLE) of psi
            llr <- function(psi) {
                -sum(p.org * dvmideal(m, limmag, psi, log=TRUE, perception.fun = vmperception.local))
            }
            opt.result <- optim(psi, llr, method='Brent', lower=-15, upper=15, hessian=TRUE)
            psi.est <- opt.result$par
            psi.var <- opt.result$hessian[1][1]
            m.mean.org <- sum(p.org * (limmag - m))
            m.mean.est <- sum(p.est * (limmag - m))

            list(
                limmag = limmag,
                psi = psi,
                psi.est = psi.est,
                psi.var = psi.var,
                m.mean.org = m.mean.org,
                m.mean.est = m.mean.est
            )
        }, model.ideal$psi, model.ideal$limmag, SIMPLIFY = FALSE)
    )
}

if (TRUE) {
    with(new.env(), {
        result.geom <- model.geom.fun(coef.model, r = seq(1.4, 3.5, 0.1))
        result.geom <- result.geom[order(result.geom$r, result.geom$limmag),]
        print(result.geom)
        result.ideal <- model.ideal.fun(coef.model, psi = seq(-4, 9, 0.5))
        result.ideal <- result.ideal[order(result.ideal$psi, result.ideal$limmag),]
        print(result.ideal)
    })
}

# Plot g(m)
if (TRUE) {
    with(new.env(), {
        m.idx <- data$m > -0.5
        g.model <- perception.fun(coef.model, data$m[m.idx])
        plot.data <- data.frame(
            m = data$m[m.idx],
            p = data$p[m.idx],
            g.model = g.model
        )
        print(plot.data)

        p <- ggplot(plot.data) +
            theme_bw() +
            geom_point(aes(x = m, y = p), color='black') +
            geom_line(aes(x = m, y = g.model), color='red') +
            scale_x_continuous(
                name = "m",
                breaks = seq(-0.5, 9.0, 0.5)
            ) +
            scale_y_continuous(
                name = "g",
                trans = "log",
                #labels = scales::comma,
                breaks = c(0.0002, 0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1.0)
            )
        print(p)
    })
}

# Plot g'(m)
if (TRUE) {
    with(new.env(), {
        m <- c(-0.499, seq(-0.45, 8.5, 0.05))
        g.model <- perception.fun(coef.model, m, deriv.degree = 1L)
        plot.data <- data.frame(
            m = m,
            g.model = g.model
        )

        p <- ggplot(plot.data) +
            theme_bw() +
            geom_line(aes(x = m, y = g.model), color='red') +
            scale_x_continuous(
                name = "m",
                breaks = seq(-0.5, 9.0, 0.5)
            ) +
            scale_y_continuous(
                name = "g'",
                trans = "log",
                #labels = scales::comma,
                breaks = c(0.001, 0.002,0.005, 0.01, 0.02, 0.05, 0.1, 0.2)
            )
        print(p)
    })
}

# Plot q(m) = g'(m) / g(m)
if (TRUE) {
    with(new.env(), {
        m <- seq(-0.4, 8.4, 0.05)
        q <- perception.fun(coef.model, m, deriv.degree = 1L)/perception.fun(coef.model, m)
        plot.data <- data.frame(
            m = m,
            q = q
        )

        p <- ggplot(plot.data) +
            theme_bw() +
            geom_line(aes(x = m, y = q), color='blue') +
            scale_x_continuous(
                name = "m",
                limits = c(-0.5, 8.5),
                breaks = seq(-0.5, 8.5, 0.5)
            ) +
            scale_y_continuous(
                name = "q",
                trans = "log",
                #labels = scales::comma,
                limits = c(0.002, 12),
                breaks = c(0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5, 10)
            )
        print(p)
    })
}

##
## Script for deriving/validating the variance-stabilizing transformation
## for the geometric model of visual meteor magnitudes (vmgeom).
##
## Purpose
## - Provide a smooth, closed-form approximation
##     t(m; limmag) ≈ a - exp(b + c * (x + 0.5)^d), with x = limmag - m,
##   whose variance (under vmgeom) is approximately constant across r.
## - Estimate parameters b, c, d as a function of the fractional limiting
##   magnitude and adjust a so that the reference point
##   t(m = 6, limmag = 5.5) ≈ 0.
## - Visualize the transformation and check stability via a regression fit
##   q = log(r) ~ poly(mean_t, 4).
##
## Contents
## - Precomputed table param.df (b, c, d) over limmag; a is determined later.
##   The fractional component (offset) of limmag is used for parameter
##   interpolation.
## - myVmgeomVstFromMagn(): implements the t-approximation for vectorized m/limmag.
## - data.t.fun(): creates a dataset with moments of m and t under vmgeom.
## - Plot 1: t(x) versus x = limmag - m for various limmag.
## - Plot 2: E[t] as a function of r across limmag families.
##
## Usage / workflow
## 1. Run the script top-to-bottom. Enable the guarded `if (FALSE)` blocks only
##    when you need to re-optimise the coefficients; they take several minutes.
## 2. Copy the resulting `param.df` (columns a–d) and polynomial coefficients
##    reported by `lm.res.t` into `R/vmgeom_vst.R` (functions
##    `vmgeomVstFromMagn` and `vmgeomVstToR`). The package stores the values as
##    literals rather than loading them from disk.
## 3. Re-run `devtools::load_all()` and the dedicated unit tests (see
##    `tests/testthat/test_vmgeom_vst.R` if present) to confirm the new
##    coefficients behave as expected before making changes public.
##
## Output locations in the package
## - `param.df` → embedded in `R/vmgeom_vst.R` (`vmgeomVstFromMagn`).
## - `lm.res.t.coef` → embedded in `R/vmgeom_vst.R` (`vmgeomVstToR`).
## - Diagnostic objects (`data.t`, `lm.res.t`) are transient and intended for
##   interactive inspection only; they are not saved automatically.
##
## Dependencies
## - Requires the 'vismeteor' package (e.g., dvmgeom) and ggplot2.
##
## Execution/Reproducibility notes
## - The if (FALSE) blocks contain the time-consuming optimization steps used to
##   obtain/refine (b, c, d) and a; they are not executed by default. The values
##   currently used are hard-coded below.
## - Important: data.t.fun(param.df) expects param.df$a to exist. When running
##   this script top-to-bottom, the first call to data.t.fun() should happen AFTER
##   param.df$a is assigned (or comment out that call temporarily). The current
##   order is preserved for documentary purposes from the original derivation.
##
##
# Approximation of the Variance-stabilizing transformation
library(vismeteor)
library(ggplot2)

r <- seq(1.4, 3.5, 0.1)
m <- seq(-200, 6, 1)

# dput(param.df)
# NOTE: Keep this table in sync with the literal `param.df` inside
# `R/vmgeom_vst.R::vmgeomVstFromMagn`. After recalibration, copy the rounded
# values there so the package picks them up without sourcing this script.
param.df <- structure(list(limmag = c(5.5, 5.52, 5.55, 5.6, 5.7, 5.8, 5.9,
    6, 6.1, 6.2, 6.3, 6.4, 6.45, 6.48, 6.5), b = c(2.176, 2.077,
    2.004, 1.945, 1.897, 1.885, 1.889, 1.906, 1.933, 1.97, 2.02,
    2.087, 2.128, 2.156, 2.176), c = c(-0.317, -0.278, -0.249, -0.225,
    -0.206, -0.2, -0.201, -0.207, -0.217, -0.231, -0.252, -0.279,
    -0.296, -0.308, -0.317), d = c(0.669, 0.748, 0.813, 0.872, 0.922,
    0.934, 0.928, 0.91, 0.881, 0.843, 0.796, 0.739, 0.706, 0.684,
    0.669)), row.names = c("5.5", "5.52", "5.55", "5.6", "5.7", "5.8",
    "5.9", "6", "6.1", "6.2", "6.3", "6.4", "6.45", "6.48", "6.5"
    ), class = "data.frame")

if (FALSE) {
    # Estimate (b, c, d) per limmag by minimizing the deviation of
    # t-variance from a target value (here 1.0)
    param.df <- with(new.env(), {
        limmag <- c(5.52, 5.55, seq(5.6, 6.4, 0.1), 6.45, 6.48, 6.5)
        myVmgeomVstFromMagn <- function(m, limmag, params) {
            a <- 0.0
            b <- params[1]
            c <- params[2]
            d <- params[3]

            x <- limmag - m
            a - exp(b + c * (x + 0.5)^d)
        }

        f.model <- function(params, limmag) {
            model <- expand.grid(r = r, limmag = limmag)
            model <- do.call(
                rbind.data.frame,
                mapply(function(r, limmag) {
                    p <- dvmgeom(m, limmag, r)
                    t <- myVmgeomVstFromMagn(m, limmag, params)
                    t.mean <- sum(p * t)
                    t.var <- sum(p * (t - t.mean)^2)

                    list(
                        r = r,
                        q = log(r),
                        limmag = limmag,
                        t.mean = t.mean,
                        t.var = t.var
                    )
                }, model$r, model$limmag , SIMPLIFY = FALSE)
            )
        }

        param.list <- list()
        for (l in limmag) {
            print(paste('limmag:', l))
            params <- as.vector(unlist(param.df[as.character(l), c('b', 'c', 'd')]))
            opt.res <- optim(params, fn = function(params, t.var.mean) {
                df <- f.model(params, l)
                sum((df$t.var - t.var.mean)^2)/nrow(df)
            }, t.var.mean = 1.0)
            params  <- c(l, opt.res[["par"]])
            param.list[[as.character(l)]] <- params
        }
        param.df <- do.call(
            rbind.data.frame,
            as.list(param.list)
        )
        colnames(param.df) <- c('limmag', 'b', 'c', 'd')

        d55 <- param.df[param.df == 6.5,]
        d55$limmag <- 5.5
        param.df <- rbind.data.frame(d55, param.df)
        param.df$b <- round(param.df$b, 3)
        param.df$c <- round(param.df$c, 3)
        param.df$d <- round(param.df$d, 3)
        param.df <- param.df[order(param.df$limmag),]
        row.names(param.df) <- as.character(param.df$limmag)
        param.df
    })
    dput(param.df)
}

param.df$offset <- round(param.df$limmag - round(param.df$limmag), 2)

#' Variance-stabilizing transform approximation t(m; limmag)
#'
#' Implements t(m; limmag) = a - exp(b + c * (x + 0.5)^d) with x = limmag - m.
#' Parameters a, b, c, d are linearly interpolated from param.df using the
#' fractional component (offset) of limmag.
#'
#' Arguments
#' - m: numeric vector of magnitudes.
#' - limmag: scalar or vector; limiting magnitude(s) corresponding to m.
#' - param.df: data.frame with columns 'offset', 'a', 'b', 'c', 'd'.
#'
#' Returns
#' - Numeric vector t with the same length as m.
myVmgeomVstFromMagn <- function(m, limmag, param.df) {
    offset <- limmag - round(limmag)
    if (1L == length(limmag)) {
        limmag <- rep(limmag, length(m))
        offset <- rep(offset, length(m))
    }

    pa.fun <- approxfun(param.df$offset, param.df$a)
    pb.fun <- approxfun(param.df$offset, param.df$b)
    pc.fun <- approxfun(param.df$offset, param.df$c)
    pd.fun <- approxfun(param.df$offset, param.df$d)

    a <- pa.fun(offset)
    b <- pb.fun(offset)
    c <- pc.fun(offset)
    d <- pd.fun(offset)

    # x is the excess relative to the limiting magnitude
    x <- limmag - m
    a - exp(b + c * (x + 0.5)^d)
}

limmag <- c(5.5, 5.52, 5.55, seq(5.6, 6.4, 0.1), 6.45, 6.48, 6.5)
#' Generate derived quantities under vmgeom for given parameters
#'
#' For a grid of r and limmag, computes the moments of m and the transformed
#' quantity t when m ~ vmgeom(limmag, r).
#'
#' Arguments
#' - param.df: parameter frame with columns 'a', 'b', 'c', 'd' (and 'offset').
#'
#' Returns
#' - data.frame with columns r, q=log(r), limmag, m.mean, m.var, t.mean, t.var.
data.t.fun <- function(param.df) {
    model <- expand.grid(r = r, limmag = limmag)
    model <- do.call(
        rbind.data.frame,
        mapply(function(r, limmag) {
            p <- dvmgeom(m, limmag, r)

            m.mean <- sum(p * (limmag - m))
            m.var <- sum(p * (limmag - m - m.mean)^2)

            t <- myVmgeomVstFromMagn(m, limmag, param.df)
            t.mean <- sum(p * t)
            t.var <- sum(p * (t - t.mean)^2)

            list(
                r = r,
                q = log(r),
                limmag = limmag,
                m.mean = m.mean,
                m.var = m.var,
                t.mean = t.mean,
                t.var = t.var
            )
        }, model$r, model$limmag , SIMPLIFY = FALSE)
    )
}
data.t <- data.t.fun(param.df)

# Pre-calibrated a values (as a function of limmag offset); later shifted
# globally so that t(6; 5.5) ≈ 0.
# These `a` coefficients are copied into `vmgeomVstFromMagn` once tuned.
param.df$a <- c(8.83660725413019, 8.52231677831387, 8.31587431846188, 8.16602303316079,
    8.05750794169318, 8.0501384519307, 8.07216870699935, 8.1189428855403,
    8.19069184757109, 8.29154614666007, 8.40974020292365, 8.58969657552978,
    8.70398711417616, 8.78298253370545, 8.83710213031958)

if (FALSE) {
    # Fine-tune the a values by minimizing squared error in the regression
    # model q ~ poly(mean_t, 4).
    opt.res <- optim(param.df$a, fn = function(a) {
        param.df$a <- a
        data.t <- data.t.fun(param.df)
        param.df$a <- round(param.df$a - myVmgeomVstFromMagn(6, 5.5, param.df), 6) + 1e-06
        data.t <- data.t.fun(param.df)
        lm.res <- lm(q ~ poly(t.mean, 4, raw = TRUE), data = data.t)
        data.t$q.predicted <- predict(lm.res)
        print(sum((data.t$q - data.t$q.predicted)^2)/nrow(data.t))
    })
    dput(opt.res[["par"]])
    param.df$a <- opt.res[["par"]]
}

# Global shift of a so that t(6; 5.5) ≈ 0 (numerically near zero).
# Copy the rounded result back to the package table after verifying the shift.
param.df$a <- round(param.df$a - myVmgeomVstFromMagn(6, 5.5, param.df), 6) + 1e-06
print(paste(c('should be zero:', myVmgeomVstFromMagn(6, 5.5, param.df))))
dput(param.df)
# regenerate summary moments with the shifted coefficients.
data.t <- data.t.fun(param.df)

# Regression model: q = log(r) ~ polynomial in E[t]
lm.res.t <- lm(q ~ poly(t.mean, 4, raw = TRUE), data = data.t)
lm.res.t.coef <- coef(lm.res.t)
# Paste `lm.res.t.coef` into `R/vmgeom_vst.R::vmgeomVstToR` (polynomial on tm).
print(lm.res.t.coef)
data.t$pred <- exp(predict(lm.res.t))
data.t$pred.diff <- data.t$r - data.t$pred

if (TRUE) {
    with(new.env(), {
        m <- seq(7.0, -4.5, -0.1)
        limmag <- seq(5.6, 6.5, 0.1)

        p <- ggplot(data.t) + theme_bw()

        for (l in limmag) {
            # Plot 1: shape of the transform t versus x = limmag - m
            plot.data <- expand.grid(limmag=limmag, m=m)
            plot.data$dm <- plot.data$limmag - plot.data$m
            plot.data <- plot.data[order(plot.data$dm),]
            plot.data$x <-  myVmgeomVstFromMagn(m = plot.data$m, limmag = plot.data$limmag, param.df = param.df)
            plot.data <- subset(plot.data, plot.data$x > 0)
            p <- p + geom_line(aes(x = dm, y = x), color='blue', data = subset(plot.data, l == plot.data$limmag))
        }

        p <- p +
            scale_x_continuous(
                name = "m",
                breaks = seq(0, 11, 1.0)
            ) +
            scale_y_continuous(
                name = "t",
                breaks = seq(0, 7, 1)
            )
        print(p)
    })
}

if (TRUE) {
    with(new.env(), {
        data.t <- data.t[order(data.t$q, data.t$limmag),]
        p <- ggplot(data.t) + theme_bw()

        for (l in limmag) {
            # Plot 2: E[t] versus r (grouped by limmag)
            plot.data <- subset(data.t, data.t$limmag == l)
            p <- p + geom_line(aes(x = r, y = t.mean), color='blue', data = plot.data)
        }

        p <- p +
            scale_x_continuous(
                name = "r",
                breaks = seq(1.4, 3.3, 0.1)
            ) +
            scale_y_continuous(
                name = "t.mean",
                breaks = seq(4.2, 5.9, 0.1)
            )
        print(p)
    })
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)

## ----setup, include = FALSE---------------------------------------------------
library(vismeteor)
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)

## ----echo=TRUE----------------------------------------------------------------
set.seed(1)
x <- seq(0, 10, length.out = 200)
y <- rpois(length(x), lambda = 50 + 30 * sin(x))
dat <- data.frame(x = x, y = y)

fit <- \(d, knots) {
    f <- if (length(knots) == 0L) {
        y ~ splines::ns(x)
    } else {
        y ~ splines::ns(x, knots = knots)
    }
    glm(f, family = poisson(), data = d)
}
score_bic <- \(d, knots) fit(d, knots) |> BIC()

result <- select_knots(
    dat,
    knot_candidates = seq(1, 9, by = 0.5),
    score_fun = score_bic
)
final_knots <- sort(c(result$knots, result$fixed_knots))
final_fit <- fit(dat, final_knots)

op <- par(mar = c(7, 4, 4, 2) + 0.1)
plot(
    x, y,
    pch = 20, col = "gray",
    main = "Poisson spline fit with selected knots",
    xlab = "x", ylab = "count"
)
lines(x, 50 + 30 * sin(x), col = "darkgreen", lwd = 2, lty = 2)
lines(x, final_fit |> predict(type = "response"), col = "blue", lwd = 2)
abline(v = final_knots, col = "red", lty = 3)
legend(
    "bottom",
    inset = c(0, -0.35),
    legend = c("true mean", "spline fit", "selected knots"),
    col = c("darkgreen", "blue", "red"),
    lty = c(2, 1, 3),
    lwd = c(2, 2, 1),
    horiz = TRUE,
    bty = "n",
    xpd = TRUE
)
par(op)

## ----echo=TRUE----------------------------------------------------------------
plot(
    vmperception,
    -0.5, 8,
    main = "Perception probability of visual meteor magnitudes",
    col = "blue",
    xlab = "limiting magnitude - meteor magnitude",
    ylab = "p"
)

## ----echo=TRUE, results='hide'------------------------------------------------
m <- seq(6, -4, -1)
limmag <- 6.5
r <- 2.0
p <- vismeteor::dvmgeom(m, limmag, r)
barplot(
    p,
    names.arg = m,
    main = paste0("Density (r = ", r, ", limmag = ", limmag, ")"),
    col = "blue",
    xlab = "m",
    ylab = "p",
    border = "blue",
    space = 0.5
)
axis(side = 2, at = pretty(p))

## ----echo=TRUE, results='hide'------------------------------------------------
m <- seq(6, -4, -1)
psi <- 5.0
limmag <- 6.5
p <- vismeteor::dvmideal(m, limmag, psi)
barplot(
    p,
    names.arg = m,
    main = paste0("Density (psi = ", psi, ", limmag = ", limmag, ")"),
    col = "blue",
    xlab = "m",
    ylab = "p",
    border = "blue",
    space = 0.5
)
axis(side = 2, at = pretty(p))

## ----echo=TRUE----------------------------------------------------------------
mt <- as.table(matrix(
    c(
        0.0, 0.0, 2.5, 0.5, 0.0, 1.0,
        0.0, 1.5, 2.0, 0.5, 0.0, 0.0,
        1.0, 0.0, 0.0, 3.0, 2.5, 0.5
    ),
    nrow = 3, ncol = 6, byrow = TRUE
))
colnames(mt) <- seq(6)
rownames(mt) <- c("A", "B", "C")
margin.table(mt, 1)
margin.table(mt, 2)

# contingency table with integer values
(mt_int <- vmtable(mt))
margin.table(mt_int, 1)
margin.table(mt_int, 2)

## ----echo=TRUE----------------------------------------------------------------
freq <- c(1, 8, 3, 3, 4, 9, 5, 0, 0, 2, 7, 8, 2, 6, 4)
f <- freq_quantile(freq, 10)
print(f)
print(tapply(freq, f, sum))

## ----setup, include = FALSE---------------------------------------------------
library(vismeteor)
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)

## ----echo=TRUE, results='hide'------------------------------------------------
observations <- with(PER_2015_magn$observations, {
    idx <- !is.na(lim_magn) & sl_start > 135.81 & sl_end < 135.87
    data.frame(
        magn_id = magn_id[idx],
        lim_magn = lim_magn[idx]
    )
})
head(observations, 5) # Example values

## ----echo=FALSE, results='asis'-----------------------------------------------
knitr::kable(head(observations, 5))

## ----echo=TRUE, results='hide'------------------------------------------------
magnitudes <- with(new.env(), {
    magnitudes <- merge(
        observations,
        as.data.frame(PER_2015_magn$magnitudes),
        by = "magn_id"
    )
    magnitudes$magn <- as.integer(as.character(magnitudes$magn))
    subset(magnitudes, (magnitudes$lim_magn - magnitudes$magn) > -0.5)
})
head(magnitudes[magnitudes$Freq > 0, ], 5) # Example values

## ----echo=FALSE, results='asis'-----------------------------------------------
knitr::kable(head(magnitudes[magnitudes$Freq > 0, ], 5))

## ----echo=TRUE, results='hide'------------------------------------------------
# maximum likelihood estimation (MLE) of psi
result <- with(magnitudes, {
    # log likelihood function
    ll <- function(psi) -sum(Freq * dvmideal(magn, lim_magn, psi, log = TRUE))
    psi_start <- 6.0 # starting value
    psi_lower <- 4.0 # lowest expected value
    psi_upper <- 10.0 # highest expected value
    # find minimum
    optim(psi_start, ll, method = "Brent", lower = psi_lower, upper = psi_upper, hessian = TRUE)
})

## ----echo=TRUE----------------------------------------------------------------
psi_mean <- result$par # mean of psi
print(psi_mean)
psi_var <- 1 / result$hessian[1][1] # variance of psi
print(psi_var)

## ----echo=TRUE, results='hide'------------------------------------------------
with(new.env(), {
    data_plot <- data.frame(psi = seq(4.0, 11, 0.1))
    data_plot$ll <- mapply(function(psi) {
        with(magnitudes, {
            # log likelihood function
            sum(Freq * dvmideal(magn, lim_magn, psi, log = TRUE))
        })
    }, data_plot$psi)
    data_plot$l <- exp(data_plot$ll - max(data_plot$ll))
    data_plot$l <- data_plot$l / sum(data_plot$l)
    plot(data_plot$psi, data_plot$l,
        type = "l",
        col = "blue",
        xlab = "psi",
        ylab = "likelihood"
    )
    abline(v = result$par, col = "red", lwd = 1)
})

## ----echo=TRUE, results='show'------------------------------------------------
tm_mean <- with(magnitudes, {
    N <- sum(Freq)
    tm <- vmideal_vst_from_magn(magn, lim_magn)
    tm_mean <- sum(Freq * tm) / N
    tm_var <- sum(Freq * (tm - tm_mean)^2) / (N - 1)
    tm_mean_var <- tm_var / N
    list("val" = tm_mean, "var" = tm_mean_var, "sd" = sqrt(tm_mean_var))
})

## ----echo=TRUE, results='show'------------------------------------------------
print(paste("tm mean:", tm_mean$val))
print(paste("tm var:", tm_mean$var))

## ----echo=TRUE, results='show'------------------------------------------------
tm_means <- with(magnitudes, {
    N <- sum(Freq)
    tm <- vmideal_vst_from_magn(magn, lim_magn)
    replicate(50000, {
        mean(sample(tm, size = N, replace = TRUE, prob = Freq))
    })
})

## ----echo=TRUE, results='show'------------------------------------------------
with(new.env(), {
    tm_min <- tm_mean$val - 3 * tm_mean$sd
    tm_max <- tm_mean$val + 3 * tm_mean$sd
    tm_means <- subset(tm_means, tm_means > tm_min & tm_means < tm_max)
    brks <- seq(min(tm_means) - 0.02, max(tm_means) + 0.02, by = 0.02)
    hist(tm_means,
        breaks = brks,
        col = "skyblue",
        border = "black",
        main = "Histogram of mean tm",
        xlab = "tm",
        ylab = "count",
        xaxt = "n"
    )
    axis(1, at = seq(round(min(brks), 1), round(max(brks), 1) + 0.1, by = 0.1))
    abline(v = 0, col = "red", lwd = 1)
})

## ----echo=TRUE----------------------------------------------------------------
lim_magn_mean <- with(magnitudes, {
    N <- sum(Freq)
    sum(Freq * lim_magn) / N
})
print(paste("lim_magn_mean:", lim_magn_mean))
print(paste("mean psi:", vmideal_vst_to_psi(tm_mean$val, lim_magn_mean)))

## ----echo=TRUE----------------------------------------------------------------
print(vmideal_vst_to_psi(qnorm(0.90, tm_mean$val, tm_mean$sd), lim_magn_mean))

## ----echo=TRUE, results='asis'------------------------------------------------
psi_mean <- vmideal_vst_to_psi(tm_mean$val, lim_magn_mean)
magnitudes$p <- with(magnitudes, dvmideal(m = magn, lm = lim_magn, psi_mean))

## ----echo=TRUE, results='hide'------------------------------------------------
magn_min <- min(magnitudes$magn)

## ----echo=TRUE, results='asis'------------------------------------------------
idx <- magnitudes$magn == magn_min
magnitudes$p[idx] <- with(
    magnitudes[idx, ],
    pvmideal(m = magn + 1L, lm = lim_magn, psi_mean, lower.tail = TRUE)
)

## ----echo=TRUE----------------------------------------------------------------
magnitutes_observed <- xtabs(Freq ~ magn_id + magn, data = magnitudes)
magnitutes_observed_mt <- margin.table(magnitutes_observed, margin = 2)
print(magnitutes_observed_mt)

## ----echo=TRUE----------------------------------------------------------------
magnitudes$magn[magnitudes$magn <= 0] <- "0-"
magnitudes$magn[magnitudes$magn >= 4] <- "4+"
magnitutes_observed <- xtabs(Freq ~ magn_id + magn, data = magnitudes)
print(margin.table(magnitutes_observed, margin = 2))

## ----echo=TRUE----------------------------------------------------------------
magnitutes_expected <- xtabs(p ~ magn_id + magn, data = magnitudes)
magnitutes_row_freq <- margin.table(magnitutes_observed, margin = 1)
magnitutes_expected <- sweep(magnitutes_expected, 1, magnitutes_row_freq, `*`)
magnitutes_expected <- magnitutes_expected / sum(magnitutes_expected)
print(sum(magnitudes$Freq) * margin.table(magnitutes_expected, margin = 2))

## ----echo=TRUE, results='asis'------------------------------------------------
chisq_test_result <- chisq.test(
    x = margin.table(magnitutes_observed, margin = 2),
    p = margin.table(magnitutes_expected, margin = 2)
)

## ----echo=TRUE----------------------------------------------------------------
chi2_df <- chisq_test_result$parameter - 1
chi2_pval <- pchisq(chisq_test_result$statistic, df = chi2_df, lower.tail = FALSE)
print(chi2_pval)

## ----fig.show='hold'----------------------------------------------------------
chisq_test_residuals <- with(new.env(), {
    chisq_test_residuals <- residuals(chisq_test_result)
    v <- as.vector(chisq_test_residuals)
    names(v) <- names(chisq_test_residuals)
    v
})

plot(
    chisq_test_residuals,
    main = "Residuals of the chi-square goodness-of-fit test",
    xlab = "m",
    ylab = "Residuals",
    ylim = c(-3, 3),
    xaxt = "n"
)
abline(h = 0.0, lwd = 2)
axis(1, at = seq_along(chisq_test_residuals), labels = names(chisq_test_residuals))

##
## Script for deriving/validating the variance-stabilizing transformation
## for the ideal magnitude model (vmideal).
##
## Purpose
## - Provide a smooth approximation t(m; limmag) for vmideal using a
##   spline in x = limmag − m. The spline shape depends on the fractional
##   component (offset) of limmag.
## - Choose parameters so that Var[t] under vmideal is approximately constant
##   across a wide range of psi, and align an intercept to a convenient
##   reference (via regression/optimization).
## - Fit a regression mapping from log(E[t]) to psi − limmag for quick
##   inversion (myVmidealVstToPsi).
##
## Contents
## - myVmidealVstFromMagn(): variance‑stabilizing transform t for vmideal,
##   built from a cubic spline with anchors at sx = {1, 2, 4, 6, 8, 10}.
##   Coefficients depend on limmag offset via linear interpolation.
## - param.df: precomputed coefficients for offsets derived from limmag in
##   [5.5, 6.5]; columns '1'..'6' are spline values at sx; 'intercept' is the
##   alignment term; 'offset' is limmag − round(limmag).
## - data.t.fun(): generates E[t] and Var[t] over grids of psi and limmag.
## - Optimization blocks (guarded by if(FALSE)) to refine coefficients.
## - myVmidealVstToPsi(): polynomial mapping from log(x) with x = E[t] to
##   psi, including optional derivative for Jacobians.
##
## Usage / workflow
## 1. Execute the script sequentially. Activate the `if (FALSE)` sections only
##    when you need to recompute spline anchors or intercepts; they can be slow.
## 2. Copy the updated `param.df` (columns `1`..`6`, `intercept`, `offset`) and
##    the regression coefficients `lm.coeff` into `R/vmideal_vst.R` (see
##    `vmidealVstFromMagn` and `vmidealVstToPsi`). The package keeps these
##    values inline for speed and reproducibility.
## 3. Validate the new values with `devtools::load_all()` followed by the
##    relevant unit tests (for example `test_vmideal_vst.R`) before committing
##    the changes.
##
## Output locations in the package
## - `param.df` → embedded table inside `R/vmideal_vst.R`.
## - `lm.coeff` → polynomial used by `vmidealVstToPsi` in
##   `R/vmideal_vst.R`.
## - Intermediate diagnostics (`data.t`, `lm.res`) are for exploratory use and
##   are not saved automatically.
##
## Dependencies
## - Requires 'vismeteor' for dvmideal and helper polynomial functions,
##   and base R.
##
## Notes on execution
## - if (FALSE) sections contain time‑consuming optimization and are not run by
##   default. Coefficients below are precomputed and used as is.
## - myVmidealVstToPsi expects polynomial coefficients as produced by the
##   regression at the end of this script; keep the valid x‑domain checks in
##   mind when reusing the mapping.
##
# Approximation of the Variance-stabilizing transformation
library(vismeteor)

m <- seq(6, -200, -1)

#' Variance‑stabilizing transform approximation for vmideal
#'
#' Builds t(m; limmag) from a cubic spline in x = limmag − m with anchor
#' points sx = {1, 2, 4, 6, 8, 10}. The spline values at sx depend on the
#' fractional component (offset) of limmag via linear interpolation of
#' param.df columns '1'..'6'. An offset‑specific intercept is subtracted so
#' that the transformation is centered consistently across offsets.
#'
#' Arguments
#' - m: numeric vector of magnitudes.
#' - limmag: scalar or vector of limiting magnitudes corresponding to m.
#' - param.df: data.frame with columns 'offset', '1'..'6', and 'intercept'.
#'
#' Returns
#' - Numeric vector t with the same length as m.
myVmidealVstFromMagn <- function(m, limmag, param.df) {
    offset <- limmag - round(limmag)
    if (1L == length(limmag)) {
        limmag <- rep(limmag, length(m))
        offset <- rep(offset, length(m))
    }

    sx <- c(1.0, 2.0, 4.0, 6.0, 8.0, 10.0)
    p0.fun <- approxfun(param.df$offset, param.df[['intercept']])
    p1.fun <- approxfun(param.df$offset, param.df[['1']])
    p2.fun <- approxfun(param.df$offset, param.df[['2']])
    p3.fun <- approxfun(param.df$offset, param.df[['3']])
    p4.fun <- approxfun(param.df$offset, param.df[['4']])
    p5.fun <- approxfun(param.df$offset, param.df[['5']])
    p6.fun <- approxfun(param.df$offset, param.df[['6']])

    # x is the excess relative to the limiting magnitude
    arg.data <- data.frame(
        x = limmag - m,
        offset = offset
    )

    data.f <- as.factor(arg.data$offset)
    data.s <- split(arg.data, data.f)
    y <- lapply(data.s, function(data) {
        x <- data$x
        offset <- data$offset[1]
        sy <- c(
            p1.fun(offset),
            p2.fun(offset),
            p3.fun(offset),
            p4.fun(offset),
            p5.fun(offset),
            p6.fun(offset)
        )
        f.spline <- splinefun(sx, sy)
        y <- rep(NA, length(x))

        # Linear extrapolation below the first anchor using the local slope
        idx <- x < sx[1]
        if (any(idx)) {
            y[idx] <- f.spline(sx[1], deriv = 1) * (x[idx] - sx[1]) + sy[1]
        }

        # Linear extrapolation above the last anchor
        idx <- x > sx[length(sx)]
        if (any(idx)) {
            y[idx] <- f.spline(sx[length(sx)], deriv = 1) * (x[idx] - sx[length(sx)]) + sy[length(sy)]
        }

        # Spline interpolation within the anchor interval
        idx <- x >= sx[1] & x <= sx[length(sx)]
        if (any(idx)) {
            y[idx] <- f.spline(x[idx])
        }

        y - p0.fun(offset)
    })

    unsplit(y, data.f)
}

# NOTE: Mirror any edits of this table into `R/vmideal_vst.R::vmidealVstFromMagn`.
# The runtime code uses a literal copy of these coefficients.
param.df <- structure(list(limmag = c(5.5, 5.52, 5.55, 5.6, 5.7, 5.8, 5.9,
    6, 6.1, 6.2, 6.3, 6.4, 6.45, 6.48, 6.5), `1` = c(0.379578706193683,
    0.380865978506213, 0.383646581863156, 0.386594955357005, 0.384516871380943,
    0.378119364136071, 0.39783016804395, 0.373393379287508, 0.366393502192895,
    0.345593857555874, 0.367086720870886, 0.359195376100079, 0.347082537633701,
    0.312420722892609, 0.312232925285191), `2` = c(0.758955184180252,
    0.757434042449928, 0.756545241530695, 0.754597324053272, 0.746096530130275,
    0.736273688899513, 0.754472950700777, 0.729839979801878, 0.724090251762525,
    0.706272163875503, 0.732424006749156, 0.730856947686901, 0.722541149425633,
    0.690367122987635, 0.691926213129086), `3` = c(1.89943177726206,
    1.89844069948433, 1.89825430762927, 1.89725244849407, 1.88997593310439,
    1.88072563899649, 1.89899661183851, 1.87396958675163, 1.86749053248022,
    1.84882849631795, 1.87410789172307, 1.87172629242088, 1.86305301012132,
    1.83068592572223, 1.83212684285442), `4` = c(3.27014476551314,
    3.26892787903128, 3.26844009590108, 3.26702893292816, 3.25921320867758,
    3.24969559483436, 3.26789784387066, 3.24296948683872, 3.23673983810267,
    3.21843295052407, 3.24413734533724, 3.24221462073629, 3.23377186059092,
    3.20154167652271, 3.20307252532895), `5` = c(4.50424034207103,
    4.50305721713432, 4.50261471059499, 4.5012654394704, 4.49353145072154,
    4.48405231361262, 4.50225758259143, 4.47729738300568, 4.47099710887246,
    4.45259464315401, 4.47819774376931, 4.4761817895577, 4.46769920746914,
    4.43544755919314, 4.43696568503357), `6` = c(5.65036712145262,
    5.64927790140628, 5.64896064069461, 5.6477824229466, 5.64028041416121,
    5.63092781192841, 5.64918249953455, 5.62420382559607, 5.61781703615347,
    5.5992741104315, 5.62469988859743, 5.62248449312786, 5.61389842883615,
    5.58158409117071, 5.58306055049816), offset = c(-0.5, -0.48,
    -0.45, -0.4, -0.3, -0.2, -0.1, 0, 0.1, 0.2, 0.3, 0.4, 0.45, 0.48,
    0.5), intercept = c(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0)), row.names = c("5.5", "5.52", "5.55", "5.6", "5.7", "5.8",
    "5.9", "6", "6.1", "6.2", "6.3", "6.4", "6.45", "6.48", "6.5"
    ), class = "data.frame")

if (FALSE) {
    # Estimate the anchor values (columns '1'..'6') per limmag by minimizing
    # deviation of t-variance from a target (here 1.0) across psi.
    param.df <- with(new.env(), {
        limmag <- c(5.5, 5.52, 5.55, seq(5.6, 6.4, 0.1), 6.45, 6.48, 6.5)
        psi <- c(-100, seq(-4, 9, 0.25), 100)

        f.model <- function(param.df, psi, limmag) {
            model <- expand.grid(psi = psi, limmag = limmag)
            model <- do.call(
                rbind.data.frame,
                mapply(function(psi, limmag) {
                    p <- dvmideal(m, limmag, psi)
                    t <- myVmidealVstFromMagn(m, limmag, param.df)
                    t.mean <- sum(p * t)
                    t.var <- sum(p * (t - t.mean)^2)

                    list(
                        t.mean = t.mean,
                        t.var = t.var
                    )
                }, model$psi, model$limmag , SIMPLIFY = FALSE)
            )
        }

        param.list <- list()
        for (l in limmag) {
            print(paste('limmag:', l))
            params <- as.vector(unlist(param.df[as.character(l), as.character(seq(1, 6))]))
            opt.res <- optim(params, fn = function(params, t.var.mean) {
                param.row <- as.data.frame(t(params))
                param.df <- rbind.data.frame(param.row, param.row)
                colnames(param.df) <- seq(1, length(params))
                param.df$intercept <- 0.0
                param.df$offset <- l - round(l)
                param.df$offset[1] <- param.df$offset[1] - 0.01
                param.df$offset[2] <- param.df$offset[2] + 0.01

                df <- f.model(param.df, psi, l)
                sum((df$t.var - t.var.mean)^2)/nrow(df)
            }, t.var.mean = 1.0)
            params  <- c(l, opt.res[["par"]])
            param.list[[as.character(l)]] <- params
        }
        param.df <- do.call(
            rbind.data.frame,
            as.list(param.list)
        )
        colnames(param.df) <- c('limmag', as.character(seq(1, 6)))
        param.df <- param.df[order(param.df$limmag),]
        row.names(param.df) <- as.character(param.df$limmag)
        param.df
    })
    param.df$offset <- round(param.df$limmag - round(param.df$limmag), 2)
    param.df$intercept <- 0.0
    dput(param.df)
}

#' Generate E[t] and Var[t] over grids of psi and limmag for vmideal
#'
#' Arguments
#' - param.df: parameter frame with columns 'offset', '1'..'6', 'intercept'.
#' - limmag: vector of limiting magnitudes.
#' - psi: vector of vmideal shape parameters.
#'
#' Returns
#' - data.frame with columns psi, limmag, t.mean, t.var.
data.t.fun <- function(param.df, limmag, psi) {
    model <- expand.grid(psi = psi, limmag = limmag)
    model <- do.call(
        rbind.data.frame,
        mapply(function(psi, limmag) {
            p <- dvmideal(m, limmag, psi)
            t <- myVmidealVstFromMagn(m, limmag, param.df)
            t.mean <- sum(p * t)
            t.var <- sum(p * (t - t.mean)^2)

            list(
                psi = psi,
                limmag = limmag,
                t.mean = t.mean,
                t.var = t.var
            )
        }, model$psi, model$limmag , SIMPLIFY = FALSE)
    )
}

# Intercept calibration (example MSE ~ 8.696629e-05 during tuning)
# Intercepts are offset-specific shifts; keep them synchronised with the
# constant column in `vmidealVstFromMagn`.
param.df$intercept <- c(0.0318278656993037, 0.0299076409185304, 0.0284655814974199,
    0.025667223534113, 0.0158031837415663, 0.00504408335887503, 0.0226347191720595,
    -0.00235692008208782, -0.0082983458149825, -0.0258911028895704,
    0.000925136964883697, 0.000504234344480941, -0.00707058067955687,
    -0.0387371640018533, -0.0367814650209238
)
limmag <- c(5.5, 5.52, 5.55, seq(5.6, 6.4, 0.1), 6.45, 6.48, 6.5)
psi <- c(-100, seq(-10, 9, 0.25))

if (FALSE) {
    # Fine‑tune the intercept per offset by minimizing squared error in the
    # regression psi - limmag ~ poly(log(E[t]), 8).
    opt.res <- optim(param.df$intercept, fn = function(intercept) {
        param.df$intercept <- intercept
        data.t <- data.t.fun(param.df, limmag, psi)
        data.t$x <- log(data.t$t.mean - min(data.t.fun(param.df, limmag, 100)$t.mean) + 1e-05)
        data.t$y <- data.t$psi - data.t$limmag
        lm.res <- lm(y ~ poly(x, 8, raw=TRUE), data = data.t)
        data.t$y.predicted <- predict(lm.res)
        print(sum((data.t$y - data.t$y.predicted)^2)/nrow(data.t))
    }, control = list(maxit = 250))
    dput(opt.res[["par"]])
    param.df$intercept <- opt.res[["par"]]
}
# Align intercepts by adding the minimal E[t] at high psi so that log(E[t])
# is well‑behaved for the subsequent regression/inversion mapping.
param.df$intercept <- param.df$intercept + min(data.t.fun(param.df, limmag, 1000)$t.mean)

#' Map x = E[t] to psi via polynomial in log(x)
#'
#' Arguments
#' - x: positive numeric vector (E[t]). Values outside [0.016, 8.22] are set NA.
#' - limmag: corresponding limiting magnitudes.
#' - poly.coef: numeric coefficients of a polynomial in log(x). The k-th entry
#'   corresponds to exponent (k-1).
#' - deriv: if TRUE, return derivative d psi / d x (scaled by 1/x) using the
#'   derivative polynomial.
#'
#' Returns
#' - psi (same shape as x) or its derivative if deriv = TRUE.
myVmidealVstToPsi <- function(x, limmag, poly.coef, deriv = FALSE) {
    names(poly.coef) <- seq_along(poly.coef) - 1 # exponents
    # x min 0.016 (psi approx 9 at limiting maginitde of 5.5)
    # x max 8.22(psi approx -10 at limiting maginitde of 6.5)
    x[x < 0.016 | x > 8.22] <- NA
    x <- log(x)

    if(deriv) {
        poly.coef1 <- vismeteor:::f.polynomial.coef(poly.coef, deriv.degree = 1L)
        vismeteor:::f.polynomial(x, poly.coef1) / x
    } else {
        limmag + vismeteor:::f.polynomial(x, poly.coef)
    }
}

# evaluate vmideal expectations with the current spline/intercept set.
data.t <- data.t.fun(param.df, limmag, psi)
data.t$psi.delta <- data.t$psi - data.t$limmag
data.t$x <- log(data.t$t.mean)
data.t$y <- data.t$psi - data.t$limmag
lm.res <- lm(y ~ poly(x, 8, raw=TRUE), data = data.t)
lm.coeff <- coefficients(lm.res)
names(lm.coeff) <- NULL
dput(lm.coeff)  # Copy this numeric vector into `R/vmideal_vst.R::vmidealVstToPsi`.

data.t$y.predicted <- predict(lm.res)
data.t$y.diff <- data.t$y.predicted - data.t$y
data.t$psi.est <- myVmidealVstToPsi(data.t$t.mean, data.t$limmag, lm.coeff)
#data.t$psi.est <- vmidealVstToPsi(data.t$t.mean, data.t$limmag)
data.t$psi.est.diff <- round(data.t$psi.est - data.t$psi, 3)
#View(data.t)

## ----setup, include = FALSE---------------------------------------------------
library(vismeteor)
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)

## ----echo=TRUE, results='hide'------------------------------------------------
observations <- with(PER_2015_magn$observations, {
    idx <- !is.na(lim_magn) & sl_start > 135.81 & sl_end < 135.87
    data.frame(
        magn_id = magn_id[idx],
        lim_magn = lim_magn[idx]
    )
})
head(observations, 5) # Example values

## ----echo=FALSE, results='asis'-----------------------------------------------
knitr::kable(head(observations, 5))

## ----echo=TRUE, results='hide'------------------------------------------------
magnitudes <- with(new.env(), {
    magnitudes <- merge(
        observations,
        as.data.frame(PER_2015_magn$magnitudes),
        by = "magn_id"
    )
    magnitudes$magn <- as.integer(as.character(magnitudes$magn))
    subset(magnitudes, (magnitudes$lim_magn - magnitudes$magn) > -0.5)
})
head(magnitudes, 5) # Example values

## ----echo=FALSE, results='asis'-----------------------------------------------
knitr::kable(head(magnitudes[magnitudes$Freq > 0, ], 5))

## ----echo=TRUE, results='hide'------------------------------------------------
# maximum likelihood estimation (MLE) of r
result_ml <- with(magnitudes, {
    # log likelihood function
    ll <- function(r) -sum(Freq * dvmgeom(magn, lim_magn, r, log = TRUE))
    r_start <- 2.0 # starting value
    r_lower <- 1.2 # lowest expected value
    r_upper <- 4.0 # highest expected value
    # find minimum
    optim(r_start, ll, method = "Brent", lower = r_lower, upper = r_upper, hessian = TRUE)
})

## ----echo=TRUE----------------------------------------------------------------
print(result_ml$par) # mean of r
print(1 / result_ml$hessian[1][1]) # variance of r

## ----echo=TRUE, results='hide'------------------------------------------------
with(new.env(), {
    data_plot <- data.frame(r = seq(2.0, 2.8, 0.01))
    data_plot$ll <- mapply(function(r) {
        with(magnitudes, {
            # log likelihood function
            sum(Freq * dvmgeom(magn, lim_magn, r, log = TRUE))
        })
    }, data_plot$r)
    data_plot$l <- exp(data_plot$ll - max(data_plot$ll))
    data_plot$l <- data_plot$l / sum(data_plot$l)
    brks <- seq(min(data_plot$r) - 0.02, max(data_plot$r) + 0.02, by = 0.02)
    plot(data_plot$r, data_plot$l,
        # breaks = brks,
        type = "l",
        col = "blue",
        xlab = "r",
        xaxt = "n",
        ylab = "likelihood"
    )
    xlabels <- seq(min(round(data_plot$r, 1)) - 0.1, max(round(data_plot$r, 1)) + 0.1, by = 0.1)
    axis(
        side = 1,
        at = xlabels,
        labels = sprintf("%.1f", xlabels)
    )
    abline(v = result_ml$par, col = "red", lwd = 1)
})

## ----echo=TRUE, results='hide'------------------------------------------------
result_rate <- with(magnitudes, {
    N <- sum(Freq)
    a <- vmperception(lim_magn - magn - 1) / vmperception(lim_magn - magn)
    a_mean <- as.numeric(weighted.mean(a, w = Freq))
    a_var <- as.numeric(cov.wt(cbind(a), wt = Freq)$cov) / N
    # apply the delta method and return the result
    list(
        "mean" = 1 / a_mean - a_var / a_mean^3,
        "var" = a_var / a_mean^4
    )
})

## ----echo=TRUE----------------------------------------------------------------
print(result_rate$mean) # mean of r
print(result_rate$var) # variance of r

## ----echo=TRUE, results='hide'------------------------------------------------
# Bootstrapping Method
r_means <- with(magnitudes, {
    N <- sum(Freq)
    a <- vmperception(lim_magn - magn - 1) / vmperception(lim_magn - magn)
    replicate(50000, {
        s <- sample(a, size = N, replace = TRUE, prob = Freq)
        s_mean <- mean(s)
        s_var <- var(s) / N
        1 / s_mean - s_var / s_mean^3
    })
})

## ----echo=TRUE, results='show'------------------------------------------------
with(new.env(), {
    r_sd <- sqrt(result_rate$var)
    r_min <- result_rate$mean - 3 * r_sd
    r_max <- result_rate$mean + 3 * r_sd
    r <- subset(r_means, r_means > r_min & r_means < r_max)
    brks <- seq(min(r) - 0.02, max(r) + 0.02, by = 0.02)
    hist(r,
        breaks = brks,
        col = "skyblue",
        border = "black",
        main = "Histogram of mean r",
        xlab = "r",
        xaxt = "n",
        ylab = "count"
    )
    xlabels <- seq(min(round(r, 1)) - 0.1, max(round(r, 1)) + 0.1, by = 0.1)
    axis(
        side = 1,
        at = xlabels,
        labels = sprintf("%.1f", xlabels)
    )
    abline(v = result_rate$mean, col = "red", lwd = 1)
})

## ----echo=TRUE, results='show'------------------------------------------------
result_vs <- with(magnitudes, {
    N <- sum(Freq)
    tm <- vmgeom_vst_from_magn(magn, lim_magn)
    tm_mean <- sum(Freq * tm) / N
    tm_var <- sum(Freq * (tm - tm_mean)^2) / (N - 1)
    tm_mean_var <- tm_var / N

    # Delta method: variance and variance of r
    r_hat <- vmgeom_vst_to_r(tm_mean)
    dr_dtm <- vmgeom_vst_to_r(tm_mean, deriv_degree = 1L)
    d2r_dtm2 <- vmgeom_vst_to_r(tm_mean, deriv_degree = 2L)
    r_hat <- r_hat - 0.5 * d2r_dtm2 * tm_mean_var
    r_var <- dr_dtm^2 * tm_mean_var + 0.5 * d2r_dtm2^2 * tm_mean_var^2

    list("mean" = r_hat, "var" = r_var)
})

## ----echo=TRUE, results='show'------------------------------------------------
print(paste("r mean:", result_vs$mean))
print(paste("r var:", result_vs$var))

## ----echo=TRUE, results='hide'------------------------------------------------
# Bootstrapping Method
r_means <- with(magnitudes, {
    N <- sum(Freq)
    tm <- vmgeom_vst_from_magn(magn, lim_magn)
    replicate(50000, {
        s <- sample(tm, size = N, replace = TRUE, prob = Freq)
        s_mean <- mean(s)
        s_var <- var(s) / N

        r_hat <- vmgeom_vst_to_r(s_mean)
        d2r_ds2 <- vmgeom_vst_to_r(s_mean, deriv_degree = 2L)
        r_hat - 0.5 * d2r_ds2 * s_var
    })
})

## ----echo=TRUE, results='show'------------------------------------------------
with(new.env(), {
    r_sd <- sqrt(result_vs$var)
    r_min <- as.vector(result_vs$mean - 3 * r_sd)
    r_max <- as.vector(result_vs$mean + 3 * r_sd)
    r <- subset(r_means, r_means > r_min & r_means < r_max)
    brks <- seq(min(r) - 0.02, max(r) + 0.02, by = 0.02)
    hist(r,
        breaks = brks,
        col = "skyblue",
        border = "black",
        main = "Histogram of mean r",
        xlab = "r",
        xaxt = "n",
        ylab = "count"
    )
    xlabels <- seq(min(round(r, 1)) - 0.1, max(round(r, 1)) + 0.1, by = 0.1)
    axis(
        side = 1,
        at = xlabels,
        labels = sprintf("%.1f", xlabels)
    )
    abline(v = result_vs$mean, col = "red", lwd = 1)
})

## ----echo=TRUE, results='hide'------------------------------------------------
magnitudes$p <- with(magnitudes, dvmgeom(m = magn, lm = lim_magn, result_rate$mean))

## ----echo=TRUE, results='hide'------------------------------------------------
magn_min <- min(magnitudes$magn)

## ----echo=TRUE, results='asis'------------------------------------------------
idx <- magnitudes$magn == magn_min
magnitudes$p[idx] <- with(
    magnitudes[idx, ],
    pvmgeom(m = magn + 1L, lm = lim_magn, result_rate$mean, lower.tail = TRUE)
)

## ----echo=TRUE----------------------------------------------------------------
magnitutes_observed <- xtabs(Freq ~ magn_id + magn, data = magnitudes)
magnitutes_observed_mt <- margin.table(magnitutes_observed, margin = 2)
print(magnitutes_observed_mt)

## ----echo=TRUE----------------------------------------------------------------
magnitudes$magn[magnitudes$magn <= 0] <- "0-"
magnitudes$magn[magnitudes$magn >= 4] <- "4+"
magnitutes_observed <- xtabs(Freq ~ magn_id + magn, data = magnitudes)
print(margin.table(magnitutes_observed, margin = 2))

## ----echo=TRUE----------------------------------------------------------------
magnitutes_expected <- xtabs(p ~ magn_id + magn, data = magnitudes)
magnitutes_row_freq <- margin.table(magnitutes_observed, margin = 1)
magnitutes_expected <- sweep(magnitutes_expected, 1, magnitutes_row_freq, `*`)
magnitutes_expected <- magnitutes_expected / sum(magnitutes_expected)
print(sum(magnitudes$Freq) * margin.table(magnitutes_expected, margin = 2))

## ----echo=TRUE, results='asis'------------------------------------------------
chisq_test_result <- chisq.test(
    x = margin.table(magnitutes_observed, margin = 2),
    p = margin.table(magnitutes_expected, margin = 2)
)

## ----echo=TRUE----------------------------------------------------------------
chi2_df <- chisq_test_result$parameter - 1
chi2_pval <- pchisq(chisq_test_result$statistic, df = chi2_df, lower.tail = FALSE)
print(chi2_pval)

## ----fig.show='hold'----------------------------------------------------------
chisq_test_residuals <- with(new.env(), {
    chisq_test_residuals <- residuals(chisq_test_result)
    v <- as.vector(chisq_test_residuals)
    names(v) <- names(chisq_test_residuals)
    v
})
plot(
    chisq_test_residuals,
    main = "Residuals of the chi-square goodness-of-fit test",
    xlab = "m",
    ylab = "Residuals",
    ylim = c(-3, 3),
    xaxt = "n"
)
abline(h = 0.0, lwd = 2)
axis(1, at = seq_along(chisq_test_residuals), labels = names(chisq_test_residuals))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(vismeteor)

## ----echo=TRUE----------------------------------------------------------------
data(PER_2015_rates)
obs <- PER_2015_rates$observations |> subset(
    rad_alt > 15 & moon_alt < 0 & sun_alt < -5 &
        sl_start >= 139.2 & sl_end <= 140.6
)

deg2rad <- \(x) x * pi / 180
obs$sl <- (obs$sl_start + obs$sl_end) / 2
obs$t <- with(obs, t_eff * sin(deg2rad(rad_alt)) / f)

rates <- data.frame(
    sl = obs$sl,
    t = obs$t,
    limmag = obs$lim_magn,
    freq = obs$freq
)
rates <- rates[order(rates$sl), ]
nrow(rates)

## ----echo=TRUE----------------------------------------------------------------
fit_glm <- \(rates, knots) {
    f <- if (length(knots) == 0L) {
        freq ~ splines::ns(sl) + limmag + offset(log(t))
    } else {
        freq ~ splines::ns(sl, knots = knots) +
            limmag + offset(log(t))
    }
    stats::glm(f, data = rates, family = stats::poisson())
}
score_bic <- \(rates, knots) stats::BIC(fit_glm(rates, knots))

## ----echo=TRUE----------------------------------------------------------------
rates$q <- vismeteor::freq_quantile(rates$freq, 12) # at least 12 meteors per bin
rates$qsl <- with(
    rates,
    ave(freq * sl, q, FUN = sum) / ave(freq, q, FUN = sum)
)
cand <- sort(unique(round(rates$qsl, 2)))
length(cand)

## ----echo=TRUE----------------------------------------------------------------
res_fwd <- vismeteor::select_knots(rates, cand, score_bic)
sort(c(res_fwd$knots, res_fwd$fixed_knots))
res_fwd$score

## ----echo=TRUE, fig.alt="BIC trajectory of the forward search over the number of interior knots, with the score optimum at the endpoint"----
plot(res_fwd$history$n_knots, res_fwd$history$score,
    type = "b", pch = 19, col = "blue",
    xlab = "number of interior knots", ylab = "BIC",
    main = "Forward selection: BIC trajectory"
)
points(tail(res_fwd$history$n_knots, 1L), res_fwd$score,
    pch = 19, col = "red", cex = 1.5
)
abline(h = res_fwd$score, lty = 2, col = "grey50")

## ----echo=TRUE----------------------------------------------------------------
res_fwd_plus1 <- vismeteor::select_knots(rates, cand, score_bic,
    fixed_knots = c(
        res_fwd$knots,
        res_fwd$fixed_knots
    ),
    n_steps = 1L
)
res_fwd_plus1$score - res_fwd$score # always >= 0: how much worse one step makes it

## ----echo=TRUE----------------------------------------------------------------
res_bwd <- vismeteor::select_knots(rates, cand, score_bic, backward = TRUE)
sort(c(res_bwd$knots, res_bwd$fixed_knots))
res_bwd$score

## ----echo=TRUE, fig.alt="BIC trajectory of the backward search over the number of interior knots, with the score optimum at the endpoint"----
plot(res_bwd$history$n_knots, res_bwd$history$score,
    type = "b", pch = 19, col = "blue",
    xlab = "number of interior knots", ylab = "BIC",
    main = "Backward elimination: BIC trajectory"
)
points(tail(res_bwd$history$n_knots, 1L), res_bwd$score,
    pch = 19, col = "red", cex = 1.5
)
abline(h = res_bwd$score, lty = 2, col = "grey50")

## ----echo=TRUE----------------------------------------------------------------
best <- if (res_fwd$score <= res_bwd$score) {
    c(res_fwd$knots, res_fwd$fixed_knots)
} else {
    c(res_bwd$knots, res_bwd$fixed_knots)
}
fit <- fit_glm(rates, sort(best))

# Data-driven population index of the shower, with 95% Wald CI
# (log-scale CI on the limmag coefficient, exponentiated).
limmag_row <- summary(fit)$coefficients["limmag", ]
r_hat <- exp(limmag_row["Estimate"])
r_ci <- exp(limmag_row["Estimate"] +
    c(-1, 1) * 1.96 * limmag_row["Std. Error"])
c(
    estimate = unname(r_hat),
    lower = unname(r_ci[1]),
    upper = unname(r_ci[2])
)

## ----echo=TRUE, fig.alt="Fitted ZHR curve over solar longitude with a 95% confidence band and selected knots marked on the axis"----
sl_grid <- seq(139.2, 140.6, length.out = 400)
pred_df <- data.frame(sl = sl_grid, limmag = 6.5, t = 1)
link <- predict(fit, newdata = pred_df, type = "link", se.fit = TRUE)
zhr <- exp(link$fit)
zhr_lo <- exp(link$fit - 1.96 * link$se.fit)
zhr_hi <- exp(link$fit + 1.96 * link$se.fit)

plot(sl_grid, zhr,
    type = "n",
    ylim = range(c(zhr_lo, zhr_hi)),
    xlab = "solar longitude (deg)", ylab = "ZHR",
    main = "PER 2015 - fitted activity profile (95% CI)"
)
polygon(c(sl_grid, rev(sl_grid)),
    c(zhr_lo, rev(zhr_hi)),
    col = adjustcolor("blue", alpha.f = 0.2), border = NA
)
lines(sl_grid, zhr, col = "blue", lwd = 2)
rug(best, col = "red")

## ----echo=TRUE, fig.alt="Binned observed-over-expected ratios with one-sigma error bars over solar longitude"----
# Use a coarser binning here than for the candidate grid above
# (>= 500 meteors per bin) so that bin uncertainties are small enough
# to read systematic deviations from the diagnostic plot.
rates$qres <- vismeteor::freq_quantile(rates$freq, 500)
rates$pred <- predict(fit, type = "response")

bins <- with(rates, {
    obs_n <- tapply(freq, qres, sum)
    pred_n <- tapply(pred, qres, sum)
    data.frame(
        sl    = tapply(sl * freq, qres, sum) / obs_n,
        ratio = obs_n / pred_n,
        se    = sqrt(obs_n) / pred_n
    )
})
nrow(bins)

plot(bins$sl, bins$ratio,
    pch = 19, col = "blue",
    ylim = range(c(
        bins$ratio - 2 * bins$se,
        bins$ratio + 2 * bins$se
    )),
    xlab = "solar longitude (deg)",
    ylab = "observed / expected",
    main = "PER 2015 - binned residual ratios (+/- 1 sigma)"
)
arrows(bins$sl, bins$ratio - bins$se,
    bins$sl, bins$ratio + bins$se,
    angle = 90, code = 3, length = 0.03, col = "blue"
)
abline(h = 1, lty = 2, col = "grey50")

#' @title Greedy stepwise knot selection for a regression spline
#' @description
#' Selects a parsimonious subset of \code{knot_candidates} as interior knots
#' for a regression spline by greedy forward addition or backward elimination,
#' scored by a user-supplied criterion (BIC, AIC, or any cross-validation
#' function). The function is model-agnostic: it only chooses which
#' knot positions to include and passes that vector to your \code{score_fun};
#' your \code{score_fun} decides which basis (\code{splines::ns},
#' \code{splines::bs} of any degree, …) and which model family to fit.
#' Typical use: fitting a smooth, parsimonious trend through noisy data — for
#' example an activity profile of a meteor shower across solar longitude —
#' where the number and position of knots should remain small and inspectable.
#'
#' @details
#' The typical use case --- meteor shower rates over the solar longitude or
#' time, for example --- is a smooth, slowly varying signal whose shape is
#' not known in advance and whose curvature changes locally: too rigid for
#' a low-order polynomial, too noisy for a histogram. A regression spline
#' (typically cubic, e.g. \code{splines::ns} or
#' \code{splines::bs(degree = 3)}) with a modest number of well-placed
#' interior knots gives a smooth, locally flexible fit with interpretable
#' degrees of freedom. Picking that number is a bias/variance trade-off:
#' too many knots overfit (ringing, unstable derivatives, slow fits), too
#' few introduce bias. \code{select_knots()} automates the trade-off by
#' greedy forward addition or backward elimination, scored by a
#' user-defined criterion; you control which knots are even allowed
#' (\code{knot_candidates}) and what counts as "better" (\code{score_fun}).
#'
#' Your \code{score_fun} must take \code{(data, knots)} and return a single
#' \code{numeric} value; \emph{lower is better}. Typically it fits a model
#' with these interior knots (\code{length(knots) == 0L} means "no interior
#' knots") and reports an information criterion (\code{stats::BIC},
#' \code{stats::AIC}) or a held-out score. \code{knots} arrives sorted, so
#' the fit code does not need to sort it again. A divergent or failed fit
#' should ideally return \code{Inf}; the function additionally wraps each
#' call in \code{tryCatch()} and treats errors as \code{Inf} so the search
#' continues robustly. See \code{Examples} for a runnable template.
#'
#' Backward elimination can drop multiple knots per round in "bulk" mode.
#' When \code{backward = TRUE} and \code{bulk_gap >= 1L}, each round removes
#' several well-separated knots at once (minimum index gap \code{bulk_gap}
#' in the current sorted knot list). For a B-spline basis of degree
#' \code{d}, each basis function has support over \code{d + 1} consecutive
#' knot intervals, so removing knots that are at least \code{d + 1}
#' positions apart has nearly additive effect on the score --- giving
#' roughly a \code{bulk_gap}-fold speed-up at the cost of a small
#' approximation (mitigated by a verify-fit after each bulk round). The
#' default \code{4L} matches \code{d + 1} for cubic bases such as
#' \code{splines::ns} or \code{splines::bs(degree = 3)}; for other bases
#' pick \code{bulk_gap = degree + 1}, or set \code{bulk_gap = 0L} for
#' strict one-knot-per-round behaviour.
#'
#' The result splits the final knot set into two disjoint vectors:
#' \code{knots} (positions the algorithm itself selected) and
#' \code{fixed_knots} (the user-supplied anchors, echoed back, deduplicated
#' and sorted). The full vector to fit a model on is
#' \code{c(knots, fixed_knots)} (or its sorted form). With the default
#' \code{n_steps = NULL} the loop stops as soon as no further move improves
#' the score, so the end state \emph{is} the score-optimum; with
#' \code{n_steps > 0L} the loop runs that many iterations regardless of
#' improvement, and the score-best point along the trajectory is recoverable
#' from \code{history} via the row at \code{which.min(history$score)}.
#'
#' If the starting state already is (locally) optimal --- typical when
#' \code{fixed_knots} alone overfits in forward mode, or when
#' \code{knot_candidates} is so small that the full pool already minimises
#' the score in backward mode --- the \code{n_steps = NULL} run terminates
#' immediately with \code{history} containing only the initial row and
#' \code{score} equal to the starting score; an explicit \code{n_steps = N}
#' run will instead take \code{N} worsening steps so the post-optimum
#' landscape is still observable.
#'
#' Knots sit next to extrema, not on them. \code{select_knots()} chooses
#' knots for a \emph{good fit}, not for detecting extrema of the response.
#' Knots typically land next to peaks or troughs --- where the curvature is
#' highest --- and not on them. Read local extrema off the \emph{shape} of
#' the fitted curve, not from \code{knots}. A knot supplied
#' through \code{fixed_knots} is the exception: it sits wherever the user
#' puts it, since it is a constraint imposed on the search rather than a
#' finding produced by it.
#'
#' @section When to use this:
#' \code{select_knots()} is for situations where knot \emph{positions} are
#' meaningful and you want a small, inspectable set of interior knots scored
#' under a criterion you choose. If you instead want a continuous roughness
#' penalty over a dense knot grid, the \pkg{mgcv} package's penalised
#' B-splines (\code{bs = "ps"}) or adaptive smoothers (\code{bs = "ad"}) are
#' usually a better fit; for greedy stepwise selection on a hinge-function
#' basis, see the \pkg{earth} package.
#'
#' Like every greedy stepwise procedure, \code{select_knots()} performs a
#' \emph{local} search: in each round it commits to the locally best
#' add/drop. It therefore returns a local optimum of the score, which is
#' usually but not provably the global optimum --- a knot \emph{combination}
#' that beats every individually-best move can be unreachable once an
#' earlier, locally-attractive knot has been picked. The only way to
#' guarantee globality would be exhaustive enumeration over all
#' \eqn{2^{|knot\_candidates|}} subsets, which is exponential and
#' impractical for any realistic candidate pool. In practice the greedy
#' optimum is close; repeating the search in the opposite direction
#' (\code{backward = TRUE}) is a cheap robustness check.
#'
#' @param data Data passed unchanged to \code{score_fun}; typically a
#'   \code{data.frame}. The function itself does not inspect or mutate it.
#' @param knot_candidates Numeric vector of candidate interior-knot positions.
#'   Duplicates and unsorted input are tolerated (sorted/uniq'd internally).
#' @param score_fun \code{\\(data, knots) -> numeric scalar}. Lower is
#'   better. The caller is responsible for fitting whatever model they want
#'   and returning a single criterion value (BIC, AIC, cross-validation
#'   error, ...).
#' @param backward Logical. \code{FALSE} (default) = forward selection,
#'   \code{TRUE} = backward elimination.
#' @param n_steps \code{NULL} (default) or a positive integer.
#'   \code{NULL} runs the greedy search until the next move would not
#'   improve the score and stops there --- so the end state of the search
#'   IS the score optimum (under the current direction and constraints).
#'   A positive integer \code{N} runs exactly \code{N} iterations regardless
#'   of whether each one improves the score; this is a deliberate
#'   exploration mode, intended for inspecting the score landscape just
#'   past a previously-found optimum. The canonical recipe is to first
#'   call with \code{n_steps = NULL} to find the optimum, then re-call with
#'   \code{fixed_knots = prev$knots} and
#'   \code{n_steps = 1L} (or 2L, 3L) to take one or more controlled steps
#'   onward and see how rapidly the score deteriorates. The value
#'   \code{0L} is not allowed (use \code{NULL} for "no steps past the
#'   optimum"); non-integer or negative values error out.
#' @param bulk_gap Integer (>= 0). Minimum index gap between knots removed in
#'   the same round of backward elimination. Ignored when
#'   \code{backward = FALSE}. \code{0L} disables bulk removal. The default
#'   \code{4L} matches the support of a cubic spline basis
#'   (\code{degree + 1} for cubic B-splines; the same value also fits
#'   \code{splines::ns}, which is cubic by construction); for higher-degree
#'   B-splines pick \code{degree + 1} accordingly. Note that
#'   \code{bulk_gap = 1L} imposes no real separation constraint and therefore
#'   removes \emph{every} improving knot in one round -- a very aggressive
#'   setting that defeats the per-round verify-fit's role as a safety net.
#' @param fixed_knots Numeric vector of knot positions that must be present
#'   in every fitted model during the search. In forward mode they are
#'   set from the start, and further knots may be added on top; in
#'   backward mode they are never proposed for removal, neither singly nor
#'   in bulk. Need not be a subset of \code{knot_candidates} -- included
#'   regardless. Duplicates and unsorted input are tolerated. Default
#'   \code{numeric(0)} (no fixed knots).
#' @param verbose Logical. If \code{TRUE}, prints per-round progress
#'   (\code{cat()} to stdout). Default \code{FALSE}.
#' @param n_cores Integer (>= 1). Performance-only knob: the per-round
#'   candidate scoring runs in parallel across \code{n_cores} workers. Default
#'   \code{1L} (serial, no extra dependency). When \code{n_cores > 1L} the
#'   base-R package \pkg{parallel} is loaded and \code{parallel::mclapply()}
#'   is used (fork-based on macOS/Linux; falls back to serial on Windows).
#'   Quick recipes:
#'   \itemize{
#'     \item \code{n_cores = 1L} -- safe default, no extra package loaded.
#'     \item \code{n_cores = max(1L, parallel::detectCores() - 1L)} -- use
#'       all cores except one; good default for interactive multi-tasking.
#'     \item \code{n_cores = parallel::detectCores()} -- use every core;
#'       fastest but resource-hungry (the whole machine is busy).
#'   }
#'   The function errors out with a clear message if \code{n_cores > 1L} but
#'   \pkg{parallel} is not installed.
#'
#'   Reproducibility deserves a note when \code{n_cores > 1L}:
#'   \code{mclapply()} is fork-based and inherits the parent RNG state. If
#'   your \code{score_fun} uses randomness (e.g. cross-validation splits),
#'   set \code{RNGkind("L'Ecuyer-CMRG")} and seed via \code{set.seed()}
#'   \emph{before} calling \code{select_knots()} to get reproducible
#'   per-worker streams; otherwise results can differ run to run and across
#'   \code{n_cores}.
#'
#' @return A list with elements (in this order):
#' \describe{
#'   \item{\code{backward}}{Logical — the direction used.}
#'   \item{\code{knots}}{Sorted numeric vector — the interior knots the
#'     algorithm itself selected, with \code{fixed_knots} \emph{excluded}.
#'     Disjoint from \code{fixed_knots} by construction. The full knot
#'     vector to fit a model on is \code{c(knots, fixed_knots)} (or its
#'     sorted form). With the default \code{n_steps = NULL} this is the
#'     score-optimal selection; with \code{n_steps > 0L} it can be a state
#'     past the optimum.}
#'   \item{\code{fixed_knots}}{Sorted numeric vector — the user-supplied
#'     fixed knots echoed back (deduplicated and sorted). Empty vector if
#'     the caller did not supply any.}
#'   \item{\code{score}}{Numeric — the score at the end state, i.e. at
#'     \code{c(knots, fixed_knots)}.}
#'   \item{\code{n_steps}}{The \code{n_steps} value used (\code{NULL} or a
#'     positive integer).}
#'   \item{\code{history}}{\code{data.frame} of per-round records:
#'     \code{step}, \code{n_knots} (total interior knot count including
#'     \code{fixed_knots}), \code{changed_knot} (the knot added or removed
#'     in that round; \code{NA} for the initial state and for bulk-removal
#'     rounds), \code{score}, \code{extra} (boolean: \code{TRUE} when that
#'     step worsened the score). For \code{n_steps = NULL} runs no
#'     worsening step is taken, so \code{extra} is always \code{FALSE}.}
#' }
#'
#' @examples
#' \dontrun{
#' # Greedy knot selection on a simple synthetic signal.
#' set.seed(1)
#' n <- 200
#' x <- seq(0, 10, length.out = n)
#' y <- stats::rpois(n, lambda = 50 + 30 * sin(x))
#' dat <- data.frame(x = x, y = y)
#'
#' # score_fun: fit a Poisson GLM with a natural cubic spline, return BIC.
#' # Lower is better.
#' fit <- \(d, knots) {
#'     f <- if (length(knots) == 0L) {
#'         y ~ splines::ns(x)
#'     } else {
#'         y ~ splines::ns(x, knots = knots)
#'     }
#'     stats::glm(f, family = stats::poisson(), data = d)
#' }
#' score_bic <- \(d, knots) stats::BIC(fit(d, knots))
#'
#' cand <- seq(1, 9, by = 0.5)
#' res <- select_knots(dat, cand, score_bic, verbose = TRUE)
#' # Full knot vector to fit the final model on:
#' final_knots <- sort(c(res$knots, res$fixed_knots))
#' }
#'
#' @seealso \code{\link[splines]{ns}}, \code{\link[splines]{bs}},
#'   \code{\link[stats]{BIC}}, \code{\link[stats]{AIC}}
#' @export
select_knots <- \(data,
    knot_candidates,
    score_fun,
    backward = FALSE,
    n_steps = NULL,
    bulk_gap = 4L,
    fixed_knots = numeric(0),
    verbose = FALSE,
    n_cores = 1L) {
    stopifnot(
        is.function(score_fun),
        is.logical(backward), length(backward) == 1L, !is.na(backward),
        is.logical(verbose), length(verbose) == 1L, !is.na(verbose),
        is.numeric(bulk_gap), length(bulk_gap) == 1L, bulk_gap >= 0L,
        is.null(n_steps) || (is.numeric(n_steps) &&
            length(n_steps) == 1L &&
            !is.na(n_steps) &&
            n_steps >= 1L &&
            n_steps == as.integer(n_steps)),
        is.numeric(n_cores), length(n_cores) == 1L, n_cores >= 1L,
        is.numeric(fixed_knots), !anyNA(fixed_knots),
        all(is.finite(fixed_knots))
    )

    forward <- !backward
    # !duplicated() + boolean indexing instead of unique(): for atomic
    # vectors unique() strips names, which would silently drop the labels
    # of a named knot_candidates / fixed_knots input. The order() variant
    # of sort() likewise preserves names.
    cand_pool <- knot_candidates[!duplicated(knot_candidates)]
    cand_pool <- cand_pool[order(cand_pool)]
    fixed_knots <- fixed_knots[!duplicated(fixed_knots)]
    fixed_knots <- fixed_knots[order(fixed_knots)]
    dir_lbl <- if (backward) "backward" else "forward"
    verb <- if (forward) "add" else "remove"
    bulk_active <- backward && (bulk_gap >= 1L) # bulk only for backward; may flip off mid-run

    # --- private helpers ---
    # Score for one knot set (Inf if the user's score_fun errors out).
    # Knots are sorted before they reach score_fun so the user's fit code
    # never has to defensively sort.
    score_of <- \(knots) {
        tryCatch(
            suppressWarnings(score_fun(data, sort(knots))),
            error = \(e) Inf
        )
    }
    # Scores for many candidate knot sets. Serial (no extra dep) by default;
    # parallel only when explicitly requested via n_cores > 1L.
    par_score <- \(items, make_knots) {
        if (n_cores > 1L) {
            if (!requireNamespace("parallel", quietly = TRUE)) {
                stop("select_knots: n_cores > 1 requires the 'parallel' package; ",
                    "install it or use n_cores = 1L.",
                    call. = FALSE
                )
            }
            res <- parallel::mclapply(items, \(it) score_of(make_knots(it)),
                mc.cores = n_cores
            )
        } else {
            res <- lapply(items, \(it) score_of(make_knots(it)))
        }
        # a crashed worker returns a try-error object -> treat it as Inf
        vapply(res, \(x) if (is.numeric(x) && length(x) == 1L) x else Inf, numeric(1))
    }

    if (forward) {
        current_knots <- fixed_knots
        # cand_pool[!cand_pool %in% fixed_knots] instead of setdiff(): the
        # latter strips names.
        remaining <- cand_pool[!cand_pool %in% fixed_knots]
    } else {
        combined <- c(cand_pool, fixed_knots)
        combined <- combined[!duplicated(combined)]
        current_knots <- combined[order(combined)]
        remaining <- NULL
    }
    current_score <- score_of(current_knots)

    # best_score is tracked internally to control the bulk-removal gate (bulk
    # is only worthwhile while we are still improving). It is not returned;
    # the returned score is the score at the loop's end state.
    best_score <- current_score

    history <- data.frame(
        step         = 0,
        n_knots      = length(current_knots),
        changed_knot = NA_real_,
        score        = current_score,
        extra        = FALSE
    )

    if (verbose) {
        fixed_lbl <- if (length(fixed_knots) > 0L) {
            sprintf(" (%d fixed)", length(fixed_knots))
        } else {
            ""
        }
        cat(sprintf(
            "Start (%s, %d core%s): %d knots%s,  score = %.2f\n",
            dir_lbl, n_cores, if (n_cores == 1) "" else "s",
            length(current_knots), fixed_lbl, current_score
        ))
    }

    step <- 1

    repeat {
        # -- Score of every one-step neighbour of the current knot set (parallel) --
        # In backward mode, fixed knots are excluded from the removal candidates;
        # `removable_idx` maps from score_candidates positions back into
        # current_knots positions so the bulk-gap distance check (which is in
        # current_knots index space, where the d+1 basis-support argument
        # lives) and the single-knot removal both stay correct.
        if (forward) {
            if (length(remaining) == 0) break
            score_candidates <- par_score(remaining, \(k) c(current_knots, k))
            cand_ids <- remaining
            removable_idx <- NULL
        } else {
            removable_idx <- which(!current_knots %in% fixed_knots)
            if (length(removable_idx) == 0) break
            score_candidates <- par_score(
                removable_idx,
                \(i) current_knots[-i]
            )
            cand_ids <- current_knots[removable_idx]
        }

        # -- Bulk removal: drop multiple well-separated improving knots at once.
        #    Active only when going backward, bulk_gap >= 1, and we are still
        #    below the score minimum. Falls through to the single-knot path if
        #    fewer than 2 non-adjacent improving candidates exist, or if the
        #    verify-fit shows no improvement (in which case bulk mode is
        #    permanently disabled for the rest of the run).
        if (bulk_active && min(score_candidates) < best_score) {
            ord <- order(score_candidates)
            improving <- ord[score_candidates[ord] < best_score]
            # picked stores positions in current_knots (mapped via
            # removable_idx) so the gap check matches the d+1 basis-support
            # argument from the docs.
            picked <- integer(0)
            for (idx in improving) {
                pos <- removable_idx[idx]
                if (length(picked) == 0L || all(abs(pos - picked) >= bulk_gap)) {
                    picked <- c(picked, pos)
                }
            }
            if (length(picked) >= 2L) {
                proposed_knots <- current_knots[-picked]
                verify_score <- score_of(proposed_knots)
                if (verify_score < best_score) {
                    if (verbose) {
                        cat(sprintf(
                            "Step %2d: remove %d knots [bulk]  ->  score %.2f -> %.2f  (delta = %.4f)\n",
                            step, length(picked), current_score, verify_score,
                            verify_score - current_score
                        ))
                    }
                    current_knots <- proposed_knots
                    current_score <- verify_score
                    best_score <- current_score
                    history <- rbind(history, data.frame(
                        step         = step,
                        n_knots      = length(current_knots),
                        changed_knot = NA_real_,
                        score        = current_score,
                        extra        = FALSE
                    ))
                    step <- step + 1
                    if (!is.null(n_steps) && step > n_steps) break
                    next
                } else {
                    if (verbose) {
                        cat(sprintf(
                            paste0(
                                "Step %2d: bulk removal of %d knots did not improve ",
                                "(delta = %.4f) -> switching to single-knot mode\n"
                            ),
                            step, length(picked), verify_score - current_score
                        ))
                    }
                    bulk_active <- FALSE
                    # fall through to single-knot path (uses the same score_candidates)
                }
            }
            # length(picked) < 2 -> fall through to single-knot path; keep bulk_active for next round
        }

        best_idx <- which.min(score_candidates)
        cand_score <- score_candidates[best_idx]
        changed_knot <- cand_ids[best_idx]
        improved <- cand_score < best_score

        # NULL-mode: exit before applying a non-improving step, so the loop
        # end state IS the score optimum. The failed candidate is not recorded
        # in history -- use n_steps = 1L for an explicit one-step look-ahead.
        if (is.null(n_steps) && !improved) {
            if (verbose) {
                cat(sprintf(
                    "Step %2d: best %s would worsen score by %.4f -> stopping at optimum\n",
                    step, verb, cand_score - current_score
                ))
            }
            break
        }

        if (verbose) {
            step_lbl <- if (!is.null(n_steps)) {
                sprintf("  [step %d/%d]", step, n_steps)
            } else {
                ""
            }
            cat(sprintf(
                "Step %2d: %s knot %.4f  ->  score %.2f -> %.2f  (delta = %.4f)%s\n",
                step, verb, changed_knot, current_score, cand_score, cand_score - current_score,
                step_lbl
            ))
        }

        if (forward) {
            current_knots <- c(current_knots, changed_knot)
            remaining <- remaining[-best_idx]
        } else {
            current_knots <- current_knots[-removable_idx[best_idx]]
        }
        current_score <- cand_score

        if (improved) best_score <- cand_score

        history <- rbind(history, data.frame(
            step         = step,
            n_knots      = length(current_knots),
            changed_knot = changed_knot,
            score        = current_score,
            extra        = !improved
        ))

        step <- step + 1

        if (!is.null(n_steps) && step > n_steps) break
    }

    free_knots <- current_knots[!current_knots %in% fixed_knots]

    list(
        backward    = backward,
        knots       = free_knots[order(free_knots)],
        fixed_knots = fixed_knots,
        score       = current_score,
        n_steps     = n_steps,
        history     = history
    )
}

#' @name vmgeom_vst
#' @aliases vmgeom_vst_from_magn
#' @aliases vmgeom_vst_to_r
#' @title Variance-Stabilizing Transformation for Geometric Visual Meteor Magnitudes
#'
#' @description
#' Applies a variance-stabilizing transformation to visual meteor magnitudes
#' under the geometric model.
#'
#' @param m integer; meteor magnitude.
#' @param lm numeric; limiting magnitude.
#' @param tm numeric; transformed magnitude.
#' @param log logical; if `TRUE`, the logarithm of the population index `r` is returned.
#' @param deriv_degree integer; the order of the derivative at `tm` to return
#'   instead of `r` or `log(r)`. Must be `0`, `1`, or `2`.
#'
#' @details
#' Many linear models require the variance of visual meteor magnitudes to be
#' homoscedastic. The function `vmgeom_vst_from_magn` applies a transformation
#' that produces homoscedastic distributions of visual meteor magnitudes if the
#' underlying distribution follows a geometric model.
#'
#' The geometric model of visual meteor magnitudes
#' depends on the [population index][vismeteor::vmgeom] `r` and the limiting magnitude `lm`,
#' resulting in a two-parameter distribution. Without detection probabilities,
#' the magnitude distribution is purely geometric, and for integer limiting
#' magnitudes the variance depends only on the population index `r`. Since the
#' limiting magnitude `lm` is a fixed parameter and never estimated
#' statistically, the magnitudes can be transformed such that, for example,
#' the mean of the transformed magnitudes directly provides an estimate of `r`
#' using the function `vmgeom_vst_to_r`.
#'
#' A key advantage of this transformation is that the limiting magnitude `lm`
#' is already incorporated into subsequent analyses. In this sense, the
#' transformation acts as a normalization of meteor magnitudes and yields a
#' variance close to `1.0`.
#'
#' This transformation is valid for \eqn{1.4 \le r \le 3.5}.
#' The numerical form of the transformation is version-specific and may change
#' substantially in future releases. Do not rely on equality of transformed
#' values across package versions.
#'
#' @return
#' * `vmgeom_vst_from_magn`: numeric value, the transformed meteor magnitude.
#' * `vmgeom_vst_to_r`: numeric value of the population index `r`, derived from
#'   the mean of `tm`.
#'
#' The argument `deriv_degree` can be used to apply the delta method.
#' If `log = TRUE`, the logarithm of `r` is returned.
#'
#' @note
#' The internal approximations used here are derived from the perception
#' probabilities produced by [vismeteor::vmperception].
#' For details on the derivation, see the script `inst/derivation/vmgeom_vst.R` in the
#' package's source code.
#'
#' @seealso [vismeteor::vmgeom] [vismeteor::vmperception]
#' @examples
#' N <- 100
#' r <- 2.0
#' limmag <- 6.3
#'
#' # Simulate magnitudes
#' m <- rvmgeom(N, limmag, r)
#'
#' # Variance-stabilizing transformation
#' tm <- vmgeom_vst_from_magn(m, limmag)
#' tm_mean <- mean(tm)
#' tm_var <- var(tm)
#'
#' # Estimator for r from the transformed mean
#' r_hat <- vmgeom_vst_to_r(tm_mean)
#'
#' # Derivative dr/d(tm) at tm_mean (needed for the delta method)
#' dr_dtm <- vmgeom_vst_to_r(tm_mean, deriv_degree = 1L)
#'
#' # Variance of the sample mean of tm
#' var_tm.mean <- tm_var / N
#'
#' # Delta method: variance and standard error of r_hat
#' var_r.hat <- (dr_dtm^2) * var_tm.mean
#' se_r.hat <- sqrt(var_r.hat)
#'
#' # Results
#' print(r_hat)
#' print(se_r.hat)

#' @rdname vmgeom_vst
#' @export
vmgeom_vst_from_magn <- function(m, lm) {
    offset <- lm - round(lm)
    if (1L == length(lm)) {
        offset <- rep(offset, length(m))
    }

    a <- .vmgeom_vst_from_magn_params$pa_fun(offset)
    b <- .vmgeom_vst_from_magn_params$pb_fun(offset)
    c <- .vmgeom_vst_from_magn_params$pc_fun(offset)
    d <- .vmgeom_vst_from_magn_params$pd_fun(offset)

    x <- lm - m
    a - exp(b + c * (x + 0.5)^d)
}

#' @rdname vmgeom_vst
#' @export
vmgeom_vst_to_r <- function(tm, log = FALSE, deriv_degree = 0L) {
    poly_coef0 <- c(10.31637299, -5.51610811, 1.38003791, -0.18014183, 0.00946975)
    names(poly_coef0) <- seq_along(poly_coef0) - 1 # exponents

    # tm min 3.96 (r approx 3.5)
    # tm max 5.74 (r approx 1.4)
    tm[tm < 3.96 | tm > 5.74] <- NA

    if (deriv_degree > 0L) {
        poly_coef1 <- .f_polynomial_coef(poly_coef0, deriv_degree = 1L)
    }
    if (deriv_degree > 1L) {
        poly_coef2 <- .f_polynomial_coef(poly_coef1, deriv_degree = 1L)
    }
    if (deriv_degree > 2L) {
        stop(paste("deriv_degree", deriv_degree, "not implemented!"))
    }

    if (log) {
        if (2L == deriv_degree) {
            .f_polynomial(tm, poly_coef2)
        } else if (1L == deriv_degree) {
            .f_polynomial(tm, poly_coef1)
        } else {
            .f_polynomial(tm, poly_coef0)
        }
    } else {
        if (2L == deriv_degree) {
            exp(.f_polynomial(tm, poly_coef0)) * (
                .f_polynomial(tm, poly_coef1)^2 + .f_polynomial(tm, poly_coef2)
            )
        } else if (1L == deriv_degree) {
            .f_polynomial(tm, poly_coef1) * exp(.f_polynomial(tm, poly_coef0))
        } else {
            exp(.f_polynomial(tm, poly_coef0))
        }
    }
}

#' @keywords internal
.vmgeom_vst_from_magn_params <- (function() {
    param_df <- data.frame(
        offset = c(
            -0.5, -0.48, -0.45, -0.4, -0.3, -0.2, -0.1, 0,
            0.1, 0.2, 0.3, 0.4, 0.45, 0.48, 0.5
        ),
        a = c(
            8.810993, 8.496702, 8.29026, 8.140408, 8.031893, 8.024524,
            8.046554, 8.093328, 8.165077, 8.265932, 8.384126, 8.564082,
            8.678373, 8.757368, 8.811488
        ),
        b = c(
            2.176, 2.077, 2.004, 1.945, 1.897, 1.885, 1.889, 1.906,
            1.933, 1.97, 2.02, 2.087, 2.128, 2.156, 2.176
        ),
        c = c(
            -0.317, -0.278, -0.249, -0.225, -0.206, -0.2, -0.201, -0.207,
            -0.217, -0.231, -0.252, -0.279, -0.296, -0.308, -0.317
        ),
        d = c(
            0.669, 0.748, 0.813, 0.872, 0.922, 0.934, 0.928, 0.91,
            0.881, 0.843, 0.796, 0.739, 0.706, 0.684, 0.669
        )
    )

    list(
        pa_fun = stats::approxfun(param_df$offset, param_df$a),
        pb_fun = stats::approxfun(param_df$offset, param_df$b),
        pc_fun = stats::approxfun(param_df$offset, param_df$c),
        pd_fun = stats::approxfun(param_df$offset, param_df$d)
    )
})()

test_that("vmgeom_vst", {
    r <- seq(1.5, 3.3, 0.1)
    limmag <- c(5.5, 5.52, 5.55, seq(5.6, 6.4, 0.1), 6.45, 6.48, 6.5)
    m <- seq(-200, 6, 1)
    model <- expand.grid(r = r, limmag = limmag)
    model <- do.call(
        rbind.data.frame,
        mapply(function(r, limmag) {
            p <- dvmgeom(m, limmag, r)
            t <- vmgeom_vst_from_magn(m, limmag)
            t.mean <- sum(p * t)
            t.var <- sum(p * (t - t.mean)^2)

            list(
                r = r,
                q = log(r),
                limmag = limmag,
                t.mean = t.mean,
                t.var = t.var
            )
        }, model$r, model$limmag, SIMPLIFY = FALSE)
    )

    #
    # test vmgeom_vst_from_magn
    #
    expect_true(vmgeom_vst_from_magn(6, 5.5) > 0.0)
    expect_true(all(abs(model$t.var - 1.0) < 0.018))

    #
    # test vmgeom_vst_to_r
    #
    model$r_est <- vmgeom_vst_to_r(model$t.mean)
    expect_true(all(abs(model$r - model$r_est) < 0.013))

    # test with non-exotic values
    model0 <- with(model, {
        subset(model, r < 2.7)
    })
    expect_true(all(abs(model0$r - model0$r_est) < 0.007))

    # test first derivative
    f <- function(x) {
        vmgeom_vst_to_r(x, deriv_degree = 1L)
    }
    y <- vmgeom_vst_to_r(5.5) - vmgeom_vst_to_r(4.5)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)

    # test second derivative
    f <- function(x) {
        vmgeom_vst_to_r(x, deriv_degree = 2L)
    }
    y <- vmgeom_vst_to_r(5.5, deriv_degree = 1L) - vmgeom_vst_to_r(4.5, deriv_degree = 1L)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)

    # log ...
    model$q.est <- vmgeom_vst_to_r(model$t.mean, log = TRUE)
    expect_true(all(abs(model$q - model$q.est) < 0.004))

    # test with non-exotic values
    model0 <- with(model, {
        subset(model, r < 2.7)
    })
    expect_true(all(abs(model0$q - model0$q.est) < 0.004))

    # test log first derivative
    f <- function(x) {
        vmgeom_vst_to_r(x, log = TRUE, deriv_degree = 1L)
    }
    y <- vmgeom_vst_to_r(5.5, log = TRUE) - vmgeom_vst_to_r(4.5, log = TRUE)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)

    # test log second derivative
    f <- function(x) {
        vmgeom_vst_to_r(x, log = TRUE, deriv_degree = 2L)
    }
    y <- vmgeom_vst_to_r(5.5, log = TRUE, deriv_degree = 1L) -
        vmgeom_vst_to_r(4.5, log = TRUE, deriv_degree = 1L)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)
})

#' @name load_vmdb
#' @aliases load_vmdb_rates
#' @aliases load_vmdb_magnitudes
#' @import httr2
#' @title Loading visual meteor observations via the imo-vmdb REST API
#' @description
#' Loads visual meteor observations from an
#' \href{https://pypi.org/project/imo-vmdb/}{imo-vmdb} web server via its
#' REST API.
#' @note Angle values are expected and returned in degrees.
#' @param base_url character; base URL of the imo-vmdb API, e.g.
#'   \code{"http://localhost:8000/api/v1"}.
#' @param shower character; selects by meteor shower codes.
#'   \code{NA} loads sporadic meteors.
#' @param period selects an observation time range by minimum/maximum.
#'   Accepts \code{POSIXct}, \code{Date}, or character.  Character
#'   elements may be \code{"YYYY-MM-DDTHH:MM:SS"},
#'   \code{"YYYY-MM-DD HH:MM:SS"}, or date-only \code{"YYYY-MM-DD"}.
#'   Date-only inputs are expanded to midnight (lower bound) and
#'   23:59:59 (upper bound) of the given day, in UTC.  IMO data is UTC
#'   by convention; the timezone marker is omitted on the wire.
#' @param sl numeric; selects a range of solar longitudes by minimum/maximum.
#' @param lim_magn numeric; selects a range of limiting magnitudes by
#'   minimum/maximum.
#' @param sun_alt_max numeric; selects the maximum altitude of the sun
#'   (rates only).
#' @param moon_alt_max numeric; selects the maximum altitude of the moon
#'   (rates only).
#' @param session_id integer; selects by session ids.
#' @param rate_id integer; selects rate observations by ids.
#' @param magn_id integer; selects magnitude observations by ids.
#' @param with_sessions logical; if \code{TRUE}, also load the corresponding
#'   session data.
#' @param with_magnitudes logical; if \code{TRUE}, also load the corresponding
#'   magnitude observations.
#' @details
#' \code{sl}, \code{period} and \code{lim_magn} expect a vector with
#' successive minimum and maximum values.
#' \code{sun_alt_max} and \code{moon_alt_max} are expected to be scalar values.
#'
#' \strong{Note:} Unlike the previous DBI-based version, only a single range
#' per filter parameter is supported.  If you previously passed a matrix with
#' multiple rows to \code{period}, \code{sl}, or \code{lim_magn}, flatten
#' them to a single min/max pair or issue multiple calls and combine with
#' \code{rbind()}.
#' @return
#' Both functions return a list, with
#'
#' \tabular{ll}{
#'     \code{observations} \tab data frame, rate or magnitude observations,\cr
#'     \code{sessions} \tab data frame; session data of observations,\cr
#'     \code{magnitudes} \tab table; contingency table of meteor magnitude
#'       frequencies.
#' }
#'
#' \code{observations} depends on the function call.
#' \code{load_vmdb_rates} returns a data frame with columns:
#'
#' \tabular{ll}{
#' \code{rate_id} \tab unique identifier of the rate observation,\cr
#' \code{shower} \tab IAU code of the shower. \code{NA} for sporadic.\cr
#' \code{period_start} \tab \code{POSIXct} (UTC); start of observation,\cr
#' \code{period_end} \tab \code{POSIXct} (UTC); end of observation,\cr
#' \code{sl_start} \tab solar longitude at start,\cr
#' \code{sl_end} \tab solar longitude at end,\cr
#' \code{session_id} \tab reference to the session,\cr
#' \code{freq} \tab count of observed meteors,\cr
#' \code{lim_magn} \tab limiting magnitude,\cr
#' \code{t_eff} \tab net observed time in hours,\cr
#' \code{f} \tab correction factor of cloud cover,\cr
#' \code{sidereal_time} \tab sidereal time,\cr
#' \code{sun_alt} \tab altitude of the sun,\cr
#' \code{sun_az} \tab azimuth of the sun,\cr
#' \code{moon_alt} \tab altitude of the moon,\cr
#' \code{moon_az} \tab azimuth of the moon,\cr
#' \code{moon_illum} \tab illumination of the moon (\code{0.0 .. 1.0}),\cr
#' \code{field_alt} \tab altitude of the field of view (optional),\cr
#' \code{field_az} \tab azimuth of the field of view (optional),\cr
#' \code{rad_alt} \tab altitude of the radiant (optional),\cr
#' \code{rad_az} \tab azimuth of the radiant (optional),\cr
#' \code{magn_id} \tab reference to the magnitude observations (optional),\cr
#' \code{magn_solo} \tab \code{TRUE} if this rate is the sole contributor
#'   to its linked magnitude observation; \code{FALSE} if the magnitude
#'   aggregates this rate with others; \code{NA} when \code{magn_id} is
#'   \code{NA} (optional).
#' }
#'
#' \code{load_vmdb_magnitudes} returns an \code{observations} data frame with:
#'
#' \tabular{ll}{
#' \code{magn_id} \tab unique identifier of the magnitude observation,\cr
#' \code{shower} \tab IAU code of the shower. \code{NA} for sporadic.\cr
#' \code{period_start} \tab \code{POSIXct} (UTC); start of observation,\cr
#' \code{period_end} \tab \code{POSIXct} (UTC); end of observation,\cr
#' \code{sl_start} \tab solar longitude at start,\cr
#' \code{sl_end} \tab solar longitude at end,\cr
#' \code{session_id} \tab reference to the session,\cr
#' \code{freq} \tab count of observed meteors,\cr
#' \code{magn_mean} \tab mean magnitude,\cr
#' \code{lim_magn} \tab limiting magnitude (optional).
#' }
#'
#' The \code{sessions} data frame contains
#'
#' \tabular{ll}{
#' \code{session_id} \tab unique identifier of the session,\cr
#' \code{longitude} \tab location's longitude,\cr
#' \code{latitude} \tab location's latitude,\cr
#' \code{elevation} \tab height above mean sea level in km,\cr
#' \code{country} \tab country name,\cr
#' \code{location_name} \tab location name,\cr
#' \code{observer_id} \tab observer id (optional),\cr
#' \code{observer_name} \tab observer name (optional).
#' }
#'
#' \code{magnitudes} is a contingency table of meteor magnitude frequencies.
#' Row names are magnitude observation IDs; column names are magnitude classes.
#'
#' @references \url{https://pypi.org/project/imo-vmdb/}
#' @examples
#' \dontrun{
#' # Load rate observations including session and magnitude data
#' data <- load_vmdb_rates(
#'     base_url       = "http://localhost:8000/api/v1",
#'     shower         = "PER",
#'     sl             = c(135.5, 145.5),
#'     period         = c("2015-08-01", "2015-08-31"),
#'     lim_magn       = c(5.3, 6.7),
#'     with_magnitudes = TRUE,
#'     with_sessions   = TRUE
#' )
#'
#' # Load magnitude observations
#' data <- load_vmdb_magnitudes(
#'     base_url     = "http://localhost:8000/api/v1",
#'     shower       = "PER",
#'     sl           = c(135.5, 145.5),
#'     period       = c("2015-08-01", "2015-08-31"),
#'     lim_magn     = c(5.3, 6.7),
#'     with_sessions = TRUE
#' )
#'
#' # Period can also be given as POSIXct or as full ISO 8601 datetime
#' # strings — useful for narrowing to a specific time window within a day.
#' data <- load_vmdb_rates(
#'     base_url = "http://localhost:8000/api/v1",
#'     shower   = "PER",
#'     period   = c("2015-08-12T20:00:00", "2015-08-13T04:00:00")
#' )
#' }

#' @rdname load_vmdb
#' @export
load_vmdb_rates <- function(
  base_url,
  shower = NULL,
  period = NULL,
  sl = NULL,
  lim_magn = NULL,
  sun_alt_max = NULL,
  moon_alt_max = NULL,
  session_id = NULL,
  rate_id = NULL,
  with_sessions = FALSE,
  with_magnitudes = FALSE
) {
    p <- .build_params(
        shower, period, sl, lim_magn,
        sun_alt_max = sun_alt_max,
        moon_alt_max = moon_alt_max,
        session_id = session_id,
        id_param = "rate_id",
        id_values = rate_id,
        with_sessions = with_sessions,
        with_magnitudes = with_magnitudes
    )

    body <- .api_get(base_url, "rates", p$scalar, p$multi)

    if (length(body$observations) == 0) {
        observations <- data.frame()
    } else {
        observations <- .remap_cols(as.data.frame(body$observations), .rate_col_map)
        observations$shower <- .normalize_shower(observations)
        observations$session_id <- factor(observations$session_id)
        observations$magn_id <- factor(observations$magn_id)
        observations$magn_solo <- as.logical(observations$magn_solo)
        observations$period_start <- .parse_dt(observations$period_start)
        observations$period_end <- .parse_dt(observations$period_end)
        row.names(observations) <- observations$rate_id
    }

    list(
        observations = observations,
        sessions     = .parse_sessions(body$sessions),
        magnitudes   = .parse_magnitudes(body$magnitude_details)
    )
}

#' @rdname load_vmdb
#' @export
load_vmdb_magnitudes <- function(
  base_url,
  shower = NULL,
  period = NULL,
  sl = NULL,
  lim_magn = NULL,
  session_id = NULL,
  magn_id = NULL,
  with_sessions = FALSE,
  with_magnitudes = TRUE
) {
    p <- .build_params(
        shower, period, sl, lim_magn,
        session_id = session_id,
        id_param = "magn_id",
        id_values = magn_id,
        with_sessions = with_sessions,
        with_magnitudes = with_magnitudes
    )

    body <- .api_get(base_url, "magnitudes", p$scalar, p$multi)

    if (length(body$observations) == 0) {
        observations <- data.frame()
    } else {
        observations <- .remap_cols(as.data.frame(body$observations), .magn_col_map)
        observations$shower <- .normalize_shower(observations)
        observations$session_id <- factor(observations$session_id)
        observations$period_start <- .parse_dt(observations$period_start)
        observations$period_end <- .parse_dt(observations$period_end)
        row.names(observations) <- observations$magn_id
    }

    list(
        observations = observations,
        sessions     = .parse_sessions(body$sessions),
        magnitudes   = .parse_magnitudes(body$magnitude_details)
    )
}

# Build a named list of scalar query parameters from filter arguments.
# Multi-value parameters (shower, session_id, rate_id, magn_id) are returned
# separately as a named list of vectors.
.build_params <- function(
  shower, period, sl, lim_magn,
  sun_alt_max = NULL, moon_alt_max = NULL,
  session_id = NULL, id_param = NULL, id_values = NULL,
  with_sessions = FALSE, with_magnitudes = FALSE
) {
    params <- list()
    multi <- list()

    if (!is.null(shower)) {
        multi$shower <- ifelse(is.na(shower), "SPO", shower)
    }

    if (!is.null(period)) {
        stopifnot(length(period) == 2)
        params$period_start <- .fmt_period(period[1], "lower")
        params$period_end <- .fmt_period(period[2], "upper")
    }

    if (!is.null(sl)) {
        stopifnot(length(sl) == 2)
        params$sl_min <- sl[1]
        params$sl_max <- sl[2]
    }

    if (!is.null(lim_magn)) {
        stopifnot(length(lim_magn) == 2)
        params$lim_magn_min <- lim_magn[1]
        params$lim_magn_max <- lim_magn[2]
    }

    if (!is.null(sun_alt_max)) params$sun_alt_max <- sun_alt_max
    if (!is.null(moon_alt_max)) params$moon_alt_max <- moon_alt_max

    if (!is.null(session_id)) multi$session_id <- as.integer(session_id)
    if (!is.null(id_values)) multi[[id_param]] <- as.integer(id_values)

    include <- character(0)
    if (with_sessions) include <- c(include, "sessions")
    if (with_magnitudes) include <- c(include, "magnitude_details")
    if (length(include) > 0) params$include <- paste(include, collapse = ",")

    list(scalar = params, multi = multi)
}


# Send a GET request to the API and return the parsed JSON body.
.api_get <- function(base_url, path, scalar_params, multi_params) {
    req <- httr2::request(base_url) |>
        httr2::req_url_path_append(path)

    if (length(scalar_params) > 0) {
        req <- do.call(httr2::req_url_query, c(list(req), scalar_params))
    }

    for (name in names(multi_params)) {
        vals <- multi_params[[name]]
        req <- do.call(
            httr2::req_url_query,
            c(list(req, .multi = "explode"), stats::setNames(list(vals), name))
        )
    }

    resp <- httr2::req_perform(req)
    httr2::resp_body_json(resp, simplifyVector = TRUE)
}


# Semantic renames only: disambiguate the generic `id` column against
# foreign-key columns in the same data frame, and avoid shadowing base R's
# `mean()`. Columns not listed here are passed through from the imo-vmdb
# API unchanged (it already uses snake_case).
.rate_col_map <- c(
    id = "rate_id"
)

.magn_col_map <- c(
    id   = "magn_id",
    mean = "magn_mean"
)

.session_col_map <- c(
    id = "session_id"
)

.remap_cols <- function(df, col_map) {
    new_names <- col_map[names(df)]
    names(df) <- ifelse(is.na(new_names), names(df), new_names)
    df
}


# Normalize the response-side shower column.  The imo-vmdb server omits
# `null`-valued fields entirely from JSON observations, so a fully
# sporadic page has no `shower` column at all (NULL after data.frame
# conversion), and a mixed page leaves sporadic rows as NA.  Both shapes
# need NA -> "SPO" mapped, matching the request-side encoding in
# `.build_params`.
.normalize_shower <- function(observations) {
    sh <- observations$shower
    if (is.null(sh)) {
        sh <- rep(NA_character_, nrow(observations))
    }
    sh <- ifelse(is.na(sh), "SPO", as.character(sh))
    factor(sh)
}


# Parse a strict ISO 8601 datetime string from the imo-vmdb REST API to
# POSIXct (UTC).  Matches the imo-vmdb 2.0 wire format
# (`YYYY-MM-DDTHH:MM:SS`, T separator, no timezone marker — IMO data is
# UTC by convention).  R's strptime treats the literal `T` in the format
# string.  Non-canonical input yields NA.
.parse_dt <- function(x) {
    as.POSIXct(x, tz = "UTC", format = "%Y-%m-%dT%H:%M:%S")
}


# Format a user-supplied period value to the strict ISO 8601 form
# imo-vmdb 2.0 expects (`YYYY-MM-DDTHH:MM:SS`, UTC, no timezone marker).
#
# Accepted shapes per element:
# - POSIXct                          -> formatted in UTC.
# - Date                             -> midnight (lower) / 23:59:59 (upper) UTC.
# - character "YYYY-MM-DDTHH:MM:SS"  -> reformatted (identity in practice).
# - character "YYYY-MM-DD HH:MM:SS"  -> parsed as datetime, reformatted with T.
# - character "YYYY-MM-DD"           -> midnight (lower) / 23:59:59 (upper) UTC.
# - anything else / parse failure    -> stop() with a clear message.
.fmt_period <- function(x, bound = c("lower", "upper")) {
    bound <- match.arg(bound)
    pt <- if (inherits(x, "POSIXct")) {
        as.POSIXct(format(x, tz = "UTC"), tz = "UTC")
    } else if (inherits(x, "Date")) {
        as.POSIXct(as.character(x), tz = "UTC") +
            (if (bound == "upper") 86399 else 0)
    } else if (is.character(x)) {
        v <- as.POSIXct(x, tz = "UTC", format = "%Y-%m-%dT%H:%M:%S")
        if (is.na(v)) {
            v <- as.POSIXct(x, tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
        }
        if (is.na(v)) {
            v <- as.POSIXct(x, tz = "UTC", format = "%Y-%m-%d")
            if (!is.na(v) && bound == "upper") {
                v <- v + 86399
            }
        }
        v
    } else {
        stop(
            "period values must be POSIXct, Date, or character; got ",
            class(x)[1]
        )
    }
    if (is.na(pt)) {
        stop("Cannot parse period value: ", x)
    }
    format(pt, format = "%Y-%m-%dT%H:%M:%S", tz = "UTC")
}


# Parse the sessions array from the API response into a data.frame.
.parse_sessions <- function(sessions_list) {
    if (is.null(sessions_list) || length(sessions_list) == 0) {
        return(NULL)
    }
    s <- .remap_cols(as.data.frame(sessions_list), .session_col_map)
    s$country <- factor(s$country)
    s$location_name <- factor(s$location_name)
    s$observer_id <- factor(s$observer_id)
    s$observer_name <- factor(s$observer_name)
    row.names(s) <- s$session_id
    s
}


# Parse the magnitudes array from the API response into an xtabs table.
.parse_magnitudes <- function(magnitudes_list) {
    if (is.null(magnitudes_list) || length(magnitudes_list) == 0) {
        return(NULL)
    }
    m <- as.data.frame(magnitudes_list)
    if (nrow(m) == 0) {
        return(NULL)
    }
    m$magn <- factor(
        m$magn,
        levels  = sort(unique(m$magn), decreasing = TRUE),
        ordered = TRUE
    )
    names(m)[names(m) == "id"] <- "magn_id"
    stats::xtabs(freq ~ magn_id + magn, data = m)
}

test_that(".build_params: shower encoding", {
    p <- vismeteor:::.build_params("PER", NULL, NULL, NULL)
    expect_equal(p$multi$shower, "PER")

    # NA → SPO (sporadic)
    p <- vismeteor:::.build_params(NA, NULL, NULL, NULL)
    expect_equal(p$multi$shower, "SPO")

    # mixed: NA becomes SPO, named entries kept
    p <- vismeteor:::.build_params(c("PER", NA), NULL, NULL, NULL)
    expect_equal(p$multi$shower, c("PER", "SPO"))

    # NULL → no shower param
    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL)
    expect_null(p$multi$shower)
})

test_that(".build_params: range params", {
    # Date-only character input is expanded to strict ISO-T form:
    # lower bound -> midnight, upper bound -> 23:59:59 UTC.
    p <- vismeteor:::.build_params(NULL, c("2015-08-01", "2015-08-31"), NULL, NULL)
    expect_equal(p$scalar$period_start, "2015-08-01T00:00:00")
    expect_equal(p$scalar$period_end, "2015-08-31T23:59:59")

    p <- vismeteor:::.build_params(NULL, NULL, c(135.5, 145.5), NULL)
    expect_equal(p$scalar$sl_min, 135.5)
    expect_equal(p$scalar$sl_max, 145.5)

    p <- vismeteor:::.build_params(NULL, NULL, NULL, c(5.5, 6.5))
    expect_equal(p$scalar$lim_magn_min, 5.5)
    expect_equal(p$scalar$lim_magn_max, 6.5)
})

test_that(".parse_dt: strict ISO-T parser", {
    v <- vismeteor:::.parse_dt("2015-08-12T22:00:00")
    expect_s3_class(v, "POSIXct")
    expect_equal(attr(v, "tzone"), "UTC")
    expect_equal(
        as.numeric(v),
        as.numeric(as.POSIXct("2015-08-12 22:00:00", tz = "UTC"))
    )

    # Non-canonical forms are NOT tolerated.
    expect_true(is.na(vismeteor:::.parse_dt("2015-08-12 22:00:00")))
    expect_true(is.na(vismeteor:::.parse_dt("2015-08-12")))
    expect_true(is.na(vismeteor:::.parse_dt("not a date")))
})

test_that(".fmt_period: strict ISO-T formatter", {
    # POSIXct in -> T-form out, UTC.
    pt <- as.POSIXct("2015-08-12 22:00:00", tz = "UTC")
    expect_equal(vismeteor:::.fmt_period(pt, "lower"), "2015-08-12T22:00:00")
    expect_equal(vismeteor:::.fmt_period(pt, "upper"), "2015-08-12T22:00:00")

    # Date -> midnight (lower) / 23:59:59 (upper).
    d <- as.Date("2015-08-12")
    expect_equal(vismeteor:::.fmt_period(d, "lower"), "2015-08-12T00:00:00")
    expect_equal(vismeteor:::.fmt_period(d, "upper"), "2015-08-12T23:59:59")

    # character T-form -> identity.
    expect_equal(
        vismeteor:::.fmt_period("2015-08-12T22:00:00", "lower"),
        "2015-08-12T22:00:00"
    )

    # character space-form -> reformatted with T.
    expect_equal(
        vismeteor:::.fmt_period("2015-08-12 22:00:00", "lower"),
        "2015-08-12T22:00:00"
    )

    # character date-only -> bound expansion.
    expect_equal(
        vismeteor:::.fmt_period("2015-08-12", "lower"),
        "2015-08-12T00:00:00"
    )
    expect_equal(
        vismeteor:::.fmt_period("2015-08-12", "upper"),
        "2015-08-12T23:59:59"
    )

    # Unparseable -> error.
    expect_error(vismeteor:::.fmt_period("not a date", "lower"))
    expect_error(vismeteor:::.fmt_period(list(), "lower"))
})

test_that(".build_params: scalar altitude / id params", {
    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL, sun_alt_max = -10)
    expect_equal(p$scalar$sun_alt_max, -10)

    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL, moon_alt_max = 5)
    expect_equal(p$scalar$moon_alt_max, 5)

    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL, session_id = c(1L, 2L))
    expect_equal(p$multi$session_id, c(1L, 2L))

    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL,
        id_param = "rate_id", id_values = 100L
    )
    expect_equal(p$multi$rate_id, 100L)
})

test_that(".build_params: include parameter", {
    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL)
    expect_null(p$scalar$include)

    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL, with_sessions = TRUE)
    expect_equal(p$scalar$include, "sessions")

    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL, with_magnitudes = TRUE)
    expect_equal(p$scalar$include, "magnitude_details")

    p <- vismeteor:::.build_params(NULL, NULL, NULL, NULL,
        with_sessions = TRUE, with_magnitudes = TRUE
    )
    expect_equal(p$scalar$include, "sessions,magnitude_details")
})

test_that(".parse_sessions: correct data.frame with factors and row names", {
    df <- data.frame(
        id = 1L,
        longitude = 10.0,
        latitude = 50.0,
        elevation = 0.3,
        country = "DE",
        location_name = "Somewhere",
        observer_id = "XX",
        observer_name = "Doe, J.",
        stringsAsFactors = FALSE
    )
    s <- vismeteor:::.parse_sessions(df)
    expect_true(is.data.frame(s))
    expect_equal(names(s), c(
        "session_id", "longitude", "latitude", "elevation",
        "country", "location_name", "observer_id", "observer_name"
    ))
    expect_equal(s$session_id, 1L)
    expect_true(is.factor(s$country))
    expect_true(is.factor(s$location_name))
    expect_true(is.factor(s$observer_id))
    expect_true(is.factor(s$observer_name))
    expect_equal(row.names(s), "1")

    expect_null(vismeteor:::.parse_sessions(NULL))
    expect_null(vismeteor:::.parse_sessions(list()))
})

test_that(".parse_magnitudes: builds xtabs table with correct dims", {
    df <- data.frame(
        id = c(200L, 200L, 200L, 201L, 201L),
        magn = c(3L, 2L, 1L, 3L, 2L),
        freq = c(2.5, 4.0, 5.5, 3.0, 5.0),
        stringsAsFactors = FALSE
    )
    m <- vismeteor:::.parse_magnitudes(df)
    expect_true(methods::is(m, "table"))
    expect_setequal(row.names(m), c("200", "201"))
    expect_setequal(colnames(m), c("1", "2", "3"))
    expect_equal(m["200", "2"], 4.0)
    expect_equal(m["201", "3"], 3.0)

    expect_null(vismeteor:::.parse_magnitudes(NULL))
    expect_null(vismeteor:::.parse_magnitudes(as.data.frame(list())))
})

# Integration tests: full pipeline with mocked HTTP responses.
# Fixture files live under tests/testthat/fixtures/{scenario}/
# and map to http://example.com/api/v1/{endpoint}.json

test_that("load_vmdb_rates: empty observations", {
    testthat::skip_if_not_installed("httptest2")
    httptest2::with_mock_dir("fixtures/empty", {
        res <- load_vmdb_rates("http://example.com/api/v1")
        expect_type(res, "list")
        expect_true(is.data.frame(res$observations))
        expect_equal(nrow(res$observations), 0)
        expect_null(res$sessions)
        expect_null(res$magnitudes)
    })
})

test_that("load_vmdb_rates: parses observations, sessions, magnitudes", {
    testthat::skip_if_not_installed("httptest2")
    httptest2::with_mock_dir("fixtures/full", {
        res <- load_vmdb_rates("http://example.com/api/v1")
        expect_type(res, "list")

        obs <- res$observations
        expect_true(is.data.frame(obs))
        expect_equal(nrow(obs), 1)
        expect_equal(obs$rate_id, 100L)
        expect_true(is.factor(obs$shower))
        expect_equal(as.character(obs$shower), "PER")
        expect_true(is.factor(obs$session_id))
        expect_true(is.factor(obs$magn_id))
        expect_true(is.logical(obs$magn_solo))
        expect_true(obs$magn_solo)
        expect_s3_class(obs$period_start, "POSIXct")
        expect_s3_class(obs$period_end, "POSIXct")
        expect_equal(attr(obs$period_start, "tzone"), "UTC")
        expect_equal(attr(obs$period_end, "tzone"), "UTC")
        expect_equal(row.names(obs), "100")

        sess <- res$sessions
        expect_true(is.data.frame(sess))
        expect_equal(row.names(sess), "1")
        expect_true(is.factor(sess$country))

        magn <- res$magnitudes
        expect_true(methods::is(magn, "table"))
        expect_equal(row.names(magn), "200")
        expect_setequal(colnames(magn), c("1", "2", "3"))
    })
})

test_that("load_vmdb_rates: all-sporadic response maps to SPO factor", {
    testthat::skip_if_not_installed("httptest2")
    httptest2::with_mock_dir("fixtures/sporadic_only", {
        res <- load_vmdb_rates("http://example.com/api/v1")
        obs <- res$observations
        expect_equal(nrow(obs), 2)
        expect_true(is.factor(obs$shower))
        expect_true("SPO" %in% levels(obs$shower))
        expect_equal(as.character(obs$shower), c("SPO", "SPO"))
        expect_false(any(is.na(obs$shower)))
    })
})

test_that("load_vmdb_rates: mixed PER/sporadic response gets SPO mapping", {
    testthat::skip_if_not_installed("httptest2")
    httptest2::with_mock_dir("fixtures/sporadic_mixed", {
        res <- load_vmdb_rates("http://example.com/api/v1")
        obs <- res$observations
        expect_equal(nrow(obs), 2)
        expect_true(is.factor(obs$shower))
        expect_setequal(as.character(obs$shower), c("PER", "SPO"))
        expect_false(any(is.na(obs$shower)))
    })
})

test_that("load_vmdb_magnitudes: all-sporadic response maps to SPO factor", {
    testthat::skip_if_not_installed("httptest2")
    httptest2::with_mock_dir("fixtures/sporadic_only_magn", {
        res <- load_vmdb_magnitudes("http://example.com/api/v1",
                                    with_magnitudes = FALSE)
        obs <- res$observations
        expect_equal(nrow(obs), 2)
        expect_true(is.factor(obs$shower))
        expect_equal(as.character(obs$shower), c("SPO", "SPO"))
    })
})

test_that("load_vmdb_magnitudes: mixed PER/sporadic response gets SPO mapping", {
    testthat::skip_if_not_installed("httptest2")
    httptest2::with_mock_dir("fixtures/sporadic_mixed_magn", {
        res <- load_vmdb_magnitudes("http://example.com/api/v1",
                                    with_magnitudes = FALSE)
        obs <- res$observations
        expect_equal(nrow(obs), 2)
        expect_setequal(as.character(obs$shower), c("PER", "SPO"))
    })
})

test_that(".build_params: POSIXct period preserves class", {
    pt <- as.POSIXct(c("2015-08-12 00:00:00", "2015-08-13 23:59:59"),
                     tz = "UTC")
    p <- vismeteor:::.build_params(NULL, pt, NULL, NULL)
    expect_equal(p$scalar$period_start, "2015-08-12T00:00:00")
    expect_equal(p$scalar$period_end, "2015-08-13T23:59:59")
})

test_that("load_vmdb_magnitudes: parses observations, sessions, magnitudes", {
    testthat::skip_if_not_installed("httptest2")
    # with_magnitudes defaults to TRUE → suppress include param for clean fixture path
    httptest2::with_mock_dir("fixtures/full", {
        res <- load_vmdb_magnitudes("http://example.com/api/v1", with_magnitudes = FALSE)
        expect_type(res, "list")

        obs <- res$observations
        expect_true(is.data.frame(obs))
        expect_equal(nrow(obs), 1)
        expect_equal(obs$magn_id, 200L)
        expect_true(is.factor(obs$shower))
        expect_equal(as.character(obs$shower), "PER")
        expect_s3_class(obs$period_start, "POSIXct")
        expect_s3_class(obs$period_end, "POSIXct")
        expect_equal(attr(obs$period_start, "tzone"), "UTC")
        expect_equal(attr(obs$period_end, "tzone"), "UTC")
        expect_equal(row.names(obs), "200")

        expect_true(is.data.frame(res$sessions))
        expect_true(methods::is(res$magnitudes, "table"))
    })
})

test_that("per_2015", {
    if (FALSE) {
        base_url <- "http://localhost:8000/api/v1"
        data_dir <- system.file("data", package = "vismeteor")

        PER_2015_rates <- load_vmdb_rates(
            base_url,
            shower = "PER",
            period = c("2015-01-01", "2015-12-31"),
            with_magnitudes = TRUE,
            with_sessions = TRUE
        )
        save(
            PER_2015_rates,
            file = normalizePath(paste0(data_dir, "/PER_2015_rates.RData")),
            compress = "xz"
        )

        PER_2015_magn <- load_vmdb_magnitudes(
            base_url,
            shower = "PER",
            period = c("2015-01-01", "2015-12-31"),
            with_magnitudes = TRUE,
            with_sessions = TRUE
        )
        save(
            PER_2015_magn,
            file = normalizePath(paste0(data_dir, "/PER_2015_magn.RData")),
            compress = "xz"
        )
    }

    expect_type(PER_2015_rates, "list")
    sessions <- PER_2015_rates$sessions
    expect_type(sessions, "list")
    expect_true(is(sessions, "data.frame"))
    observations <- PER_2015_rates$observations
    expect_type(observations, "list")
    expect_true(is(observations, "data.frame"))
    magnitudes <- PER_2015_rates$magnitudes
    expect_type(magnitudes, "double")
    expect_true(is(magnitudes, "table"))

    expect_type(PER_2015_magn, "list")
    sessions <- PER_2015_magn$sessions
    expect_type(sessions, "list")
    expect_true(is(sessions, "data.frame"))
    observations <- PER_2015_magn$observations
    expect_type(observations, "list")
    expect_true(is(observations, "data.frame"))
    magnitudes <- PER_2015_magn$magnitudes
    expect_type(magnitudes, "double")
    expect_true(is(magnitudes, "table"))
})

#' @title Quantiles with a minimum frequency
#' @description
#' This function generates quantiles with a minimum frequency.
#' These quantiles are formed from a vector `freq` of frequencies.
#' Each quantile then has the minimum total frequency `min`.
#' @param freq integer; A vector of frequencies.
#' @param min integer; Minimum total frequency per quantile.
#' @details
#' The frequencies `freq` are grouped in the order in which they
#' are passed as a vector.
#' The minimum `min` must be greater than `0`.
#' @return
#' A factor of indices is returned.
#' The index references the corresponding passed frequency `freq`.
#' @examples
#' freq <- c(1, 2, 3, 4, 5, 6, 7, 8, 9)
#' cumsum(freq)
#' (f <- freq_quantile(freq, 10))
#' tapply(freq, f, sum)
#' @export
freq_quantile <- function(freq, min) {
    if (0 >= min) {
        stop(paste0('min must be greater than 0 instead of "', min, '"!'))
    }

    n_sum <- 0
    id_last <- 1L
    id <- integer(length(freq))
    for (i in seq_along(freq)) {
        n_i <- freq[i]
        if ((n_i + n_sum) < min) {
            id[i] <- id_last
            n_sum <- n_i + n_sum
        } else {
            id[i] <- id_last
            n_sum <- 0
            id_last <- id_last + 1L
        }
    }

    if (0 == n_sum) {
        id_last <- id_last - 1L
    }

    id_max <- id[length(id)]
    if (id_max > 1 && n_sum > 0 && sum(freq[id == id_last]) < min) {
        id[id == id_max] <- id_max - 1L
    }

    factor(id, ordered = TRUE)
}

#' @title Rounds a contingency table of meteor magnitude frequencies
#' @description
#' The meteor magnitude contingency table of VMDB contains half meteor counts (e.g. `3.5`).
#' This function converts these frequencies to integer values.
#' @param mt table; A two-dimensional contingency table of meteor magnitude frequencies.
#' @details
#' The contingency table of meteor magnitudes `mt` must be two-dimensional.
#' The row names refer to the magnitude observations.
#' Column names must be integer meteor magnitude values.
#' Also, the columns must be sorted in ascending or descending order of meteor magnitude.
#'
#' A sum-preserving algorithm is used for rounding.
#' It ensures that the total frequency of meteors per observation is preserved.
#' The marginal frequencies of the magnitudes are also preserved with
#' the restriction that the deviation is at most \eqn{\pm 0.5}.
#' If the total sum of a meteor magnitude is integer,
#' then the deviation is \eqn{\pm 0}.
#'
#' The algorithm is unbiased: for a fixed observation order it preserves the original
#' totals without introducing systematic drift, even though each run follows the
#' deterministic sequence dictated by the observed counts and their ordering.
#'
#' @note Internally the counts are doubled to half-meteor units, leftover halves are
#' alternated between rows so column margins stay within \eqn{\pm 0.5}, and when the
#' grand total is odd the matrix is temporarily mirrored so the unavoidable surplus
#' meteor originates from the opposite end of the magnitude scale rather than always
#' favouring the faintest bin. The mirroring is only the initial condition; the loop
#' then processes the table cell by cell so the rounding direction alternates between
#' bright and faint magnitudes depending on the current row and column state.
#'
#' @return
#' A rounded contingency table of meteor magnitudes is returned.
#' @examples
#' # For example, create a contingency table of meteor magnitudes
#' mt <- as.table(matrix(
#'     c(
#'         0.0, 0.0, 2.5, 0.5, 0.0, 1.0,
#'         0.0, 1.5, 2.0, 0.5, 0.0, 0.0,
#'         1.0, 0.0, 0.0, 3.0, 2.5, 0.5
#'     ),
#'     nrow = 3, ncol = 6, byrow = TRUE
#' ))
#' colnames(mt) <- seq(6)
#' rownames(mt) <- c("A", "B", "C")
#' mt
#' margin.table(mt, 1)
#' margin.table(mt, 2)
#'
#' # contingency table with integer values
#' (mt_int <- vmtable(mt))
#' margin.table(mt_int, 1)
#' margin.table(mt_int, 2)
#' @export
vmtable <- function(mt) {
    if (!methods::is(mt, "table")) {
        stop(paste0("Magnitude table is not a table!"))
    }

    if (2L != length(dim(mt))) {
        stop(paste0("Magnitude table is not two-dimensional!"))
    }

    mt_m <- as.matrix(mt) # work with raw numeric matrix to control rounding precisely
    ncol_mt <- ncol(mt_m)
    # "Fair" rounding. When the total count is odd we mirror the columns so that the surplus
    # meteor is taken from the opposite end of the magnitude scale instead of consistently
    # favouring the faintest magnitudes during the rounding pass
    from_right <- 0L != sum(mt_m) %% 2L
    if (from_right) {
        mt_m <- t(apply(mt_m, 1L, rev))
    }
    mt2c <- round(2L * mt_m) # work in half-meteor units so integer rounding keeps the original totals

    # phase 1: round column margin and add dummy row
    margin_v <- as.integer(as.vector(margin.table(mt2c, 2L)))
    # dummy row absorbs the 0.5 remainder per column so every column margin stays intact after rounding
    dummy_row <- diff(c(0L, sapply(cumsum(margin_v), \(freq) {
        2L * (freq %/% 2L) # round down
    }))) - margin_v
    mt2c <- rbind(mt2c, dummy_row) # add dummy row
    nrow_mt2c <- nrow(mt2c)

    # phase 2: cumsum column-wise and round
    mt2c_v <- as.integer(as.vector(mt2c)) # traverse cells column-by-column to match VMDB layout
    mt2c_v_cs <- cumsum(mt2c_v) # cumulative sum acts as running total for the sum-preserving rounding logic
    # retain row membership so remainders alternate per observation
    mt2c_row_ids <- rep(seq_len(nrow_mt2c), times = ncol_mt)

    rows_reminder <- rep(0L, nrow_mt2c) # tracks whether a row already kept the previous 0.5 remainder
    mt2c_f <- integer(length(mt2c_v_cs))
    # walk through the cumulative frequencies and distribute every odd remainder to alternating rows
    for (i in seq_along(mt2c_v_cs)) {
        row_id <- mt2c_row_ids[i]
        freq <- mt2c_v_cs[i] # running total for this cell including all previous entries of the column
        freq_quotient <- freq %/% 2L # largest whole-meteor multiple we can commit without breaking the total
        freq_reminder <- freq %% 2L # 1 when the cumulative sum would otherwise leave a dangling half meteor

        freq <- 2L * freq_quotient
        if (0L != freq_reminder) {
            row_reminder_org <- rows_reminder[row_id]
            row_reminder_new <- 1L
            if (1L == row_reminder_org) {
                freq <- freq + 2L # give the extra meteor to the opposite row on alternating encounters
                row_reminder_new <- 0L
            }

            if (row_reminder_org != row_reminder_new) {
                rows_reminder[row_id] <- row_reminder_new # remember which rows have already claimed the half meteor
            }
        }
        mt2c_f[i] <- freq # store the adjusted cumulative total for later differencing
    }

    mt2c_v <- diff(append(mt2c_f, 0L, after = 0L)) # undo cumulative form to recover per-cell counts
    mt2c_m <- matrix(mt2c_v, ncol = ncol_mt) # restore table shape after column-wise walk
    mt2c_m <- utils::head(mt2c_m, -1L) # remove dummy_row
    if (from_right) {
        mt2c_m <- t(apply(mt2c_m, 1L, rev)) # reapply original magnitude ordering if we flipped earlier
    }

    result <- as.table(mt2c_m %/% 2L) # collapse half-meteor counts back to whole meteors
    dimnames(result) <- dimnames(mt) # restore dimnames

    result
}

test_that("freq_quantile", {
    # simple example I
    freq <- c(1, 2, 3, 4, 5, 6, 7, 8, 9)
    f <- vismeteor::freq_quantile(freq, 10)
    expect_type(f, "integer")
    expect_true(is.factor(f))
    expect_equal(as.integer(f), c(1, 1, 1, 1, 2, 2, 3, 3, 3))
    a <- sapply(split(freq, f), sum)
    expect_true(all(a >= 10))

    # simple example II
    freq <- c(9, 8, 7, 6, 5, 4, 3, 2, 1)
    f <- vismeteor::freq_quantile(freq, 10)
    expect_type(f, "integer")
    expect_true(is.factor(f))
    expect_equal(as.integer(f), c(1, 1, 2, 2, 3, 3, 3, 3, 3))
    a <- sapply(split(freq, f), sum)
    expect_true(all(a >= 10))

    # simple example III
    freq <- rep(10, 6)
    f <- vismeteor::freq_quantile(freq, 10)
    expect_type(f, "integer")
    expect_true(is.factor(f))
    expect_equal(as.integer(f), c(1, 2, 3, 4, 5, 6))
    a <- sapply(split(freq, f), sum)
    expect_true(all(a >= 10))

    # simple example IV
    freq <- rep(5, 12)
    f <- vismeteor::freq_quantile(freq, 10)
    expect_type(f, "integer")
    expect_true(is.factor(f))
    expect_equal(as.integer(f), rep(c(1, 2, 3, 4, 5, 6), each = 2))
    a <- sapply(split(freq, f), sum)
    expect_true(all(a >= 10))

    # test overflow
    freq <- c(1, 10)
    f <- vismeteor::freq_quantile(freq, 10)
    expect_type(f, "integer")
    expect_true(is.factor(f))
    expect_equal(as.integer(f), c(1, 1))
    a <- sapply(split(freq, f), sum)
    expect_true(all(a >= 10))

    # test single large frequency
    freq <- 12
    f <- vismeteor::freq_quantile(freq, 10)
    expect_type(f, "integer")
    expect_true(is.factor(f))
    expect_equal(as.integer(f), 1)

    # test single small frequency
    freq <- 2
    f <- vismeteor::freq_quantile(freq, 10)
    expect_type(f, "integer")
    expect_true(is.factor(f))
    expect_equal(as.integer(f), 1)
})

test_that("vmtable", {
    test_mt <- function(mt) {
        mt_int <- vismeteor::vmtable(mt)
        expect_type(mt_int, "integer")
        expect_true(isa(mt_int, "table"))
        expect_equal(
            as.vector(margin.table(mt, 1L)),
            as.vector(margin.table(mt_int, 1L))
        )

        diff <- as.vector(margin.table(mt, 2L)) -
            as.vector(margin.table(mt_int, 2L))
        expect_true(any(abs(diff) <= 0.5))
        expect_true(0.0 == sum(mt - mt_int))
        expect_false(any(abs(mt - mt_int) > 1.0))

        mt_int
    }

    # null observation
    mt <- as.table(matrix(
        c(
            0.0, 0.0, 0.0, 0.0
        ),
        nrow = 1, ncol = 4, byrow = TRUE
    ))
    mt_int <- test_mt(mt)
    expected <- as.table(matrix(
        c(
            0L, 0L, 0L, 0L
        ),
        nrow = 1, ncol = 4, byrow = TRUE
    ))
    expect_equal(expected, mt_int)

    # single observations
    mt <- as.table(matrix(
        c(
            1.0, 0.5, 0.5, 0.0
        ),
        nrow = 1, ncol = 4, byrow = TRUE
    ))
    mt_int <- test_mt(mt)
    expected <- as.table(matrix(
        c(
            1L, 0L, 1L, 0L
        ),
        nrow = 1, ncol = 4, byrow = TRUE
    ))
    expect_equal(expected, mt_int)

    mt <- as.table(matrix(
        c(
            1.5, 0.5, 0.5, 0.5
        ),
        nrow = 1, ncol = 4, byrow = TRUE
    ))
    mt_int <- test_mt(mt)
    expected <- as.table(matrix(
        c(
            2L, 0L, 1L, 0L
        ),
        nrow = 1, ncol = 4, byrow = TRUE
    ))
    expect_equal(expected, mt_int)

    # multi observation I
    mt <- as.table(matrix(
        c(
            0.0, 0.0, 0.0, 1.0,
            0.0, 0.0, 0.5, 0.5,
            0.0, 0.5, 0.5, 0.0,
            0.0, 0.0, 0.0, 0.0,
            0.5, 0.5, 0.0, 0.0,
            1.0, 0.0, 0.0, 0.0
        ),
        nrow = 6, ncol = 4, byrow = TRUE
    ))
    mt_int <- test_mt(mt)
    expected <- as.table(matrix(
        c(
            0L, 0L, 0L, 1L,
            0L, 0L, 1L, 0L,
            0L, 1L, 0L, 0L,
            0L, 0L, 0L, 0L,
            1L, 0L, 0L, 0L,
            1L, 0L, 0L, 0L
        ),
        nrow = 6, ncol = 4, byrow = TRUE
    ))
    expect_equal(expected, mt_int)

    # multi observation II
    mt <- as.table(matrix(
        c(
            0.0, 0.0, 0.0, 1.0,
            0.5, 0.5, 0.5, 0.5,
            0.0, 0.5, 1.0, 0.5,
            0.0, 0.0, 0.0, 0.0,
            0.5, 1.0, 0.5, 0.0,
            1.0, 2.0, 3.0, 4.0
        ),
        nrow = 6, ncol = 4, byrow = TRUE
    ))
    mt_int <- test_mt(mt)
    expected <- as.table(matrix(
        c(
            0L, 0L, 0L, 1L,
            1L, 0L, 1L, 0L,
            0L, 1L, 0L, 1L,
            0L, 0L, 0L, 0L,
            0L, 1L, 1L, 0L,
            1L, 2L, 3L, 4L
        ),
        nrow = 6, ncol = 4, byrow = TRUE
    ))
    expect_equal(expected, mt_int)

    # Perseids margin check
    mt <- PER_2015_magn$magnitudes
    mt_margin <- as.data.frame(margin.table(mt, 1))
    colnames(mt_margin)[2] <- "freq0"
    mt_int <- test_mt(mt)

    parts <- 20L
    # split by row
    f <- floor(parts * (seq(1L, nrow(mt)) - 1L) / nrow(mt))
    for (i in (seq_len(parts) - 1L)) {
        mti <- mt[i == f, ]
        test_mt(mti)
        mti_int <- mt_int[i == f, ]
        diff <- as.vector(margin.table(mti, 2L)) - as.vector(margin.table(mti_int, 2L))
        expect_true(any(abs(diff) <= 0.5))
        diff <- sum(mti - mti_int)
        expect_true(0.0 == diff)
    }

    if (FALSE) {
        # create meteors from gamma distribution
        obs_n <- 200L
        m_n <- 1000L
        set.seed(13)
        m_rnd <- rgamma(m_n, 3.0)
        m_gamma_mean <- mean(m_rnd)
        m_gamma_var <- var(m_rnd)
        expect_true(abs(m_gamma_mean - 3.0) < 0.03)
        expect_true(abs(m_gamma_var - 3.0) < 0.1)
        m_rnd <- round(2.0 * m_rnd) / 2.0 # round 0.5
        m_obs <- seq(1, obs_n)
        names(m_rnd) <- rep(m_obs, each = m_n %/% obs_n)
        m_df <- do.call(
            rbind.data.frame,
            tapply(m_rnd, names(m_rnd), function(m) {
                obs <- as.integer(names(m)[1])
                m <- c(floor(m), floor(m + 0.5))
                t <- table(m) / 2.0
                df <- as.data.frame(t)
                df$m <- as.integer(levels(df$m))
                df$obs <- rep(obs, nrow(df))
                df
            })
        )
        m_df$m <- factor(as.character(m_df$m), levels = as.character(sort(unique(m_df$m))), ordered = TRUE)
        mt <- xtabs(Freq ~ obs + m, data = m_df)
        mt_r <- margin.table(mt, 1)
        mt_c <- margin.table(mt, 2)
        mt_mean <- sum(as.integer(names(mt_c)) * mt_c) / m_n
        mt_var <- sum((as.integer(names(mt_c)) - mt_mean)^2 * mt_c) / (m_n - 1)
        expect_true(abs(mt_mean - m_gamma_mean) < 0.01)
        expect_true(abs(mt_var - m_gamma_var) < 0.2)

        mt_int <- test_mt(mt)
        mt_int_c <- margin.table(mt_int, 2)
        mt_int_mean <- sum(as.integer(names(mt_int_c)) * mt_int_c) / m_n
        mt_int_var <- sum((as.integer(names(mt_int_c)) - mt_mean)^2 * mt_int_c) / (m_n - 1)
        expect_true(abs(mt_int_mean - m_gamma_mean) < 0.01)
        expect_true(abs(mt_int_var - m_gamma_var) < 0.2)
        expect_true(abs(mt_int_mean - mt_mean) < 0.01)
        expect_true(abs(mt_int_var - mt_var) < 0.03)
    }
})

#' @title Perception Probabilities of Visual Meteor Magnitudes
#' @description
#' Provides the perception probability of visual meteor magnitudes.
#' @param dm numeric; difference between the limiting magnitude and the meteor magnitude.
#' @details
#' The perception probabilities of _Koschack R., Rendtel J., 1990b_
#' are estimated with the formula
#' \deqn{
#'     p(dm) = \begin{cases}
#'         1.0 - \exp\left(-z(dm + 0.5)\right)\  & \text{ if } dm > -0.5,\\
#'         0.0 \  & \text{ otherwise,}
#'     \end{cases}
#' }
#' where
#' \deqn{
#' z(x) = 0.0037 \, x + 0.0019 \, x^2 + 0.00271 \, x^3 + 0.0009 \, x^4
#' }
#' and `dm` is the difference between the limiting magnitude and the meteor magnitude.
#' @return This function returns the visual perception probabilities.
#' @references
#' Koschack R., Rendtel J., 1990b
#' _Determination of spatial number density and mass index from visual meteor observations (II)._
#' WGN 18, 119–140.
#' @examples
#' # Perception probability of visually estimated meteor of magnitude 3.0
#' # with a limiting magnitude of 5.6.
#' vmperception(5.6 - 3.0)
#'
#' # plot
#' old_par <- par(mfrow = c(1, 1))
#' plot(
#'     vmperception,
#'     -0.5, 8,
#'     main = paste(
#'         "perception probability of",
#'         "visual meteor magnitudes"
#'     ),
#'     col = "blue",
#'     xlab = "dm",
#'     ylab = "p"
#' )
#'
#' par(old_par)
#' @export
vmperception <- function(dm) {
    poly_coef <- c(0.0, 0.0037, 0.0019, 0.00271, 0.0009)
    names(poly_coef) <- seq(along = poly_coef) - 1 # exponents

    dm <- dm + 0.5
    p <- rep(0.0, length(dm))
    idx <- dm > .Machine$double.eps
    if (any(idx)) {
        f0 <- .f_polynomial(dm[idx], poly_coef)
        p[idx] <- 1.0 - exp(-f0)
    }

    p
}

#' build polynomial sum
#'
#' @noRd
.f_polynomial <- function(m, poly_coef) {
    exponents <- as.numeric(names(poly_coef))
    margin.table(poly_coef * t(outer(m, exponents, "^")), 2)
}

#' returns polynomial coefficients
#'
#' @noRd
.f_polynomial_coef <- function(poly_coef, deriv_degree = 1L) {
    if (0L == deriv_degree) {
        return(poly_coef)
    }

    if (1 == length(poly_coef)) {
        return(0)
    }

    exponents <- as.numeric(names(poly_coef))
    poly_coef <- poly_coef * exponents
    if (0L %in% exponents) {
        intercept_idx <- 0L == exponents
        poly_coef <- poly_coef[!intercept_idx]
        exponents <- exponents[!intercept_idx]
    }
    exponents <- exponents - 1L
    names(poly_coef) <- exponents

    .f_polynomial_coef(poly_coef, deriv_degree - 1L)
}

test_that("vmperception", {
    # test approximation of `KOSREN90` perceptions published in
    # J. Rendtel, Handbook for meteor observers 2022 edition, p. 149, (c) International Meteor Organization
    data <- data.frame(
        m = seq(-0.4, 7.6, 0.2),
        p = c(
            0.00046, 0.0011, 0.0023, 0.0046, 0.0081, 0.0122, 0.0182,
            0.0257, 0.03444, 0.04365, 0.05495, 0.06918, 0.08511, 0.10351,
            0.13, 0.16, 0.2, 0.24, 0.29, 0.34674, 0.4, 0.46, 0.52, 0.57544,
            0.63096, 0.67764, 0.71, 0.74, 0.77, 0.79, 0.81, 0.83, 0.85,
            0.87, 0.89, 0.91, 0.93, 0.94, 0.96, 0.98, 1.0
        )
    )
    data <- rbind(data.frame(m = -0.5, p = 0.0), data)
    p_fun <- approxfun(data$m, data$p, yleft = 0.0, yright = 1.0)

    # ignore lower magnitudes
    data0 <- subset(data, data$m >= 0.29)

    # check perceptions
    p <- vmperception(data0$m)
    expect_true(all(p < 1.0))
    expect_true(all(p > 0.0))
    pdiff <- (p - data0$p) / data0$p
    expect_lt(mean(abs(pdiff)), 0.057)
    expect_true(all(abs(pdiff) < 0.14))

    # Checks whether the approximation formula changes the r-value
    model <- with(new.env(), {
        r <- seq(1.6, 3.0, 0.1)
        limmag <- seq(5.6, 6.4, 0.2)
        m <- seq(-100, 6, 1)

        df <- expand.grid(r = r, limmag = limmag)
        do.call(
            rbind.data.frame,
            mapply(function(r, limmag) {
                p <- dvmgeom(m, limmag, r, perception_fun = p_fun)

                # maximum likelihood estimation (MLE) of r
                llr <- function(r) {
                    -sum(p * dvmgeom(m, limmag, r, log = TRUE))
                }
                r_est <- optim(r, llr, method = "Brent", lower = 1.1, upper = 5, hessian = FALSE)$par

                list(
                    r = r,
                    limmag = limmag,
                    r_est = r_est
                )
            }, df$r, df$limmag, SIMPLIFY = FALSE)
        )
    })

    rdiff <- (model$r - model$r_est)
    expect_lt(mean(abs(rdiff)), 0.0041)
    expect_true(all(abs(rdiff) < 0.022))

    # lower r values
    model_rlow <- subset(model, model$r < 2.7)

    rdiff <- (model_rlow$r - model_rlow$r_est)
    expect_lt(mean(abs(rdiff)), 0.0021)
    expect_true(all(abs(rdiff) < 0.009))
})

test_that("dvmgeom", {
    r <- 1.8
    lm <- 6.5

    # from documentation
    dp_fun <- function(m) {
        stats::dgeom(m, 1 - 1 / r) * vismeteor::vmperception(m + 0.5)
    }
    m <- seq(0, 40)
    expected_p <- dp_fun(m)
    expected_p <- expected_p / sum(expected_p)
    m <- 6.0 - m

    # density of meteor magnitudes
    p <- vismeteor::dvmgeom(m, lm, r)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_true(all(abs(p - expected_p) < 1e-10))
    expect_true(all(abs(log(p) - log(expected_p)) < 1e-09))

    p <- vismeteor::dvmgeom(c(6, 6, 6), lm = c(5.4, 5.5, 5.6), r)
    expect_type(p, "double")
    expect_length(p, 3)
    expect_equal(p[1], 0.0)
    expect_equal(p[2], 0.0)
    expect_gt(p[3], 0.0)

    # log density of meteor magnitudes
    p <- vismeteor::dvmgeom(m, lm, r, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_true(all(abs(p - log(expected_p)) < 1e-09))

    # density of infinite meteor magnitudes
    p <- vismeteor::dvmgeom(c(-Inf, Inf), lm, r)
    expect_type(p, "double")
    expect_length(p, 2)
    expect_equal(p[1], 0.0)
    expect_equal(p[2], 0.0)

    # log density of infinite meteor magnitudes
    p <- vismeteor::dvmgeom(c(-Inf, Inf), lm, r, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 2)
    expect_equal(p[1], -Inf)
    expect_equal(p[2], -Inf)

    # test order of densities of meteor magnitudes
    lm <- seq(2.5, 6.5, 0.1)
    p <- vismeteor::dvmgeom(rep(3L, length(lm)), lm, r)
    expect_equal(p[order(p)], p)

    # test densities of meteor magnitudes with different parameters
    p1 <- vismeteor::dvmgeom(3, 6.5, 1.8)
    p2 <- vismeteor::dvmgeom(4, 6.3, 2.0)
    p <- vismeteor::dvmgeom(c(3, 4), c(6.5, 6.3), c(1.8, 2.0))
    expect_equal(p, c(p1, p2))

    # Is the density sum of meteor magnitudes equal to 1.0 ?
    for (lm in seq(5.4, 6.4, 0.1)) {
        p <- sum(vismeteor::dvmgeom(as.integer(seq(6, -100, -1)), lm, r))
        expect_equal(round(p, 6), 1.0, label = paste("lm =", lm))
    }

    # test mean of meteor magnitudes
    q <- log(r)
    f <- function(m) {
        exp(-q * m) * vmperception(m)
    }
    norm <- stats::integrate(f, -0.5, Inf)$value
    f <- function(m) {
        m * exp(-q * m) * vmperception(m) / norm
    }
    expected_m_mean <- stats::integrate(f, -0.5, Inf)$value
    expect_equal(round(expected_m_mean, 2), 4.42)

    f <- function(m) {
        m * vismeteor::dvmgeom(m, lm, r)
    }
    m_mean <- sum(f(as.integer(6 - seq(0, 100, 1))))
    expect_equal(round(lm - m_mean, 2), round(expected_m_mean, 2))

    # test maximum likelihood estimation (MLE) of q of meteor magnitudes
    lm <- 5.8
    m <- as.integer(seq(6, -60, -1))
    p <- vismeteor::dvmgeom(m, lm, r)
    ll <- function(q) { # log likelihood function
        -sum(p * vismeteor::dvmgeom(m, lm, exp(q), log = TRUE))
    }
    est <- optim(1, ll, method = "Brent", lower = 0.01, upper = 4, hessian = TRUE)
    expect_equal(round(est$par, 6), round(base::log(r), 6))

    # density of meteor magnitudes equals geometric distribution
    perception_const <- function(m, log = FALSE) {
        rep(ifelse(log, 0.0, 1.0), length(m))
    }

    m <- as.integer(seq(6, -30, -1))
    p <- vismeteor::dvmgeom(m, lm, r, perception_fun = perception_const)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(stats::dgeom(6 - m, 1 - 1 / r), p)
})

test_that("cvmideal", {
    psi <- 4.0
    lm <- 6.5

    # from documentation
    dp_fun <- function(m, psi) {
        r <- 10^0.4
        1.5 * log(r) * sqrt(r^(3 * psi + 2 * m) / ((r^psi + r^m)^5))
    }

    # test convolution
    m <- seq(6, -15, -1)
    expected_p <- sum(sapply(m, function(m) {
        stats::integrate(function(m) dp_fun(m, psi), m - 0.5, m + 0.5)$value *
            vismeteor::vmperception(lm - m)
    }))

    p <- vismeteor::cvmideal(lm, psi)
    expect_type(p, "double")
    expect_length(p, 1)
    expect_equal(round(p, 5), round(expected_p, 5))
    expect_equal(round(log(p), 3), round(log(expected_p), 3))

    p <- vismeteor::cvmideal(lm, psi, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 1)
    expect_equal(round(p, 3), round(log(expected_p), 3))

    p <- vismeteor::cvmideal(c(-15, 20), psi)
    expect_type(p, "double")
    expect_length(p, 2)
    expect_equal(round(p, 5), c(0.0, 1.0))

    p <- vismeteor::cvmideal(lm, c(-15, 20))
    expect_type(p, "double")
    expect_length(p, 2)
    expect_equal(round(p, 5), c(1.0, 0.0))

    p <- suppressWarnings(
        vismeteor::cvmideal(c(-Inf, -Inf, Inf, Inf), c(-Inf, Inf, -Inf, Inf))
    )
    expect_type(p, "double")
    expect_length(p, 4)
    expect_true(is.na(p[1]))
    expect_equal(p[2], 0.0)
    expect_equal(p[3], 1.0)
    expect_true(is.na(p[4]))

    p <- suppressWarnings(
        vismeteor::cvmideal(c(-Inf, -Inf, Inf, Inf), c(-Inf, Inf, -Inf, Inf), log = TRUE)
    )
    expect_type(p, "double")
    expect_length(p, 4)
    expect_true(is.na(p[1]))
    expect_equal(p[2], -Inf)
    expect_equal(p[3], 0.0)
    expect_true(is.na(p[4]))
})

test_that("dvmideal", {
    lm <- 6.5
    psi <- 4.0

    # from documentation
    dp_fun <- function(m, psi) {
        r <- 10^0.4
        1.5 * log(r) * sqrt(r^(3 * psi + 2 * m) / ((r^psi + r^m)^5))
    }

    for (current.psi in c(-10.0, psi, 5)) {
        m <- seq(6, current.psi - 25, -1)
        expected_p <- sapply(m, function(m) {
            stats::integrate(function(m) dp_fun(m, current.psi), m - 0.5, m + 0.5)$value *
                vismeteor::vmperception(lm - m)
        })
        expected_p <- expected_p / sum(expected_p)

        # density of meteor magnitudes
        p <- vismeteor::dvmideal(m, lm, current.psi)
        expect_type(p, "double")
        expect_length(p, length(m))
        expect_false(any(abs(log(p) - log(expected_p)) > 0.03), label = paste("psi =", current.psi))

        # log density of meteor magnitudes
        p <- vismeteor::dvmideal(m, lm, current.psi, log = TRUE)
        expect_type(p, "double")
        expect_length(p, length(m))
        expect_false(any(abs(p - log(expected_p)) > 0.03), label = paste("psi =", current.psi))

        # Is the density sum of meteor magnitudes equal to 1.0 ?
        p <- sum(vismeteor::dvmideal(m, lm, current.psi))
        expect_equal(round(p, 6), 1.0, label = paste("psi =", current.psi))
    }

    p <- vismeteor::dvmideal(7, lm, psi)
    expect_type(p, "double")
    expect_length(p, 1)
    expect_equal(p, 0.0)

    # log density of meteor magnitudes
    p <- vismeteor::dvmideal(7, lm, psi, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 1)
    expect_equal(p, -Inf)

    p <- vismeteor::dvmideal(c(6, 6, 6), lm = c(5.4, 5.5, 5.6), psi)
    expect_type(p, "double")
    expect_length(p, 3)
    expect_equal(p[1], 0.0)
    expect_equal(p[2], 0.0)
    expect_gt(p[3], 0.0)

    # density of infinite meteor magnitudes
    p <- vismeteor::dvmideal(c(-Inf, Inf), lm, psi)
    expect_type(p, "double")
    expect_length(p, 2)
    expect_equal(p[1], 0.0)
    expect_equal(p[2], 0.0)

    # log density of infinite meteor magnitudes
    p <- vismeteor::dvmideal(c(-Inf, Inf), lm, psi, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 2)
    expect_equal(p[1], -Inf)
    expect_equal(p[2], -Inf)

    # test order of densities of meteor magnitudes
    lm <- seq(2.5, 6.5, 0.1)
    p <- vismeteor::dvmideal(rep(3, length(lm)), lm, psi)
    expect_equal(p[order(p)], p)

    # test maximum likelihood estimation (MLE) of psi
    lm <- 5.8
    m <- as.integer(seq(-20, 6))
    p <- vismeteor::dvmideal(m, lm, psi)
    ll <- function(psi) { # log likelihood function
        -sum(p * vismeteor::dvmideal(m, lm, psi, log = TRUE))
    }
    est <- optim(1, ll, method = "Brent", lower = 0, upper = 7, hessian = TRUE)
    expect_equal(round(est$par, 6), psi)

    # density of meteor magnitudes equals geometric distribution
    lm <- 6.3
    m <- as.integer(seq(-20, 6))

    p <- vismeteor::dvmideal(m, lm, 16.25, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    diff.p <- abs(p - dvmgeom(m, lm, 10^0.4, log = TRUE))
    expect_true(any(diff.p > 0.0))
    expect_false(any(diff.p > 1e-08))

    p <- vismeteor::dvmideal(m, lm, 16.35, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, dvmgeom(m, lm, 10^0.4, log = TRUE))
})

#' @name vmgeom
#' @aliases dvmgeom
#' @aliases pvmgeom
#' @aliases qvmgeom
#' @aliases rvmgeom
#' @title Geometric Model of Visual Meteor Magnitudes
#' @description
#' Density, distribution function, quantile function, and random generation
#' for the geometric model of visual meteor magnitudes.
#' @param m integer; the meteor magnitude.
#' @param p numeric; probability.
#' @param lm numeric; limiting magnitude.
#' @param r numeric; the population index.
#' @param n numeric; count of meteor magnitudes.
#' @param log logical; if `TRUE`, probabilities p are given as `log(p)`.
#' @param lower.tail logical; if `TRUE` (default) probabilities are
#'     \eqn{P[M < m]}, otherwise, \eqn{P[M \ge m]}.
#' @param perception_fun function; perception probability function (optional).
#'     Default is [vismeteor::vmperception].
#' @details
#' In visual meteor observations, magnitudes are estimated as integer values.
#' Consequently, the distribution of observed magnitudes is discrete, and its
#' probability mass function is given by
#'
#' \deqn{
#' P[M = m] \sim
#' \begin{cases}
#'   f(m_{\mathrm{lim}} - m)\, r^m, & \text{if } m_{\mathrm{lim}} - m > -0.5,\\[5pt]
#'   0 & \text{otherwise,}
#' \end{cases}
#' }
#'
#' where \eqn{m_{\mathrm{lim}}} denotes the limiting (non-integer) magnitude of the observation,
#' and \eqn{m} the integer meteor magnitude.
#' The function \eqn{f(\cdot)} denotes the [perception probability function][vismeteor::vmperception].
#'
#' Thus, the distribution is the product of the perception probabilities and the
#' underlying [geometric distribution][stats::Geometric] of meteor magnitudes.
#' Therefore, the parameter `p` of the geometric distribution is given by `p = 1 - 1/r`.
#'
#' The parameter `lm` specifies the limiting magnitude for the meteor magnitude `m`.
#' `m` must be an integer meteor magnitude.
#' The length of the vector `lm` must either equal the length of the vector `m`,
#' or `lm` must be a scalar value.
#' In the case of `rvmgeom`, the length of the vector `lm` must equal `n`,
#' or `lm` must be a scalar value.
#'
#' If a different perception probability function `perception_fun` is provided,
#' it must have the signature `function(x)` and return the perception probability of
#' the difference `x` between the limiting magnitude and the meteor magnitude.
#' If `x >= 15.0`, the function `perception_fun` should return a perception probability of `1.0`.
#' The argument `perception_fun` is resolved using [match.fun].
#'
#' @return
#' - `dvmgeom`: density
#' - `pvmgeom`: distribution function
#' - `qvmgeom`: quantile function
#' - `rvmgeom`: random generation
#'
#' The length of the result is determined by `n` for `rvmgeom`, and by the maximum
#' of the lengths of the numeric vector arguments for the other functions.
#' All arguments are vectorized; standard R recycling rules apply.
#'
#' Since the distribution is discrete, `qvmgeom` and `rvmgeom` always return integer values.
#' `qvmgeom` may return `NaN` with a warning.
#' @seealso [vismeteor::vmperception]
#'   [stats::Geometric]
#' @examples
#' N <- 100
#' r <- 2.0
#' limmag <- 6.5
#' (m <- seq(6, -7))
#'
#' # discrete density of `N` meteor magnitudes
#' (freq <- round(N * dvmgeom(m, limmag, r)))
#'
#' # log likelihood function
#' lld <- function(r) {
#'     -sum(freq * dvmgeom(m, limmag, r, log = TRUE))
#' }
#'
#' # maximum likelihood estimation (MLE) of r
#' est <- optim(2, lld, method = "Brent", lower = 1.1, upper = 4)
#'
#' # estimations
#' est$par # mean of r
#'
#' # generate random meteor magnitudes
#' m <- rvmgeom(N, r, lm = limmag)
#'
#' # log likelihood function
#' llr <- function(r) {
#'     -sum(dvmgeom(m, limmag, r, log = TRUE))
#' }
#'
#' # maximum likelihood estimation (MLE) of r
#' est <- optim(2, llr, method = "Brent", lower = 1.1, upper = 4, hessian = TRUE)
#'
#' # estimations
#' est$par # mean of r
#' sqrt(1 / est$hessian[1][1]) # standard deviation of r
#'
#' m <- seq(6, -4, -1)
#' p <- vismeteor::dvmgeom(m, limmag, r)
#' barplot(
#'     p,
#'     names.arg = m,
#'     main = paste0("Density (r = ", r, ", limmag = ", limmag, ")"),
#'     col = "blue",
#'     xlab = "m",
#'     ylab = "p",
#'     border = "blue",
#'     space = 0.5
#' )
#' axis(side = 2, at = pretty(p))

#' @rdname vmgeom
#' @export
dvmgeom <- function(m, lm, r, log = FALSE, perception_fun = vmperception) {
    if (anyNA(m) || anyNA(lm) || anyNA(r)) {
        stop("NA's are not allowed!")
    }

    if (any(is.infinite(lm))) {
        stop("Infinite limiting magnitudes are not allowed!")
    }

    if (any(is.infinite(r))) {
        stop("Infinite r values are not allowed!")
    }

    if (any(1.0 > r)) {
        stop(paste0('r must be greater than 1.0 instead of "', r, '"!'))
    }

    is_wholenumber <- function(x, tol = .Machine$double.eps^0.5) all(is.infinite(x) | abs(x - round(x)) < tol)
    if (!is_wholenumber(m)) {
        stop("magnitudes must be integer values!")
    }

    perception_fun <- match.fun(perception_fun)

    p_geom <- 1.0 - 1.0 / r

    if (1 == length(r)) {
        r <- rep(r, length(m))
        p_geom <- rep(p_geom, length(m))
    }

    std_res <- .vmgeom_std(m, lm)
    m <- std_res$m
    offset <- std_res$offset

    f_density <- function(m, offset, p_geom) {
        m_max <- 15L
        idx <- m <= m_max
        d <- stats::dgeom(m, p_geom, log = TRUE)
        if (any(idx)) {
            d[idx] <- d[idx] + base::log(perception_fun(m[idx] + offset))
        }

        d - base::log(.vmgeom_norm(offset, p_geom, m_max, perception_fun))
    }

    arg_data <- data.frame(
        m = m,
        offset = offset,
        p_geom = p_geom
    )

    data_f <- as.factor(paste0(offset, "/", p_geom))
    data_s <- split(arg_data, data_f)
    d <- lapply(data_s, \(data) {
        m <- data$m
        offset <- data$offset[1]
        p_geom <- data$p_geom[1]
        d <- rep(-Inf, length(m))

        idx <- m > -1
        if (any(idx)) {
            d[idx] <- f_density(m[idx], offset, p_geom)
        }

        if (!log) {
            d[idx] <- exp(d[idx])
            d[!idx] <- 0.0
        }

        d
    })

    unsplit(d, data_f)
}

#' @rdname vmgeom
#' @export
pvmgeom <- function(m, lm, r, lower.tail = TRUE, log = FALSE, perception_fun = vmperception) {
    if (anyNA(m) || anyNA(lm) || anyNA(r)) {
        stop("NA's are not allowed!")
    }

    if (any(is.infinite(lm))) {
        stop("Infinite limiting magnitudes are not allowed!")
    }

    if (any(is.infinite(r))) {
        stop("Infinite r values are not allowed!")
    }

    if (any(1.0 > r)) {
        stop(paste0('r must be greater than 1.0 instead of "', r, '"!'))
    }

    is_wholenumber <- function(x, tol = .Machine$double.eps^0.5) all(is.infinite(x) | abs(x - round(x)) < tol)
    if (!is_wholenumber(m)) {
        stop("magnitudes must be integer values!")
    }

    perception_fun <- match.fun(perception_fun)

    p_geom <- 1.0 - 1.0 / r

    if (1 == length(r)) {
        r <- rep(r, length(m))
        p_geom <- rep(p_geom, length(m))
    }

    std_res <- .vmgeom_std(m, lm)
    m <- std_res$m
    offset <- std_res$offset

    f_density <- function(m, offset, p_geom) {
        stats::dgeom(m, p_geom) * perception_fun(m + offset)
    }

    f_sum <- Vectorize(\(m, offset, p_geom) {
        m <- as.integer(seq(0, m))
        sum(f_density(m, offset, p_geom))
    })

    f_prob <- function(m, offset, p_geom) {
        m_max <- 15L
        norm <- .vmgeom_norm(offset, p_geom, m_max, perception_fun)
        p <- rep(0.0, length(m))

        if (lower.tail) {
            idx <- m <= m_max
            if (any(idx)) {
                p[idx] <- 1.0 - f_sum(m[idx], offset, p_geom) / norm
            }

            idx <- m > m_max
            if (any(idx)) {
                p[idx] <- stats::pgeom(m[idx], p_geom, lower.tail = FALSE) / norm
            }
        } else {
            idx <- m <= m_max
            if (any(idx)) {
                p[idx] <- f_sum(m[idx], offset, p_geom) / norm
            }

            idx <- m > m_max
            if (any(idx)) {
                p[idx] <- 1.0 - stats::pgeom(m[idx], p_geom, lower.tail = FALSE) / norm
            }
        }

        p
    }

    arg_data <- data.frame(
        m = m,
        offset = offset,
        p_geom = p_geom
    )

    data_f <- as.factor(paste0(offset, "/", p_geom))
    data_s <- split(arg_data, data_f)
    p <- lapply(data_s, \(data) {
        m <- data$m
        offset <- data$offset[1]
        p_geom <- data$p_geom[1]

        if (lower.tail) {
            p <- rep(1.0, length(m))
        } else {
            p <- rep(0.0, length(m))
        }

        idx <- m > -1
        if (any(idx)) {
            p[idx] <- f_prob(m[idx], offset, p_geom)
        }

        if (log) {
            p[idx] <- base::log(p[idx])
            if (lower.tail) {
                p[!idx] <- 0.0
            } else {
                p[!idx] <- -Inf
            }
        }

        p
    })

    unsplit(p, data_f)
}

#' @rdname vmgeom
#' @export
qvmgeom <- function(p, lm, r, lower.tail = TRUE, perception_fun = vmperception) {
    if (anyNA(p) || anyNA(lm) || anyNA(r)) {
        stop("NA's are not allowed!")
    }

    if (any(is.infinite(lm))) {
        stop("Infinite limiting magnitudes are not allowed!")
    }

    if (any(is.infinite(r))) {
        stop("Infinite r values are not allowed!")
    }

    if (any(1.0 > r)) {
        stop(paste0('r must be greater than 1.0 instead of "', r, '"!'))
    }

    perception_fun <- match.fun(perception_fun)

    p_geom <- 1.0 - 1.0 / r

    if (1 == length(r)) {
        r <- rep(r, length(p))
        p_geom <- rep(p_geom, length(p))
    }

    lm_round <- round(lm)
    offset <- lm - lm_round
    lm <- lm_round

    idx <- -0.5 == offset
    lm[idx] <- lm[idx] - 1L
    offset[idx] <- offset[idx] + 1.0

    if (1 == length(offset)) {
        offset <- rep(offset, length(p))
    }

    arg_data <- data.frame(
        p = p,
        offset = offset,
        p_geom = p_geom,
        r = r
    )

    data_f <- as.factor(paste0(offset, "/", p_geom))
    data_s <- split(arg_data, data_f)
    m <- lapply(data_s, \(data) {
        m_max <- 15L
        p <- data$p
        offset <- data$offset[1]
        p_geom <- data$p_geom[1]
        r <- data$r[1]
        m <- rep(NA, length(p))

        if (lower.tail) {
            m[1.0 == p] <- 0L
            p_max <- 1.0 - vismeteor::pvmgeom(0, m_max + offset, r, lower.tail = FALSE, perception_fun = perception_fun)
            idx <- p >= 0.0 & p < p_max
            if (any(idx)) {
                m[idx] <- m_max + 1 + stats::qgeom(p[idx] / p_max, p_geom, lower.tail = FALSE)
            }

            idx <- p >= p_max & p < 1.0
            if (any(idx)) {
                m0 <- seq(-m_max, 0, 1)
                p0 <- c(vismeteor::pvmgeom(m0, offset, r, lower.tail = TRUE, perception_fun = perception_fun), 1.0)
                m0 <- c(m0, 1L)
                p_idx <- findInterval(p[idx], p0, left.open = FALSE)
                m[idx] <- -m0[p_idx]
            }
        } else {
            m[0.0 == p] <- 0L
            p_max <- vismeteor::pvmgeom(0, m_max + offset, r, lower.tail = FALSE, perception_fun = perception_fun)
            idx <- p > p_max & p <= 1.0
            if (any(idx)) {
                m[idx] <- m_max + 1L + stats::qgeom((p[idx] - p_max) / (1.0 - p_max), p_geom)
            }

            idx <- p > 0.0 & p <= p_max
            if (any(idx)) {
                m0 <- seq(1, -m_max, -1)
                p0 <- vismeteor::pvmgeom(m0, offset, r, lower.tail = FALSE, perception_fun = perception_fun)
                p_idx <- findInterval(p[idx], p0, left.open = TRUE) + 1
                m[idx] <- -m0[p_idx]
                m[m < 0] <- NA
            }
        }

        m
    })

    m <- lm - unsplit(m, data_f)

    if (anyNA(m)) {
        warning("NaNs produced")
    }

    as.numeric(m)
}

#' @rdname vmgeom
#' @export
rvmgeom <- function(n, lm, r, perception_fun = vmperception) {
    if (anyNA(lm) || anyNA(r)) {
        stop("NA's are not allowed!")
    }

    if (any(is.infinite(lm))) {
        stop("Infinite limiting magnitudes are not allowed!")
    }

    if (any(is.infinite(r))) {
        stop("Infinite r values are not allowed!")
    }

    if (any(1.0 > r)) {
        stop(paste0('r must be greater than 1.0 instead of "', r, '"!'))
    }

    perception_fun <- match.fun(perception_fun)

    p <- stats::runif(n)
    m <- rep(NA, n)
    idx <- p < 0.5

    if (any(idx)) {
        m[idx] <- vismeteor::qvmgeom(p[idx], lm, r, lower.tail = TRUE, perception_fun = perception_fun)
    }

    if (any(!idx)) {
        m[!idx] <- vismeteor::qvmgeom(1.0 - p[!idx], lm, r, lower.tail = FALSE, perception_fun = perception_fun)
    }

    m
}

#' standardization
#'
#' @noRd
.vmgeom_std <- function(m, lm) {
    m <- lm - m
    m_round <- round(m)
    offset <- rep(0.0, length(m))
    idx <- is.infinite(m)
    offset[!idx] <- m[!idx] - m_round[!idx]
    m <- m_round

    idx <- -0.5 == offset
    m[idx] <- m[idx] - 1L
    offset[idx] <- offset[idx] + 1.0

    list(
        m = m,
        offset = offset
    )
}

#' normalization
#'
#' @noRd
.vmgeom_norm <- function(offset, p_geom, m_max, perception_fun) {
    m <- as.integer(seq(0, m_max))
    sum(stats::dgeom(m, p_geom) * perception_fun(m + offset)) +
        stats::pgeom(m_max, p_geom, lower.tail = FALSE)
}

test_that("qvmgeom", {
    r <- 1.8

    p <- vismeteor::pvmgeom(6, 5.5, r, lower.tail = FALSE)
    q <- suppressWarnings(
        vismeteor::qvmgeom(c(-1.0, 0.0, p - 1e-06, p, p + 1e-06, 1.0), 5.5, r, lower.tail = FALSE)
    )
    expect_equal(q, c(NA, 5, NA, 5, 5, -Inf))

    p <- vismeteor::pvmgeom(6, 5.5, r, lower.tail = TRUE)
    q <- suppressWarnings(
        vismeteor::qvmgeom(c(-1.0, 1.0, p + 1e-06, p, p - 1e-06, 0.0), 5.5, r, lower.tail = TRUE)
    )
    expect_equal(q, c(NA, 5, NA, 5, 5, -Inf))

    lms <- seq(5.6, 6.5, 0.1)

    for (lm in lms) {
        q <- suppressWarnings(
            vismeteor::qvmgeom(c(-1.0, 0.0, 1.0), lm, r, lower.tail = FALSE)
        )
        expect_equal(q, c(NA, 6, -Inf), label = paste("lm =", lm))

        q <- suppressWarnings(
            vismeteor::qvmgeom(c(-1.0, 0.0, 1.0), lm, r, lower.tail = TRUE)
        )
        expect_equal(q, c(NA, -Inf, 6), label = paste("lm =", lm))
    }

    for (lm in lms) {
        p <- vismeteor::pvmgeom(6, lm, r, lower.tail = FALSE)
        q <- vismeteor::qvmgeom(c(p - 1e-06, p, p + 1e-06), lm, r, lower.tail = FALSE)
        expect_equal(q, c(6, 6, 5), label = paste("lm =", lm))

        p <- vismeteor::pvmgeom(6, lm, r, lower.tail = TRUE)
        q <- vismeteor::qvmgeom(c(p + 1e-06, p - 1e-06), lm, r, lower.tail = TRUE)
        expect_equal(q, c(6, 5), label = paste("lm =", lm))
    }

    for (lm in lms) {
        p <- vismeteor::pvmgeom(5, lm, r, lower.tail = FALSE)
        q <- vismeteor::qvmgeom(c(p - 1e-06, p, p + 1e-06), lm, r, lower.tail = FALSE)
        expect_equal(q, c(5, 5, 4), label = paste("lm =", lm))

        p <- vismeteor::pvmgeom(5, lm, r, lower.tail = TRUE)
        q <- vismeteor::qvmgeom(c(p + 1e-06, p - 1e-06), lm, r, lower.tail = TRUE)
        expect_equal(q, c(5, 4), label = paste("lm =", lm))
    }

    # quantile of meteor magnitudes equals geometric distribution

    perception_const <- function(m, log = FALSE) {
        rep(ifelse(log, 0.0, 1.0), length(m))
    }

    m <- as.integer(seq(0, 17, 1))
    p <- round(vismeteor::pvmgeom(6 - m, 6.5, r, lower.tail = FALSE, perception_fun = perception_const), 6)
    m <- vismeteor::qvmgeom(p, 6.5, r, lower.tail = FALSE, perception_fun = perception_const)
    expect_type(m, "double")
    expect_length(m, length(m))
    expect_equal(stats::qgeom(p, 1 - 1 / r, lower.tail = TRUE), 6 - m)

    m <- as.integer(seq(10, 30, 1))
    p <- round(vismeteor::pvmgeom(6 - m, 6.5, r, lower.tail = TRUE, perception_fun = perception_const), 6)
    m <- vismeteor::qvmgeom(p, 6.5, r, lower.tail = TRUE, perception_fun = perception_const)
    expect_type(m, "double")
    expect_length(m, length(m))
    expect_equal(stats::qgeom(p, 1 - 1 / r, lower.tail = FALSE), 6 - m)
})

test_that("pvmgeom", {
    r <- 1.8
    lm <- 6.5

    # from documentation
    dp_fun <- function(m) {
        stats::dgeom(m, 1 - 1 / r) * vismeteor::vmperception(m + 0.5)
    }

    # probability lower tail

    m <- seq(40, 0)
    expected_p <- dp_fun(m)
    expected_p[1] <- expected_p[1] + stats::pgeom(m[1], 1 - 1 / r, lower.tail = FALSE)
    expected_p <- expected_p / sum(expected_p)
    expected_p <- cumsum(expected_p) # lower tail
    m <- 6.0 - m

    p <- vismeteor::pvmgeom(m + 1, lm, r, lower.tail = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, expected_p)
    expect_equal(log(p), log(expected_p))

    p <- vismeteor::pvmgeom(m + 1, lm, r, lower.tail = TRUE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, log(expected_p))

    # probability upper tail

    m <- seq(0, 40)
    expected_p <- dp_fun(m)
    expected_p <- expected_p / sum(expected_p)
    expected_p <- cumsum(expected_p) # upper tail
    m <- 6.0 - m

    # probability
    p <- vismeteor::pvmgeom(m, lm, r, lower.tail = FALSE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, expected_p)
    expect_equal(log(p), log(expected_p))

    p <- vismeteor::pvmgeom(m, lm, r, lower.tail = FALSE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, log(expected_p))

    # sum lower.tail and upper.tail
    p <- vismeteor::pvmgeom(m, lm, r, lower.tail = FALSE) + vismeteor::pvmgeom(m, lm, r, lower.tail = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, rep(1.0, length(m)))
    expect_equal(log(p), rep(0.0, length(m)))

    p <- vismeteor::pvmgeom(6, c(5.4, 5.5, 5.6, 100), r, lower.tail = FALSE)
    expect_type(p, "double")
    expect_length(p, 4)
    expect_equal(p[1], 0.0)
    expect_equal(p[2], 0.0)
    expect_equal(p[3], vismeteor::dvmgeom(6, 5.6, r))
    expect_equal(p[4], 1.0)

    p <- vismeteor::pvmgeom(6, c(5.4, 5.5, 5.6, 100), r, lower.tail = FALSE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 4)
    expect_equal(p[1], -Inf)
    expect_equal(p[2], -Inf)
    expect_equal(p[3], vismeteor::dvmgeom(6, 5.6, r, log = TRUE))
    expect_equal(p[4], 0.0)

    p <- vismeteor::pvmgeom(6, c(5.4, 5.5, 5.6, 100), r, lower.tail = TRUE)
    expect_type(p, "double")
    expect_length(p, 4)
    expect_equal(p[1], 1.0)
    expect_equal(p[2], 1.0)
    expect_equal(p[3], 1 - vismeteor::dvmgeom(6, 5.6, r))
    expect_equal(p[4], 0.0)

    p <- vismeteor::pvmgeom(6, c(5.4, 5.5, 5.6, 100), r, lower.tail = TRUE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 4)
    expect_equal(p[1], 0.0)
    expect_equal(p[2], 0.0)
    expect_equal(p[3], log(1 - vismeteor::dvmgeom(6, 5.6, r)))
    expect_lt(p[4], -50)

    # test probabilities of meteor magnitudes with different parameters
    p1 <- vismeteor::pvmgeom(3, 6.5, 1.8)
    p2 <- vismeteor::pvmgeom(4, 6.3, 2.0)
    p <- vismeteor::pvmgeom(c(3, 4), c(6.5, 6.3), c(1.8, 2.0))
    expect_equal(p, c(p1, p2))

    # test order of probabilities
    lm <- seq(2.5, 6.5, 0.1)
    p <- vismeteor::pvmgeom(rep(3L, length(lm)), lm, r, lower.tail = FALSE)
    expect_equal(p, p[order(p, decreasing = FALSE)])

    lm <- seq(2.5, 6.5, 0.1)
    p <- vismeteor::pvmgeom(rep(3L, length(lm)), lm, r, lower.tail = TRUE)
    expect_equal(p, p[order(p, decreasing = TRUE)])

    # test lower probabilities
    lm <- seq(5.4, 6.4, 0.1)
    p1 <- vismeteor::dvmgeom(rep(6L, length(lm)), lm, r)
    p2 <- vismeteor::pvmgeom(rep(6L, length(lm)), lm, r, lower.tail = FALSE)
    expect_equal(p1, p2)

    # test sum lower probabilities
    lm <- seq(5.4, 6.4, 0.1)
    p1 <- vismeteor::dvmgeom(rep(6L, length(lm)), lm, r) +
        vismeteor::dvmgeom(rep(5L, length(lm)), lm, r)
    p2 <- vismeteor::pvmgeom(rep(5L, length(lm)), lm, r, lower.tail = FALSE)
    expect_equal(p1, p2)

    # probability of meteor magnitudes equals geometric distribution

    perception_const <- function(m, log = FALSE) {
        rep(ifelse(log, 0.0, 1.0), length(m))
    }

    m <- seq(0, 30, 1)
    p <- vismeteor::pvmgeom(6 - m, 6.5, r, lower.tail = FALSE, perception_fun = perception_const)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, stats::pgeom(m, 1 - 1 / r, lower.tail = TRUE))

    m <- seq(0, 30, 1)
    p <- vismeteor::pvmgeom(6 - m, 6.5, r, lower.tail = TRUE, perception_fun = perception_const)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, stats::pgeom(m, 1 - 1 / r, lower.tail = FALSE))
})

test_that("rvmgeom", {
    r <- 1.8

    perception_const <- function(m, log = FALSE) {
        rep(ifelse(log, 0.0, 1.0), length(m))
    }

    with_seed <- function(seed, code) {
        code <- substitute(code)
        orig_seed <- .Random.seed
        on.exit(.Random.seed <<- orig_seed)
        set.seed(seed)
        eval.parent(code)
    }

    # meteor magnitudes <= 6.0
    m <- vismeteor::rvmgeom(1000, 6.5, r)
    expect_type(m, "double")
    expect_length(m, 1000)
    expect_false(anyNA(m))
    expect_false(anyNA(is.infinite(m)))
    expect_true(all(m <= 6.0))
    expect_equal(as.integer(m), m)

    # Tests with perception_fun
    m <- with_seed(7, vismeteor::rvmgeom(1000, 6.5, r, perception_fun = perception_const))
    expect_type(m, "double")
    expect_length(m, 1000)
    expect_false(anyNA(m))
    expect_false(anyNA(is.infinite(m)))
    expect_true(all(m <= 6.0))
    expect_equal(as.integer(m), m)
    m_geom <- with_seed(7, stats::rgeom(10000, 1.0 - 1 / r))
    expect_lt(mean(m_geom) - mean(6 - m), 0.1)
})

test_that("rvmideal", {
    psi <- 4.0

    with_seed <- function(seed, code) {
        code <- substitute(code)
        orig_seed <- .Random.seed
        on.exit(.Random.seed <<- orig_seed)
        set.seed(seed)
        eval.parent(code)
    }

    # meteor magnitudes <= 6.0
    m <- vismeteor::rvmideal(1000, 6.5, psi)
    expect_type(m, "double")
    expect_length(m, 1000)
    expect_false(anyNA(m))
    expect_false(any(is.infinite(m)))
    expect_true(all(m <= 6.0))
    expect_equal(as.integer(m), m)

    # meteor magnitudes equals geometric distribution
    lm <- 6.3
    m <- with_seed(8, vismeteor::rvmideal(100000, lm, 30))
    expect_type(m, "double")
    expect_length(m, 100000)
    expect_false(anyNA(m))
    expect_false(any(is.infinite(m)))
    expect_true(all(m <= 6))
    expect_equal(as.integer(m), m)
    m_geom <- with_seed(8, rvmgeom(100000, lm, 10^(1 / 2.5)))
    expect_equal(round(mean(m_geom), 2), round(mean(m), 2))
})

test_that("vmgeomVst", {
    r <- seq(1.5, 3.3, 0.1)
    limmag <- c(5.5, 5.52, 5.55, seq(5.6, 6.4, 0.1), 6.45, 6.48, 6.5)
    m <- seq(-200, 6, 1)
    model <- expand.grid(r = r, limmag = limmag)
    model <- do.call(
        rbind.data.frame,
        mapply(function(r, limmag) {
            p <- dvmgeom(m, limmag, r)
            t <- vmgeomVstFromMagn(m, limmag)
            t.mean <- sum(p * t)
            t.var <- sum(p * (t - t.mean)^2)

            list(
                r = r,
                q = log(r),
                limmag = limmag,
                t.mean = t.mean,
                t.var = t.var
            )
        }, model$r, model$limmag , SIMPLIFY = FALSE)
    )

    #
    # test vmgeomVstFromMagn
    #
    expect_true(vmgeomVstFromMagn(6, 5.5) > 0.0)
    expect_true(all(abs(model$t.var - 1.0) < 0.018))

    #
    # test vmgeomVstToR
    #
    model$r.est <- vmgeomVstToR(model$t.mean)
    expect_true(all(abs(model$r - model$r.est) < 0.013))

    # test with non-exotic values
    model0 <- with(model, {
        subset(model, r < 2.7)
    })
    expect_true(all(abs(model0$r - model0$r.est) < 0.007))

    # test first derivative
    f <- function(x) {
        vmgeomVstToR(x, deriv.degree = 1L)
    }
    y <- vmgeomVstToR(5.5) - vmgeomVstToR(4.5)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)

    # test second derivative
    f <- function(x) {
        vmgeomVstToR(x, deriv.degree = 2L)
    }
    y <- vmgeomVstToR(5.5, deriv.degree = 1L) - vmgeomVstToR(4.5, deriv.degree = 1L)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)

    # log ...
    model$q.est <- vmgeomVstToR(model$t.mean, log = TRUE)
    expect_true(all(abs(model$q - model$q.est) < 0.004))

    # test with non-exotic values
    model0 <- with(model, {
        subset(model, r < 2.7)
    })
    expect_true(all(abs(model0$q - model0$q.est) < 0.004))

    # test log first derivative
    f <- function(x) {
        vmgeomVstToR(x, log = TRUE, deriv.degree = 1L)
    }
    y <- vmgeomVstToR(5.5, log = TRUE) - vmgeomVstToR(4.5, log = TRUE)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)

    # test log second derivative
    f <- function(x) {
        vmgeomVstToR(x, log = TRUE, deriv.degree = 2L)
    }
    y <- vmgeomVstToR(5.5, log = TRUE, deriv.degree = 1L) -
        vmgeomVstToR(4.5, log = TRUE, deriv.degree = 1L)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)
})

#' @name vmideal_vst
#' @aliases vmideal_vst_from_magn
#' @aliases vmideal_vst_to_psi
#' @title Variance-stabilizing Transformation for the Ideal Distribution of Visual Meteor Magnitudes
#'
#' @description
#' Applies a variance-stabilizing transformation to meteor magnitudes
#' under the assumption of the ideal magnitude distribution.
#'
#' @param m integer; the meteor magnitude.
#' @param lm numeric; limiting magnitude.
#' @param tm numeric; transformed magnitude.
#' @param deriv_degree integer; the degree of the derivative at `tm` to return
#'   instead of `psi`. Must be `0`, `1` or `2`.
#'
#' @details
#' Many linear models require the variance of visual meteor magnitudes to be
#' homoscedastic. The function `vmideal_vst_from_magn` applies a transformation
#' that produces homoscedastic distributions of visual meteor magnitudes if the
#' underlying magnitudes follow the ideal magnitude distribution.
#' In this sense, the transformation acts as a normalization of meteor magnitudes
#' and yields a variance close to `1.0`.
#'
#' The ideal distribution of visual meteor magnitudes
#' depends on the [parameter psi][vismeteor::vmideal] and the limiting magnitude `lm`,
#' resulting in a two-parameter distribution. Without detection probabilities,
#' the magnitude distribution reduces to a pure [ideal magnitude distribution][vismeteor::mideal],
#' which depends only on the parameter `psi`. Since the
#' limiting magnitude `lm` is a fixed parameter and never estimated
#' statistically, the magnitudes can be transformed such that, for example,
#' the mean of the transformed magnitudes directly provides an estimate of `psi`
#' using the function `vmideal_vst_to_psi`.
#'
#' This transformation is valid for \eqn{-10 \le \texttt{psi} \le 9}.
#' The numerical form of the transformation is version-specific and may change
#' substantially in future releases. Do not rely on equality of transformed
#' values across package versions.
#'
#' @return
#' * `vmideal_vst_from_magn`: a numeric value, the transformed meteor magnitude.
#' * `vmideal_vst_to_psi`: a numeric value of the parameter `psi`, derived from
#'   the mean of `tm`.
#' The argument `deriv_degree` can be used to apply the delta method.
#'
#' @note
#' The internal approximations used here are derived from the perception
#' probabilities produced by [vismeteor::vmperception].
#' For details on the derivation, see the script `inst/derivation/vmideal_vst.R` in the
#' package's source code.
#'
#' @seealso [vismeteor::vmideal] [vismeteor::mideal] [vismeteor::vmperception]
#' @examples
#' N <- 100
#' psi <- 5.0
#' limmag <- 6.3
#'
#' # Simulate magnitudes
#' m <- rvmideal(N, limmag, psi)
#'
#' # Variance-stabilizing transformation
#' tm <- vmideal_vst_from_magn(m, limmag)
#' tm_mean <- mean(tm)
#' tm_var <- var(tm)
#'
#' # Estimator for psi from the transformed mean
#' psi_hat <- vmideal_vst_to_psi(tm_mean, limmag)
#'
#' # Derivative d(psi)/d(tm) at tm_mean (needed for the delta method)
#' dpsi_dtm <- vmideal_vst_to_psi(tm_mean, limmag, deriv_degree = 1L)
#'
#' # Variance of the sample mean of tm
#' var_tm.mean <- tm_var / N
#'
#' # Delta method: variance and standard error of psi_hat
#' var_psi.hat <- (dpsi_dtm^2) * var_tm.mean
#' se_psi.hat <- sqrt(var_psi.hat)
#'
#' # Results
#' print(psi_hat)
#' print(se_psi.hat)

#' @rdname vmideal_vst
#' @export
vmideal_vst_from_magn <- function(m, lm) {
    offset <- lm - round(lm)
    if (1L == length(lm)) {
        offset <- rep(offset, length(m))
    }

    arg_data <- data.frame(
        x = lm - m,
        offset = offset
    )

    data_f <- as.factor(arg_data$offset)
    data_s <- split(arg_data, data_f)
    y <- lapply(data_s, \(data) {
        x <- data$x
        offset <- data$offset[1]
        params <- .vmideal_vst_from_magn_params
        sy <- c(
            params$p1_fun(offset),
            params$p2_fun(offset),
            params$p3_fun(offset),
            params$p4_fun(offset),
            params$p5_fun(offset),
            params$p6_fun(offset)
        )
        sx <- params$sx
        f_spline <- stats::splinefun(sx, sy)
        y <- rep(NA, length(x))

        idx <- x < sx[1]
        if (any(idx)) {
            y[idx] <- f_spline(sx[1], deriv = 1) * (x[idx] - sx[1]) + sy[1]
        }

        idx <- x > sx[length(sx)]
        if (any(idx)) {
            y[idx] <- f_spline(sx[length(sx)], deriv = 1) * (x[idx] - sx[length(sx)]) + sy[length(sy)]
        }

        idx <- x >= sx[1] & x <= sx[length(sx)]
        if (any(idx)) {
            y[idx] <- f_spline(x[idx])
        }

        y - params$p0_fun(offset)
    })

    unsplit(y, data_f)
}

#' @rdname vmideal_vst
#' @export
vmideal_vst_to_psi <- function(tm, lm, deriv_degree = 0L) {
    poly_coef0 <- c(
        -2.97086442804517, -2.72858043751615, -0.683184284791628, -0.1971973188227,
        -0.0707494993259986, -0.0267813318470729, -0.00482163891332698,
        -1.03296948649241e-05, 6.01391927719714e-05
    )
    names(poly_coef0) <- seq_along(poly_coef0) - 1 # exponents

    # x min 0.016 (psi approx 9 at limiting maginitde of 5.5)
    # x max 8.22(psi approx -10 at limiting maginitde of 6.5)
    tm[tm < 0.016 | tm > 8.22] <- NA
    y <- log(tm)

    if (deriv_degree > 0L) {
        poly_coef1 <- .f_polynomial_coef(poly_coef0, deriv_degree = 1L)
    }
    if (deriv_degree > 1L) {
        poly_coef2 <- .f_polynomial_coef(poly_coef1, deriv_degree = 1L)
    }
    if (deriv_degree > 2L) {
        stop(paste("deriv_degree", deriv_degree, "not implemented!"))
    }

    if (2L == deriv_degree) {
        (.f_polynomial(y, poly_coef2) - .f_polynomial(y, poly_coef1)) / tm^2
    } else if (1L == deriv_degree) {
        .f_polynomial(y, poly_coef1) / tm
    } else {
        lm + .f_polynomial(y, poly_coef0)
    }
}

#' @keywords internal
.vmideal_vst_from_magn_params <- (function() {
    param_df <- data.frame(
        "p1" = c(
            0.379578706193683, 0.380865978506213,
            0.383646581863156, 0.386594955357005, 0.384516871380943, 0.378119364136071,
            0.39783016804395, 0.373393379287508, 0.366393502192895, 0.345593857555874,
            0.367086720870886, 0.359195376100079, 0.347082537633701, 0.312420722892609,
            0.312232925285191
        ), "p2" = c(
            0.758955184180252, 0.757434042449928,
            0.756545241530695, 0.754597324053272, 0.746096530130275, 0.736273688899513,
            0.754472950700777, 0.729839979801878, 0.724090251762525, 0.706272163875503,
            0.732424006749156, 0.730856947686901, 0.722541149425633, 0.690367122987635,
            0.691926213129086
        ), "p3" = c(
            1.89943177726206, 1.89844069948433,
            1.89825430762927, 1.89725244849407, 1.88997593310439, 1.88072563899649,
            1.89899661183851, 1.87396958675163, 1.86749053248022, 1.84882849631795,
            1.87410789172307, 1.87172629242088, 1.86305301012132, 1.83068592572223,
            1.83212684285442
        ), "p4" = c(
            3.27014476551314, 3.26892787903128,
            3.26844009590108, 3.26702893292816, 3.25921320867758, 3.24969559483436,
            3.26789784387066, 3.24296948683872, 3.23673983810267, 3.21843295052407,
            3.24413734533724, 3.24221462073629, 3.23377186059092, 3.20154167652271,
            3.20307252532895
        ), "p5" = c(
            4.50424034207103, 4.50305721713432,
            4.50261471059499, 4.5012654394704, 4.49353145072154, 4.48405231361262,
            4.50225758259143, 4.47729738300568, 4.47099710887246, 4.45259464315401,
            4.47819774376931, 4.4761817895577, 4.46769920746914, 4.43544755919314,
            4.43696568503357
        ), "p6" = c(
            5.65036712145262, 5.64927790140628,
            5.64896064069461, 5.6477824229466, 5.64028041416121, 5.63092781192841,
            5.64918249953455, 5.62420382559607, 5.61781703615347, 5.5992741104315,
            5.62469988859743, 5.62248449312786, 5.61389842883615, 5.58158409117071,
            5.58306055049816
        ), "offset" = c(
            -0.5, -0.48, -0.45, -0.4, -0.3,
            -0.2, -0.1, 0, 0.1, 0.2, 0.3, 0.4, 0.45, 0.48, 0.5
        ), "intercept" = c(
            1.4811934820941,
            1.47927325731333, 1.47783119789222, 1.47503283992891, 1.46516880013637,
            1.45440969975367, 1.47200033556686, 1.44700869631271, 1.44106727057982,
            1.42347451350523, 1.45029075335968, 1.44986985073928, 1.44229503571524,
            1.41062845239295, 1.41258415137388
        )
    )

    list(
        sx = c(1.0, 2.0, 4.0, 6.0, 8.0, 10.0),
        p0_fun = stats::approxfun(param_df$offset, param_df[["intercept"]]),
        p1_fun = stats::approxfun(param_df$offset, param_df[["p1"]]),
        p2_fun = stats::approxfun(param_df$offset, param_df[["p2"]]),
        p3_fun = stats::approxfun(param_df$offset, param_df[["p3"]]),
        p4_fun = stats::approxfun(param_df$offset, param_df[["p4"]]),
        p5_fun = stats::approxfun(param_df$offset, param_df[["p5"]]),
        p6_fun = stats::approxfun(param_df$offset, param_df[["p6"]])
    )
})()

test_that("vmideal_vst", {
    psi <- seq(-3.5, 6.5, 0.5)
    limmag <- c(5.5, 5.52, 5.55, seq(5.6, 6.4, 0.1), 6.45, 6.48, 6.5)
    m <- seq(-200, 6, 1)

    model <- expand.grid(psi = psi, limmag = limmag)
    model <- do.call(
        rbind.data.frame,
        mapply(function(psi, limmag) {
            p <- dvmideal(m, limmag, psi)
            t <- vmideal_vst_from_magn(m, limmag)
            t.mean <- sum(p * t)
            t.var <- sum(p * (t - t.mean)^2)

            list(
                psi = psi,
                limmag = limmag,
                t.mean = t.mean,
                t.var = t.var
            )
        }, model$psi, model$limmag, SIMPLIFY = FALSE)
    )

    #
    # test vmideal_vst_from_magn
    #
    expect_true(all(abs(model$t.var - 1.0) < 0.0017))

    #
    # test vmideal_vst_to_psi
    #
    model$psi.est <- vmideal_vst_to_psi(model$t.mean, model$limmag)
    expect_true(mean(abs(model$psi - model$psi.est)) < 0.0072)
    expect_true(all(abs(model$psi - model$psi.est) < 0.03))

    # test with low psi values
    model0 <- with(model, {
        subset(model, psi < 5.0)
    })
    expect_true(mean(abs(model0$psi - model0$psi.est)) < 0.006)
    expect_true(all(abs(model0$psi - model0$psi.est) < 0.023))

    # test first derivative
    f <- function(x) {
        vmideal_vst_to_psi(x, 6.0, deriv_degree = 1L)
    }
    y <- vmideal_vst_to_psi(5.5, 6.0) - vmideal_vst_to_psi(4.5, 6.0)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)

    # test second derivative
    f <- function(x) {
        vmideal_vst_to_psi(x, 6.0, deriv_degree = 2L)
    }
    y <- vmideal_vst_to_psi(5.5, 6.0, deriv_degree = 1L) - vmideal_vst_to_psi(4.5, 6.0, deriv_degree = 1L)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)
})

#' @name PER_2015_rates
#' @docType data
#' @keywords data
#' @title Visual rate observations of Perseids from 2015
#' @description
#' Visual rate observations of the Perseid shower from 2015.
#' @format A list with the same structure as returned by [vismeteor::load_vmdb_rates].
#' @details
#' `PER_2015_rates` are rate observations loaded with [vismeteor::load_vmdb_rates].
#' @seealso [vismeteor::load_vmdb]
NULL

#' @name PER_2015_magn
#' @docType data
#' @keywords data
#' @title Visual magnitude observations of Perseids from 2015
#' @description
#' Visual magnitude observations of the Perseid shower from 2015.
#' @format A list with the same structure as returned by [vismeteor::load_vmdb_magnitudes].
#' @details
#' `PER_2015_magn` are magnitude observations loaded with
#' [vismeteor::load_vmdb_magnitudes].
#' @seealso [vismeteor::load_vmdb]
NULL

#' @name vmideal
#' @aliases dvmideal
#' @aliases pvmideal
#' @aliases qvmideal
#' @aliases rvmideal
#' @aliases cvmideal
#' @title Ideal Distribution of Visual Meteor Magnitudes
#' @description
#' Density, distribution function, quantile function, and random generation
#' for the ideal distribution of visual meteor magnitudes.
#' @param psi numeric; the location parameter of the probability distribution.
#' @param m integer; visual meteor magnitude.
#' @param lm numeric; limiting magnitude.
#' @param p numeric; probability.
#' @param n numeric; count of meteor magnitudes.
#' @param log logical; if `TRUE`, probabilities are returned as `log(p)`.
#' @param lower.tail logical; if `TRUE` (default), probabilities are
#'     \eqn{P[M < m]}; otherwise, \eqn{P[M \ge m]}.
#' @param perception_fun function; optional perception probability function.
#'     The default is [vismeteor::vmperception].
#'
#' @details
#' The density of the [ideal distribution of meteor magnitudes][vismeteor::mideal] is
#' \deqn{
#'     {\displaystyle f(m) = \frac{\mathrm{d}p}{\mathrm{d}m} =
#'      \frac{3}{2} \, \log(r) \sqrt{\frac{r^{3 \, \psi + 2 \, m}}{(r^\psi + r^m)^5}}}
#' }
#' where \eqn{m} denotes the continuous (real-valued) meteor magnitude,
#' \eqn{r = 10^{0.4} \approx 2.51189 \dots} is a constant, and
#' \eqn{\psi} is the only parameter of this magnitude distribution.
#'
#' In visual meteor observations, magnitudes are usually estimated as integer values.
#' Hence, this distribution is discrete and its probability mass function is given by
#' \deqn{
#' P[M = m] \sim
#' \begin{cases}
#'   g(m_{\mathrm{lim}} - m) \displaystyle \int\limits_{m-0.5}^{m+0.5} f(u) \, \mathrm{d}u,
#'   & \text{if } m_{\mathrm{lim}} - m > -0.5,\\[5pt]
#'   0 & \text{otherwise,}
#' \end{cases}
#' }
#' where \eqn{m_{\mathrm{lim}}} denotes the limiting (non-integer) magnitude of the observation,
#' and \eqn{m} the integer meteor magnitude.
#' The function \eqn{f(\cdot)} is the continuous density of the ideal magnitude distribution,
#' and \eqn{g(\cdot)} denotes the [perception probability function][vismeteor::vmperception].
#'
#' If a different perception probability function `perception_fun` is supplied,
#' it must have the signature `function(x)` and return the perception probabilities of
#' the difference `x` between the limiting magnitude and the meteor magnitude.
#' If `x >= 15.0`, the `perception_fun` function should return a perception probability of `1.0`.
#' The argument `perception_fun` is resolved using [match.fun].
#'
#' @return
#' - `dvmideal`: density
#' - `pvmideal`: distribution function
#' - `qvmideal`: quantile function
#' - `rvmideal`: random generation
#' - `cvmideal`: partial convolution of the ideal distribution of meteor magnitudes
#'   with the perception probabilities.
#'
#' The length of the result is determined by `n` for `rvmideal`, and is the maximum
#' of the lengths of the numeric vector arguments for the other functions.
#' All arguments are vectorized; standard R recycling rules apply.
#'
#' Since the distribution is discrete, `qvmideal` and `rvmideal` always return integer values.
#' `qvmideal` may return `NaN` with a warning.
#' @seealso [vismeteor::mideal]
#'   [vismeteor::vmperception]
#'
#' @references Richter, J. (2018) \emph{About the mass and magnitude distributions of meteor showers}.
#'   WGN, Journal of the International Meteor Organization, vol. 46, no. 1, p. 34-38
#' @examples
#' N <- 100
#' psi <- 5.0
#' limmag <- 6.5
#' (m <- seq(6, -4))
#'
#' # discrete density of `N` meteor magnitudes
#' (freq <- round(N * dvmideal(m, limmag, psi)))
#'
#' # log likelihood function
#' lld <- function(psi) {
#'     -sum(freq * dvmideal(m, limmag, psi, log = TRUE))
#' }
#'
#' # maximum likelihood estimation (MLE) of psi
#' est <- optim(2, lld, method = "Brent", lower = 0, upper = 8, hessian = TRUE)
#'
#' # estimations
#' est$par # mean of psi
#'
#' # generate random meteor magnitudes
#' m <- rvmideal(N, limmag, psi)
#'
#' # log likelihood function
#' llr <- function(psi) {
#'     -sum(dvmideal(m, limmag, psi, log = TRUE))
#' }
#'
#' # maximum likelihood estimation (MLE) of psi
#' est <- optim(2, llr, method = "Brent", lower = 0, upper = 8, hessian = TRUE)
#'
#' # estimations
#' est$par # mean of psi
#' sqrt(1 / est$hessian[1][1]) # standard deviation of psi
#'
#' m <- seq(6, -4, -1)
#' p <- vismeteor::dvmideal(m, limmag, psi)
#' barplot(
#'     p,
#'     names.arg = m,
#'     main = paste0("Density (psi = ", psi, ", limmag = ", limmag, ")"),
#'     col = "blue",
#'     xlab = "m",
#'     ylab = "p",
#'     border = "blue",
#'     space = 0.5
#' )
#' axis(side = 2, at = pretty(p))
#'
#' plot(
#'     \(lm) vismeteor::cvmideal(lm, psi, log = TRUE),
#'     -5, 10,
#'     main = paste0(
#'         "Partial convolution of the ideal meteor magnitude distribution\n",
#'         "with the perception probabilities (psi = ", psi, ")"
#'     ),
#'     col = "blue",
#'     xlab = "lm",
#'     ylab = "log(rate)"
#' )

#' @rdname vmideal
#' @export
dvmideal <- function(m, lm, psi, log = FALSE, perception_fun = vmperception) {
    if (anyNA(m) || anyNA(lm) || anyNA(psi)) {
        stop("NA's are not allowed!")
    }

    if (any(is.infinite(lm))) {
        stop("Infinite limiting magnitudes are not allowed!")
    }

    if (any(is.infinite(psi))) {
        stop("Infinite psi values are not allowed!")
    }

    is_wholenumber <- function(x, tol = .Machine$double.eps^0.5) all(is.infinite(x) | abs(x - round(x)) < tol)
    if (!is_wholenumber(m)) {
        stop("magnitudes must be integer values!")
    }

    perception_fun <- match.fun(perception_fun)

    if (1 == length(lm)) {
        lm <- rep(lm, length(m))
    }

    if (1 == length(psi)) {
        psi <- rep(psi, length(m))
    }

    # density function
    f_density <- function(m, lm, psi, log) {
        psi_exp <- 10.0
        if (lm + psi_exp < psi) {
            return(vismeteor::dvmgeom(m, lm, 10^0.4, log = log))
        }

        norm_res <- .vmideal_norm(lm, psi, perception_fun)
        p <- norm_res$p
        m_lower <- norm_res$m_lower
        m_upper <- norm_res$m_upper

        if (log) {
            d <- rep(-Inf, length(m))
        } else {
            d <- rep(0.0, length(m))
        }

        idx <- m >= m_lower & m <= m_upper
        if (any(idx)) {
            d_tmp <- p[as.character(m[idx])]

            if (log) {
                d_tmp[0.0 == d_tmp] <- -Inf

                log_idx <- -Inf != d_tmp
                if (any(log_idx)) {
                    d_tmp[log_idx] <- base::log(d_tmp[log_idx])
                }
            }

            d[idx] <- d_tmp
        }

        idx <- m > -Inf & m < m_lower
        if (any(idx)) {
            if (log) {
                d[idx] <- .dmideal_int(m[idx], psi, log = TRUE) - base::log(norm_res$norm)
            } else {
                d[idx] <- .dmideal_int(m[idx], psi) / norm_res$norm
            }
        }

        d
    }

    arg_data <- data.frame(
        m = m,
        lm = lm,
        psi = psi
    )

    data_f <- as.factor(paste0(lm, "/", psi))
    data_s <- split(arg_data, data_f)
    d <- lapply(data_s, \(data) {
        m <- data$m
        lm <- data$lm[1]
        psi <- data$psi[1]
        f_density(m, lm, psi, log = log)
    })

    unsplit(d, data_f)
}

#' @rdname vmideal
#' @export
pvmideal <- function(m, lm, psi, lower.tail = TRUE, log = FALSE, perception_fun = vmperception) {
    if (anyNA(m) || anyNA(lm) || anyNA(psi)) {
        stop("NA's are not allowed!")
    }

    if (any(is.infinite(lm))) {
        stop("Infinite limiting magnitudes are not allowed!")
    }

    if (any(is.infinite(psi))) {
        stop("Infinite psi values are not allowed!")
    }

    is_wholenumber <- function(x, tol = .Machine$double.eps^0.5) all(is.infinite(x) | abs(x - round(x)) < tol)
    if (!is_wholenumber(m)) {
        stop("magnitudes must be integer values!")
    }

    perception_fun <- match.fun(perception_fun)

    if (1 == length(lm)) {
        lm <- rep(lm, length(m))
    }

    if (1 == length(psi)) {
        psi <- rep(psi, length(m))
    }

    # probability function
    f_prob <- function(m, lm, psi, log) {
        psi_exp <- 10.0
        if (lm + psi_exp < psi) {
            return(vismeteor::pvmgeom(m, lm, 10^0.4, lower.tail = lower.tail, log = log))
        }

        norm_res <- .vmideal_norm(lm, psi, perception_fun)
        m_lower <- norm_res$m_lower
        m_upper <- norm_res$m_upper

        if (lower.tail) {
            if (log) {
                p <- rep(0.0, length(m))
                p[-Inf == m] <- -Inf
            } else {
                p <- rep(1.0, length(m))
                p[-Inf == m] <- 0.0
            }

            idx <- m > -Inf & m <= m_lower
            if (any(idx)) {
                if (log) {
                    p[idx] <- vismeteor::pmideal(m[idx] - 0.5, psi, lower.tail = TRUE, log = TRUE) -
                        base::log(norm_res$norm)
                } else {
                    p[idx] <- vismeteor::pmideal(m[idx] - 0.5, psi, lower.tail = TRUE, log = FALSE) / norm_res$norm
                }
            }

            idx <- m > m_lower & m <= m_upper
            if (any(idx)) {
                p_gen <- cumsum(norm_res$p)
                p[idx] <- norm_res$p_lower_tail + p_gen[as.character(m[idx] - 1)]
                p[p > 1.0] <- 1.0
                if (log) {
                    log_idx <- p > 0.0
                    if (any(log_idx)) {
                        p[log_idx] <- base::log(p[log_idx])
                    }
                }
            }
        } else {
            p <- rep(0.0, length(m))
            p[-Inf == m] <- 1.0

            idx <- m > -Inf & m < m_lower
            if (any(idx)) {
                p[idx] <- 1.0 - vismeteor::pmideal(m[idx] - 0.5, psi, lower.tail = TRUE) / norm_res$norm
            }

            idx <- m >= m_lower & m <= m_upper
            if (any(idx)) {
                p_gen <- base::rev(cumsum(base::rev(norm_res$p)))
                p[idx] <- p_gen[as.character(m[idx])]
            }

            p[m > 0.0 & is.infinite(m)] <- 0.0
            p[p > 1.0] <- 1.0

            if (log) {
                log_idx <- p > 0.0
                if (any(log_idx)) {
                    p[log_idx] <- base::log(p[log_idx])
                }
                p[!log_idx] <- -Inf
            }
        }

        p
    }

    arg_data <- data.frame(
        m = m,
        lm = lm,
        psi = psi
    )

    data_f <- as.factor(paste0(lm, "/", psi))
    data_s <- split(arg_data, data_f)
    p <- lapply(data_s, \(data) {
        m <- data$m
        lm <- data$lm[1]
        psi <- data$psi[1]
        f_prob(m, lm, psi, log = log)
    })

    unsplit(p, data_f)
}

#' @rdname vmideal
#' @export
qvmideal <- function(p, lm, psi, lower.tail = TRUE, perception_fun = vmperception) {
    if (anyNA(p) || anyNA(lm) || anyNA(psi)) {
        stop("NA's are not allowed!")
    }

    if (any(is.infinite(lm))) {
        stop("Infinite limiting magnitudes are not allowed!")
    }

    if (any(is.infinite(psi))) {
        stop("Infinite psi values are not allowed!")
    }

    perception_fun <- match.fun(perception_fun)

    if (1 == length(lm)) {
        lm <- rep(lm, length(p))
    }

    if (1 == length(psi)) {
        psi <- rep(psi, length(p))
    }

    # quantile function
    f_q <- function(p, lm, psi) {
        m_max <- 15L
        psi_exp <- 10.0

        if (lm + psi_exp < psi) {
            r_lower <- 10^0.4
            return(vismeteor::qvmgeom(p, lm, r_lower, lower.tail = lower.tail))
        }

        m_upper <- .vmideal_upper_lm(lm)
        m_lower <- m_upper - m_max
        m <- rep(NA, length(p))

        if (lower.tail) {
            m[0.0 == p] <- -Inf
            m[1.0 == p] <- m_upper
            p_max <- vismeteor::pvmideal(m_lower, lm, psi, lower.tail = TRUE, perception_fun = perception_fun)

            idx <- p > 0.0 & p < p_max
            if (any(idx)) {
                p_max_ideal <- vismeteor::pmideal(m_lower - 0.5, psi, lower.tail = TRUE)
                m[idx] <- floor(0.5 + vismeteor::qmideal(
                    p_max_ideal * p[idx] / p_max,
                    psi,
                    lower.tail = TRUE
                ))
            }
            idx <- p >= p_max & p < 1.0
            if (any(idx)) {
                m_gen <- seq(m_lower, m_upper + 1)
                p_gen <- vismeteor::pvmideal(m_gen, lm, psi, lower.tail = lower.tail, perception_fun = perception_fun)
                p_idx <- findInterval(p[idx], p_gen, left.open = FALSE)
                m[idx] <- m_gen[p_idx]
            }
        } else {
            m[0.0 == p] <- m_upper
            m[1.0 == p] <- -Inf
            p_max <- vismeteor::pvmideal(m_lower, lm, psi, lower.tail = FALSE, perception_fun = perception_fun)

            idx <- p > p_max & p < 1.0
            if (any(idx)) {
                p_max_ideal <- vismeteor::pmideal(m_lower - 0.5, psi, lower.tail = FALSE)
                m[idx] <- floor(0.5 + vismeteor::qmideal(
                    1.0 - (1.0 - p[idx]) * ((1.0 - p_max_ideal) / (1.0 - p_max)),
                    psi,
                    lower.tail = FALSE
                ))
            }

            idx <- p > 0.0 & p <= p_max
            if (any(idx)) {
                m_gen <- seq(m_upper + 1, m_lower)
                p_gen <- vismeteor::pvmideal(m_gen, lm, psi, lower.tail = FALSE, perception_fun = perception_fun)
                p_idx <- findInterval(p[idx], p_gen, left.open = TRUE) + 1
                m[idx] <- m_gen[p_idx]
            }
        }

        m
    }

    arg_data <- data.frame(
        p = p,
        lm = lm,
        psi = psi
    )

    data_f <- as.factor(paste0(lm, "/", psi))
    data_s <- split(arg_data, data_f)
    m <- lapply(data_s, \(data) {
        p <- data$p
        lm <- data$lm[1]
        psi <- data$psi[1]
        f_q(p, lm, psi)
    })

    m <- unsplit(m, data_f)

    if (anyNA(m)) {
        warning("NaNs produced")
    }

    m
}

#' @rdname vmideal
#' @export
rvmideal <- function(n, lm, psi, perception_fun = vmperception) {
    if (anyNA(lm) || anyNA(psi)) {
        stop("NA's are not allowed!")
    }

    if (any(is.infinite(lm))) {
        stop("Infinite limiting magnitudes are not allowed!")
    }

    if (any(is.infinite(psi))) {
        stop("Infinite psi values are not allowed!")
    }

    perception_fun <- match.fun(perception_fun)

    p <- stats::runif(n)
    m <- rep(NA, n)

    idx <- p < 0.5
    if (any(idx)) {
        m[idx] <- vismeteor::qvmideal(p[idx], lm, psi, lower.tail = TRUE, perception_fun = perception_fun)
    }

    if (any(!idx)) {
        m[!idx] <- vismeteor::qvmideal(1.0 - p[!idx], lm, psi, lower.tail = FALSE, perception_fun = perception_fun)
    }

    m
}

#' @rdname vmideal
#' @export
cvmideal <- function(lm, psi, log = FALSE, perception_fun = vmperception) {
    if (anyNA(lm) || anyNA(psi)) {
        stop("NA's are not allowed!")
    }

    perception_fun <- match.fun(perception_fun)

    if (1 == length(psi)) {
        psi <- rep(psi, length(lm))
    }

    # Integration - similar to .vmideal_norm()
    f_integrate <- function(lm, psi) {
        m_max <- 15L
        m_upper <- .vmideal_upper_lm(lm)
        m_lower <- m_upper - m_max
        m <- as.integer(seq(m_lower, m_upper))
        p <- .dmideal_int(m, psi) * perception_fun(lm - m)
        p_lower_tail <- vismeteor::pmideal(m_lower - 0.5, psi, lower.tail = TRUE)
        sum(p) + p_lower_tail
    }

    p <- mapply(\(lm, psi) {
        if (Inf == lm & Inf == psi) {
            return(NA)
        }
        if (-Inf == lm & -Inf == psi) {
            return(NA)
        }
        if (Inf == lm & -Inf == psi) {
            return(if (log) 0.0 else 1.0)
        }
        if (Inf == psi & -Inf == lm) {
            return(if (log) -Inf else 0.0)
        }

        if (log) {
            base::log(f_integrate(lm, psi))
        } else {
            f_integrate(lm, psi)
        }
    }, lm, psi, SIMPLIFY = TRUE)

    if (anyNA(p)) {
        warning("NaNs produced")
    }

    p
}

#' upper available magnitude
#'
#' @noRd
.vmideal_upper_lm <- function(lm) {
    lm_round <- round(lm)
    offset <- lm - lm_round
    if (-0.5 == offset) {
        lm_round <- lm_round - 1L
    }

    lm_round
}

#' density of ideal integer magnitude distribution
#'
#' @noRd
.dmideal_int <- function(m, psi, log = FALSE) {
    psi_exp <- 10.0
    r_lower <- 10^0.4

    a <- -base::log(r_lower)
    d <- rep(NA, length(m))

    idx <- m > (psi + psi_exp)
    if (any(idx)) {
        if (log) {
            d[idx] <- base::log(1 - base::exp(1.5 * a)) + a * (1.5 * (m[idx] - psi) - 0.75)
        } else {
            d[idx] <- base::exp(a * (1.5 * (m[idx] - psi) - 0.75)) -
                base::exp(a * (1.5 * (m[idx] - psi) + 0.75))
        }
    }

    idx <- m < (psi - psi_exp)
    if (any(idx)) {
        if (log) {
            d[idx] <- base::log(1.5) + a * (psi - m[idx] - 0.5) + base::log(1 - base::exp(a))
        } else {
            d[idx] <- 1.5 * base::exp(a * (psi - m[idx] - 0.5)) -
                1.5 * base::exp(a * (psi - m[idx] + 0.5))
        }
    }

    idx <- is.na(d)
    if (any(idx)) {
        d[idx] <- vismeteor::pmideal(m[idx] + 0.5, psi, lower.tail = TRUE) -
            vismeteor::pmideal(m[idx] - 0.5, psi, lower.tail = TRUE)

        if (log) {
            d[idx] <- base::log(d[idx])
        }
    }

    d
}

#' normalization
#'
#' @noRd
.vmideal_norm <- function(lm, psi, perception_fun) {
    m_max <- 15L
    m_upper <- .vmideal_upper_lm(lm)
    m_lower <- m_upper - m_max
    m <- as.integer(seq(m_lower, m_upper))
    p <- .dmideal_int(m, psi) * perception_fun(lm - m)
    names(p) <- as.character(m)
    p_lower_tail <- vismeteor::pmideal(m_lower - 0.5, psi, lower.tail = TRUE)
    norm <- sum(p) + p_lower_tail

    list(
        m_lower = m_lower,
        m_upper = m_upper,
        norm = norm,
        p = p / norm,
        p_lower_tail = p_lower_tail / norm
    )
}

test_that("vmidealVst", {
    psi <- seq(-3.5, 6.5, 0.5)
    limmag <- c(5.5, 5.52, 5.55, seq(5.6, 6.4, 0.1), 6.45, 6.48, 6.5)
    m <- seq(-200, 6, 1)

    model <- expand.grid(psi = psi, limmag = limmag)
    model <- do.call(
        rbind.data.frame,
        mapply(function(psi, limmag) {
            p <- dvmideal(m, limmag, psi)
            t <- vmidealVstFromMagn(m, limmag)
            t.mean <- sum(p * t)
            t.var <- sum(p * (t - t.mean)^2)

            list(
                psi = psi,
                limmag = limmag,
                t.mean = t.mean,
                t.var = t.var
            )
        }, model$psi, model$limmag , SIMPLIFY = FALSE)
    )

    #
    # test vmidealVstFromMagn
    #
    expect_true(all(abs(model$t.var - 1.0) < 0.0017))

    #
    # test vmidealVstToPsi
    #
    model$psi.est <- vmidealVstToPsi(model$t.mean, model$limmag)
    expect_true(mean(abs(model$psi - model$psi.est)) < 0.0072)
    expect_true(all(abs(model$psi - model$psi.est) < 0.03))

    # test with low psi values
    model0 <- with(model, {
        subset(model, psi < 5.0)
    })
    expect_true(mean(abs(model0$psi - model0$psi.est)) < 0.006)
    expect_true(all(abs(model0$psi - model0$psi.est) < 0.023))

    # test first derivative
    f <- function(x) {
        vmidealVstToPsi(x, 6.0, deriv.degree = 1L)
    }
    y <- vmidealVstToPsi(5.5, 6.0) - vmidealVstToPsi(4.5, 6.0)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)

    # test second derivative
    f <- function(x) {
        vmidealVstToPsi(x, 6.0, deriv.degree = 2L)
    }
    y <- vmidealVstToPsi(5.5, 6.0, deriv.degree = 1L) - vmidealVstToPsi(4.5, 6.0, deriv.degree = 1L)
    expect_true(abs(y - stats::integrate(f, 4.5, 5.5)$value) < 1e-10)
})

test_that("qvmideal", {
    lm <- 6.0
    psi <- 4.0

    for (current.psi in c(-10.0, psi)) {
        if (psi == current.psi) {
            m <- seq(-10, 6)
        } else {
            m <- current.psi + seq(-10, 10)
        }

        # lower tail
        p <- vismeteor::pvmideal(m, lm, current.psi, lower.tail = FALSE)
        p <- c(-1.0, 0.0, p - 1e-07, p + 1e-07, 1.0)
        q <- suppressWarnings(
            vismeteor::qvmideal(p, lm, current.psi, lower.tail = FALSE)
        )
        expect_equal(q, c(NA, 6, m, m - 1, -Inf), label = paste("psi =", current.psi))

        # upper tail
        p <- vismeteor::pvmideal(m, lm, current.psi, lower.tail = TRUE)
        q <- suppressWarnings(
            vismeteor::qvmideal(c(-1.0, 1.0, p + 1e-07, p - 1e-07, 0.0), lm, current.psi, lower.tail = TRUE)
        )
        expect_equal(q, c(NA, 6, m, m - 1, -Inf), label = paste("psi =", current.psi))
    }

    lms <- seq(5.6, 6.5, 0.1)
    for (current.lm in lms) {
        q <- suppressWarnings(
            vismeteor::qvmideal(c(-1.0, 0.0, 1.0), current.lm, psi, lower.tail = TRUE)
        )
        expect_equal(q, c(NA, -Inf, 6), label = paste("lm =", current.lm))

        q <- suppressWarnings(
            vismeteor::qvmideal(c(-1.0, 0.0, 1.0), current.lm, psi, lower.tail = FALSE)
        )
        expect_equal(q, c(NA, 6, -Inf), label = paste("lm =", current.lm))
    }

    for (current.lm in lms) {
        p <- vismeteor::pvmideal(6, current.lm, psi, lower.tail = TRUE)
        q <- vismeteor::qvmideal(c(p + 1e-07, p - 1e-07), current.lm, psi, lower.tail = TRUE)
        expect_equal(q, c(6, 5), label = paste("lm =", current.lm))

        p <- vismeteor::pvmideal(6, current.lm, psi, lower.tail = FALSE)
        q <- vismeteor::qvmideal(c(p - 1e-07, p + 1e-07), current.lm, psi, lower.tail = FALSE)
        expect_equal(q, c(6, 5), label = paste("lm =", current.lm))
    }

    for (current.lm in lms) {
        p <- vismeteor::pvmideal(5, current.lm, psi, lower.tail = TRUE)
        q <- vismeteor::qvmideal(c(p + 1e-07, p - 1e-07), current.lm, psi, lower.tail = TRUE)
        expect_equal(q, c(5, 4), label = paste("lm =", current.lm))

        p <- vismeteor::pvmideal(5, current.lm, psi, lower.tail = FALSE)
        q <- vismeteor::qvmideal(c(p - 1e-07, p + 1e-07), current.lm, psi, lower.tail = FALSE)
        expect_equal(q, c(5, 4), label = paste("lm =", current.lm))
    }

    # quantile of meteor magnitudes equals geometric distribution
    psi <- 30
    m <- as.integer(seq(-10, 6, 1))
    p <- round(vismeteor::pvmideal(m, lm, psi, lower.tail = TRUE), 6)
    m <- vismeteor::qvmideal(p, lm, psi, lower.tail = TRUE)
    expect_type(m, "double")
    expect_length(m, length(m))
    expect_equal(qvmgeom(p, lm, 10^0.4, lower.tail = TRUE), m)

    m <- as.integer(seq(-10, 6, 1))
    p <- round(vismeteor::pvmideal(m, lm, psi, lower.tail = FALSE), 6)
    m <- vismeteor::qvmideal(p, lm, psi, lower.tail = FALSE)
    expect_type(m, "double")
    expect_length(m, length(m))
    expect_equal(vismeteor::qvmgeom(p, lm, 10^0.4, lower.tail = FALSE), m)
})

test_that("pvmideal", {
    lm <- 6.5
    psi <- 4.0

    # from documentation
    dp_fun <- function(m, psi) {
        r <- 10^0.4
        1.5 * log(r) * sqrt(r^(3 * psi + 2 * m) / ((r^psi + r^m)^5))
    }

    for (current.psi in c(-10.0, psi, 5)) {
        # test probabilities lower tail

        m <- seq(current.psi - 25, 6)
        expected_p <- sapply(m, function(m) {
            stats::integrate(function(m) dp_fun(m, current.psi), m - 0.5, m + 0.5)$value *
                vismeteor::vmperception(lm - m)
        })
        expected_p[1] <- expected_p[1] + vismeteor::pmideal(m[1] - 0.5, current.psi, lower.tail = TRUE)
        expected_p <- cumsum(expected_p) / sum(expected_p) # lower tail

        p <- vismeteor::pvmideal(m + 1, lm, current.psi, lower.tail = TRUE)
        expect_type(p, "double")
        expect_length(p, length(m))
        expect_false(any(abs(log(p) - log(expected_p)) > 0.02), label = paste("psi =", current.psi))

        p <- vismeteor::pvmideal(m + 1, lm, current.psi, lower.tail = TRUE, log = TRUE)
        expect_type(p, "double")
        expect_length(p, length(m))
        expect_false(any(abs(p - log(expected_p)) > 0.02), label = paste("psi =", current.psi))

        # test probabilities upper tail

        m <- seq(6, current.psi - 25, -1)
        expected_p <- sapply(m, function(m) {
            stats::integrate(function(m) dp_fun(m, current.psi), m - 0.5, m + 0.5)$value *
                vismeteor::vmperception(lm - m)
        })
        expected_p <- cumsum(expected_p) / sum(expected_p) # upper tail

        p <- vismeteor::pvmideal(m, lm, current.psi, lower.tail = FALSE)
        expect_type(p, "double")
        expect_length(p, length(m))
        expect_false(any(abs(log(p) - log(expected_p)) > 0.02), label = paste("psi =", current.psi))

        p <- vismeteor::pvmideal(m, lm, current.psi, lower.tail = FALSE, log = TRUE)
        expect_type(p, "double")
        expect_length(p, length(m))
        expect_false(any(abs(p - log(expected_p)) > 0.02), label = paste("psi =", current.psi))

        # test sum is 1.0
        m <- seq(current.psi - 25, 6)
        p <- vismeteor::pvmideal(m, lm, current.psi, lower.tail = TRUE) +
            vismeteor::pvmideal(m, lm, current.psi, lower.tail = FALSE)
        expect_type(p, "double")
        expect_length(p, length(m))
        expect_false(any(abs(1.0 - p) > 1e-06), label = paste("psi =", current.psi))
    }

    # probability of m >= 3
    expected_p <- round(sum(vismeteor::dvmideal(as.integer(seq(3, 6)), 6.3, psi)), 6)
    expect_equal(expected_p, 0.35389)

    # probability
    p <- vismeteor::pvmideal(c(3, 7, -100, -Inf), 6.3, psi, lower.tail = TRUE)
    expect_type(p, "double")
    expect_length(p, 4)
    expect_equal(round(p[1], 6), 1 - expected_p)
    expect_equal(p[2], 1.0)
    expect_equal(p[3], 0.0)
    expect_equal(p[4], 0.0)

    p <- vismeteor::pvmideal(c(3, 7, -100, -Inf), 6.3, psi, lower.tail = FALSE)
    expect_type(p, "double")
    expect_length(p, 4)
    expect_equal(round(p[1], 6), expected_p)
    expect_equal(p[2], 0.0)
    expect_equal(p[3], 1.0)
    expect_equal(p[4], 1.0)

    # log probability
    p <- vismeteor::pvmideal(c(3, 7, -Inf), 6.3, psi, lower.tail = TRUE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 3)
    expect_equal(round(p[1], 4), round(log(1 - expected_p), 4))
    expect_equal(p[2], 0.0)
    expect_equal(p[3], -Inf)

    p <- vismeteor::pvmideal(c(3, 7, -Inf), 6.3, psi, lower.tail = FALSE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 3)
    expect_equal(round(p[1], 4), round(log(expected_p), 4))
    expect_equal(p[2], -Inf)
    expect_equal(p[3], 0.0)

    # test order of probabilities
    lm <- seq(2.5, 6.5, 0.1)
    p <- vismeteor::pvmideal(rep(3L, length(lm)), lm, psi, lower.tail = TRUE)
    expect_equal(p, p[order(p, decreasing = TRUE)])

    lm <- seq(2.5, 6.5, 0.1)
    p <- vismeteor::pvmideal(rep(3L, length(lm)), lm, psi, lower.tail = FALSE)
    expect_equal(p, p[order(p, decreasing = FALSE)])

    # test lower probabilities
    lm <- seq(5.4, 6.4, 0.1)
    p1 <- vismeteor::dvmideal(rep(6L, length(lm)), lm, psi)
    p2 <- vismeteor::pvmideal(rep(6L, length(lm)), lm, psi, lower.tail = FALSE)
    expect_equal(p1, p2)

    # test sum lower probabilities
    lm <- seq(5.4, 6.4, 0.1)
    p1 <- vismeteor::dvmideal(rep(6L, length(lm)), lm, psi) +
        vismeteor::dvmideal(rep(5L, length(lm)), lm, psi)
    p2 <- vismeteor::pvmideal(rep(5L, length(lm)), lm, psi, lower.tail = FALSE)
    expect_equal(p1, p2)

    # density of meteor magnitudes equals geometric distribution
    lm <- 6.3
    m <- as.integer(seq(-20, 6))

    p <- vismeteor::pvmideal(m, lm, 16.25, lower.tail = TRUE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, pvmgeom(m, lm, 10^0.4, lower.tail = TRUE, log = TRUE))

    p <- vismeteor::pvmideal(m, lm, 16.25, lower.tail = FALSE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, pvmgeom(m, lm, 10^0.4, lower.tail = FALSE, log = TRUE))

    p <- vismeteor::pvmideal(m, lm, 16.35, lower.tail = TRUE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, pvmgeom(m, lm, 10^0.4, lower.tail = TRUE, log = TRUE))

    p <- vismeteor::pvmideal(m, lm, 16.35, lower.tail = FALSE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, length(m))
    expect_equal(p, pvmgeom(m, lm, 10^0.4, lower.tail = FALSE, log = TRUE))
})

test_that("qmideal", {
    psi <- 4.0

    p <- vismeteor::pmideal(6, psi, lower.tail = FALSE)
    m <- suppressWarnings(
        vismeteor::qmideal(c(-0.1, 0.0, p, 1.0, 1.2), psi, lower.tail = FALSE)
    )
    expect_length(m, 5)
    expect_equal(m, c(NA, Inf, 6, -Inf, NA))

    p <- vismeteor::pmideal(6, psi, lower.tail = TRUE)
    m <- suppressWarnings(
        vismeteor::qmideal(c(-0.1, 0.0, p, 1.0, 1.2), psi, lower.tail = TRUE)
    )
    expect_length(m, 5)
    expect_equal(m, c(NA, -Inf, 6, Inf, NA))

    # quantile of meteor magnitudes equals exponential distribution

    # from documentation
    dp_fun <- function(m, psi) {
        r <- 10^0.4
        1.5 * log(r) * sqrt(r^(3 * psi + 2 * m) / ((r^psi + r^m)^5))
    }

    m_expected <- seq(-20, -10, 1)
    p <- vismeteor::pmideal(m_expected, psi, lower.tail = TRUE)
    m <- vismeteor::qmideal(p, psi, lower.tail = TRUE)
    expect_equal(m, m_expected)
    p_max <- stats::integrate(function(m) dp_fun(m, psi), -Inf, -9)$value
    m <- -9 - stats::qexp(p / p_max, 0.4 * log(10), lower.tail = FALSE)
    expect_equal(round(m, 3), m_expected)

    m_expected <- seq(15, 20, 1)
    p <- vismeteor::pmideal(m_expected, psi, lower.tail = FALSE)
    m <- vismeteor::qmideal(p, psi, lower.tail = FALSE)
    expect_equal(m, m_expected)
    p_max <- stats::integrate(function(m) dp_fun(m, psi), 14, Inf)$value
    m <- 14 + stats::qexp(p / p_max, 1.5 * 0.4 * log(10), lower.tail = FALSE)
    expect_equal(round(m, 3), m_expected)
})

test_that("pmideal", {
    psi <- 4.0

    # from documentation
    dp_fun <- function(m, psi) {
        r <- 10^0.4
        1.5 * log(r) * sqrt(r^(3 * psi + 2 * m) / ((r^psi + r^m)^5))
    }

    # probability of m >= 3
    expected_p <- stats::integrate(function(m) dp_fun(m, psi), -Inf, 3)$value
    expect_equal(vismeteor::pmideal(3, psi, lower.tail = TRUE), expected_p)

    # probability
    p <- vismeteor::pmideal(c(3, -Inf, -100, 100, Inf), psi, lower.tail = TRUE)
    expect_type(p, "double")
    expect_length(p, 5)
    expect_equal(p[1], expected_p)
    expect_equal(p[2], 0.0)
    expect_equal(p[3], 0.0)
    expect_equal(p[4], 1.0)
    expect_equal(p[5], 1.0)

    p <- vismeteor::pmideal(c(3, -Inf, -100, 100, Inf), psi, lower.tail = FALSE)
    expect_type(p, "double")
    expect_length(p, 5)
    expect_equal(round(p[1], 6), round(1.0 - expected_p, 6))
    expect_equal(p[2], 1.0)
    expect_equal(p[3], 1.0)
    expect_equal(p[4], 0.0)
    expect_equal(p[5], 0.0)

    # log probability
    p <- vismeteor::pmideal(c(3, -Inf, -100, 100, Inf), psi, lower.tail = TRUE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 5)
    expect_equal(round(p[1], 6), round(log(expected_p), 6))
    expect_equal(p[2], -Inf)
    expect_equal(round(p[3], 6), -95.382075)
    expect_equal(p[4], 0)
    expect_equal(p[5], 0)

    p <- vismeteor::pmideal(c(3, -Inf, -100, 100, Inf), psi, lower.tail = FALSE, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 5)
    expect_equal(round(p[1], 6), round(log(1.0 - expected_p), 6))
    expect_equal(p[2], 0.0)
    expect_equal(p[3], 0.0)
    expect_equal(round(p[4], 6), -132.628899)
    expect_equal(p[5], -Inf)

    # density of meteor magnitudes equals exponential distribution
    m <- seq(-20, -10, 1)
    p <- vismeteor::pmideal(m, psi, lower.tail = TRUE)
    p_max <- stats::integrate(function(m) dp_fun(m, psi), -Inf, -10)$value
    expect_equal(
        round(log(p), 3),
        round(log(p_max * stats::pexp(-m - 10, 0.4 * log(10), lower.tail = FALSE)), 3)
    )

    p <- vismeteor::pmideal(m, psi, lower.tail = TRUE, log = TRUE)
    expect_equal(
        round(p, 3),
        round(log(p_max) + stats::pexp(-m - 10, 0.4 * log(10), lower.tail = FALSE, log = TRUE), 3)
    )

    m <- seq(15, 20, 1)
    p <- vismeteor::pmideal(m, psi, lower.tail = FALSE)
    p_max <- stats::integrate(function(m) dp_fun(m, psi), 15, Inf)$value
    expect_equal(
        round(log(p), 3),
        round(log(p_max * stats::pexp(m - 15, 1.5 * 0.4 * log(10), lower.tail = FALSE)), 3)
    )

    p <- vismeteor::pmideal(m, psi, lower.tail = FALSE, log = TRUE)
    expect_equal(
        round(p, 3),
        round(log(p_max) + stats::pexp(m - 15, 1.5 * 0.4 * log(10), lower.tail = FALSE, log = TRUE), 3)
    )
})

test_that("dmideal", {
    psi <- 4.0

    # from documentation
    dp_fun <- function(m, psi) {
        r <- 10^0.4
        1.5 * log(r) * sqrt(r^(3 * psi + 2 * m) / ((r^psi + r^m)^5))
    }

    expected_p <- dp_fun(3, psi)

    # density of meteor magnitudes
    p <- vismeteor::dmideal(c(-Inf, Inf, 3), psi)
    expect_type(p, "double")
    expect_length(p, 3)
    expect_equal(p[1], 0.0)
    expect_equal(p[2], 0.0)
    expect_equal(p[3], expected_p)

    # density of meteor magnitudes (shifting property)
    p <- vismeteor::dmideal(c(2, 3, 4), c(psi - 1, psi, psi + 1))
    expect_type(p, "double")
    expect_length(p, 3)
    expect_equal(p[1], expected_p)
    expect_equal(p[2], expected_p)
    expect_equal(p[3], expected_p)

    # log density of meteor magnitudes
    p <- vismeteor::dmideal(c(-Inf, Inf, 3), psi, log = TRUE)
    expect_type(p, "double")
    expect_length(p, 3)
    expect_equal(p[1], -Inf)
    expect_equal(p[2], -Inf)
    expect_equal(p[3], log(expected_p))

    # test maximum likelihood estimation (MLE) of psi
    m <- seq(-20, 20, 0.001)
    p <- vismeteor::dmideal(m, psi)
    ll <- function(psi) { # log likelihood function
        -sum(p * vismeteor::dmideal(m, psi, log = TRUE))
    }
    est <- optim(2, ll, method = "Brent", lower = 0, upper = 7, hessian = TRUE)
    expect_equal(round(est$par, 6), psi)
})

test_that("select_knots", {
    # synthetic, deterministic signal
    set.seed(1)
    n <- 200
    x <- seq(0, 10, length.out = n)
    y <- sin(x) + stats::rnorm(n, sd = 0.2)
    dat <- data.frame(x = x, y = y)

    fit <- \(d, knots) {
        f <- if (length(knots) == 0L) {
            y ~ splines::ns(x)
        } else {
            y ~ splines::ns(x, knots = knots)
        }
        stats::lm(f, data = d)
    }
    score_bic <- \(d, knots) stats::BIC(fit(d, knots))
    cand <- seq(1, 9, by = 0.5)

    # forward selection: result has the expected shape
    res <- select_knots(dat, cand, score_bic)
    expect_named(res, c(
        "backward", "knots", "fixed_knots",
        "score", "n_steps", "history"
    ))
    expect_false(res$backward)
    expect_true(is.numeric(res$knots))
    expect_true(is.numeric(res$fixed_knots))
    expect_equal(res$knots, sort(res$knots))
    expect_equal(res$fixed_knots, sort(res$fixed_knots))
    expect_true(all(res$knots %in% cand))
    expect_length(intersect(res$knots, res$fixed_knots), 0L)
    expect_null(res$n_steps)
    expect_equal(res$score, tail(res$history$score, 1L))
    expect_true(is.data.frame(res$history))
    expect_named(
        res$history,
        c("step", "n_knots", "changed_knot", "score", "extra")
    )
    # NULL-mode never records a worsening step
    expect_false(any(res$history$extra))

    # n_steps = 2L: exactly 2 total iterations (NOT 2 past optimum). In
    # forward mode without bulk that is 2 added knots and a 3-row history.
    res2 <- select_knots(dat, cand, score_bic, n_steps = 2L)
    expect_identical(res2$n_steps, 2L)
    expect_length(res2$knots, 2L)
    expect_equal(nrow(res2$history), 3L)

    # A large n_steps lets the search run through the optimum and onward;
    # the best point along that trajectory is at least as good as the
    # NULL-mode optimum (both follow the same path up to it).
    res_long <- select_knots(dat, cand, score_bic,
        n_steps = length(cand) + 5L
    )
    expect_lte(min(res_long$history$score), res$score)

    # Single-step exploration past the optimum: take one controlled step
    # from the previous optimum and check that exactly one new knot appears.
    anchor <- c(res$knots, res$fixed_knots)
    res_step1 <- select_knots(dat, cand, score_bic,
        fixed_knots = anchor, n_steps = 1L
    )
    expect_length(res_step1$knots, 1L)
    expect_false(res_step1$knots %in% anchor)
    expect_identical(sort(res_step1$fixed_knots), sort(anchor))

    # backward elimination produces a valid run. Greedy stepwise selection is
    # path-dependent (see the "When to use this" section of ?select_knots),
    # so backward and forward need not reach the same knot set; only the
    # structural invariants must hold.
    res_b <- select_knots(dat, cand, score_bic, backward = TRUE)
    expect_true(res_b$backward)
    expect_true(all(res_b$knots %in% cand))
    expect_equal(res_b$knots, sort(res_b$knots))
    expect_lt(res_b$score, score_bic(dat, numeric(0)))

    # bulk_gap = 0L disables bulk; it is a different trajectory than the
    # default bulk run, so optima can differ, but both must be valid.
    res_b0 <- select_knots(dat, cand, score_bic,
        backward = TRUE, bulk_gap = 0L
    )
    expect_true(all(res_b0$knots %in% cand))
    expect_lt(res_b0$score, score_bic(dat, numeric(0)))

    # score_fun that errors out is treated as Inf and the search still runs
    score_err <- \(d, knots) {
        if (length(knots) > 2L) stop("boom")
        score_bic(d, knots)
    }
    res_err <- select_knots(dat, cand, score_err)
    expect_true(is.finite(res_err$score))
    expect_lte(length(res_err$knots), 2L)

    # fixed_knots, forward: echoed back in $fixed_knots, never in $knots.
    res_ff <- select_knots(dat, cand, score_bic, fixed_knots = 5.25)
    expect_true(5.25 %in% res_ff$fixed_knots)
    expect_false(5.25 %in% res_ff$knots)
    expect_length(intersect(res_ff$knots, res_ff$fixed_knots), 0L)
    expect_equal(res_ff$knots, sort(res_ff$knots))

    # fixed_knots, backward: never removed -> still in $fixed_knots, not in $knots
    res_fb <- select_knots(dat, cand, score_bic,
        backward = TRUE, fixed_knots = 5.25
    )
    expect_true(5.25 %in% res_fb$fixed_knots)
    expect_false(5.25 %in% res_fb$knots)
    expect_equal(res_fb$knots, sort(res_fb$knots))

    # fixed_knots = cand in backward mode: nothing is removable, the search
    # terminates immediately. $knots is empty; $fixed_knots holds the whole pool.
    res_fall <- select_knots(dat, cand, score_bic,
        backward = TRUE, fixed_knots = cand
    )
    expect_length(res_fall$knots, 0L)
    expect_equal(sort(c(res_fall$knots, res_fall$fixed_knots)), sort(cand))

    # Edge case: forward starting from a fully-saturated fixed_knots set
    # (overfit start, no improving move possible). NULL-mode must terminate
    # at the starting state without recording any worsening step.
    res_sat <- select_knots(dat, cand, score_bic, fixed_knots = cand)
    expect_length(res_sat$knots, 0L)
    expect_equal(nrow(res_sat$history), 1L)
    expect_false(any(res_sat$history$extra))

    # Same start, but explicit n_steps = 2L: two worsening steps are taken.
    # (No candidates remain to add in forward mode -- the loop exits via the
    # `length(remaining) == 0` guard at the top. Test that branch instead by
    # using a set that overfits but still has candidates to spare.)
    overfit_anchor <- cand[seq(1L, length(cand) - 2L)]
    res_over_null <- select_knots(dat, cand, score_bic,
        fixed_knots = overfit_anchor
    )
    res_over_2 <- select_knots(dat, cand, score_bic,
        fixed_knots = overfit_anchor, n_steps = 2L
    )
    # With n_steps = 2L the loop runs both iterations regardless of improvement.
    expect_equal(nrow(res_over_2$history), 3L)
    # If the NULL run stopped at the start (no improvement possible), then the
    # explicit n_steps run necessarily explores worsening territory.
    if (nrow(res_over_null$history) == 1L) {
        expect_true(any(res_over_2$history$extra))
    }

    # score_fun receives sorted knots (no defensive sort needed in the user fit)
    seen_unsorted <- FALSE
    score_probe <- \(d, knots) {
        if (length(knots) >= 2L && is.unsorted(knots)) seen_unsorted <<- TRUE
        score_bic(d, knots)
    }
    select_knots(dat, cand, score_probe)
    expect_false(seen_unsorted)
    select_knots(dat, cand, score_probe, backward = TRUE)
    expect_false(seen_unsorted)

    # Names on knot_candidates and fixed_knots are preserved through the
    # returned $knots and $fixed_knots.
    cand_named <- setNames(cand, paste0("k", seq_along(cand)))
    res_named <- select_knots(dat, cand_named, score_bic)
    expect_false(is.null(names(res_named$knots)))
    expect_true(all(names(res_named$knots) %in% names(cand_named)))

    fk_named <- c(anchor = 5.25)
    res_fnamed <- select_knots(dat, cand_named, score_bic,
        fixed_knots = fk_named
    )
    expect_true("anchor" %in% names(res_fnamed$fixed_knots))
    expect_equal(
        unname(res_fnamed$fixed_knots[
            names(res_fnamed$fixed_knots) == "anchor"
        ]),
        5.25
    )

    # input validation
    expect_error(select_knots(dat, cand, score_bic, backward = NA))
    expect_error(select_knots(dat, cand, score_bic, n_steps = -1L))
    expect_error(select_knots(dat, cand, score_bic, n_steps = 0L))
    expect_error(select_knots(dat, cand, score_bic, n_steps = 1.5))
    expect_error(select_knots(dat, cand, score_bic, n_steps = NA_integer_))
    expect_error(select_knots(dat, cand, score_bic, n_cores = 0L))
    expect_error(select_knots(dat, cand, score_fun = "not a function"))
    expect_error(select_knots(dat, cand, score_bic, fixed_knots = c(1, NA)))
})

test_that("rmideal", {
    psi <- 4.0

    with_seed <- function(seed, code) {
        code <- substitute(code)
        orig_seed <- .Random.seed
        on.exit(.Random.seed <<- orig_seed)
        set.seed(seed)
        eval.parent(code)
    }

    m <- vismeteor::rmideal(1000, psi)
    expect_type(m, "double")
    expect_length(m, 1000)
    expect_false(anyNA(m))
    expect_false(any(is.infinite(m)))

    # MLE of psi
    m <- with_seed(8, vismeteor::rmideal(100000, psi))
    expect_type(m, "double")
    expect_length(m, 100000)
    expect_false(anyNA(m))
    expect_false(any(is.infinite(m)))
    llr <- function(psi) {
        -sum(vismeteor::dmideal(m, psi, log = TRUE))
    }
    est <- optim(2, llr, method = "Brent", lower = 0, upper = 8, hessian = TRUE)
    expect_equal(round(est$par, 2), psi)
})

---
title: "Parameter Estimation of the Geometric Model of Visual Meteor Magnitudes"
date: "`r Sys.Date()`"
output:
    rmarkdown::html_vignette:
        toc: true
        fig_width: 6
        fig_height: 4
vignette: >
  %\VignetteIndexEntry{Parameter Estimation of the Geometric Model of Visual Meteor Magnitudes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
library(vismeteor)
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

## Introduction

The geometric model of visual meteor magnitudes is a commonly used statistical approach
to describe the magnitude distribution of a meteor shower. The observable magnitude distribution
of meteors is then
$$
P[M = m] \sim
\begin{cases}
  f(m_{\mathrm{lim}} - m)\, r^m, & \text{if } m_{\mathrm{lim}} - m > -0.5,\\[5pt]
  0 & \text{otherwise,}
\end{cases}
$$

where $m_{\mathrm{lim}}$ denotes the limiting (non-integer) magnitude of the observation,
and $m$ the integer meteor magnitude.
The function $f(\cdot)$ denotes the perception probability function.

The estimation of the population index r, briefly called the r-value,
is a common task in the evaluation of meteor magnitudes.
Here we demonstrate two methods to estimate this parameter.

First, we obtain some magnitude observations from the example data set,
which also includes the limiting magnitude.

```{r, echo=TRUE, results='hide'}
observations <- with(PER_2015_magn$observations, {
    idx <- !is.na(lim_magn) & sl_start > 135.81 & sl_end < 135.87
    data.frame(
        magn_id = magn_id[idx],
        lim_magn = lim_magn[idx]
    )
})
head(observations, 5) # Example values
```

```{r, echo=FALSE, results='asis'}
knitr::kable(head(observations, 5))
```

Next, the observed meteor magnitudes are matched with the corresponding observations.
This is necessary as we need the limiting magnitudes of the observations to determine
the r-value.

Using

```{r, echo=TRUE, results='hide'}
magnitudes <- with(new.env(), {
    magnitudes <- merge(
        observations,
        as.data.frame(PER_2015_magn$magnitudes),
        by = "magn_id"
    )
    magnitudes$magn <- as.integer(as.character(magnitudes$magn))
    subset(magnitudes, (magnitudes$lim_magn - magnitudes$magn) > -0.5)
})
head(magnitudes, 5) # Example values
```

we obtain a data frame with the absolute observed frequencies `Freq` for each
observation of a magnitude class. The expression
`subset(magnitudes, (magnitudes$lim_magn - magnitudes$magn) > -0.5`
ensures that meteors fainter than the limiting magnitude are not used if they exist.

```{r, echo=FALSE, results='asis'}
knitr::kable(head(magnitudes[magnitudes$Freq > 0, ], 5))
```

This data frame contains a total of `r sum(magnitudes$Freq)` meteors.
This is a sufficiently large number to estimate the r-value.


## Maximum Likelihood Method

The maximum likelihood method can be used to estimate the r-value in an asymptotically
unbiased manner. For this, the function `dvmgeom()` is needed, which returns the
probability density of the observable meteor magnitudes when the r-value and the
limiting magnitudes are known.

The following algorithm estimates the r-value by maximizing the likelihood with
the `optim()` function. The function `ll()` returns the negative log-likelihood,
as `optim()` identifies a minimum.

```{r, echo=TRUE, results='hide'}
# maximum likelihood estimation (MLE) of r
result_ml <- with(magnitudes, {
    # log likelihood function
    ll <- function(r) -sum(Freq * dvmgeom(magn, lim_magn, r, log = TRUE))
    r_start <- 2.0 # starting value
    r_lower <- 1.2 # lowest expected value
    r_upper <- 4.0 # highest expected value
    # find minimum
    optim(r_start, ll, method = "Brent", lower = r_lower, upper = r_upper, hessian = TRUE)
})
```

This gives the expected value and the variance of the r-value:

```{r, echo=TRUE}
print(result_ml$par) # mean of r
print(1 / result_ml$hessian[1][1]) # variance of r
```

We can additionally visualize the likelihood function here.

```{r, echo=TRUE, results='hide'}
with(new.env(), {
    data_plot <- data.frame(r = seq(2.0, 2.8, 0.01))
    data_plot$ll <- mapply(function(r) {
        with(magnitudes, {
            # log likelihood function
            sum(Freq * dvmgeom(magn, lim_magn, r, log = TRUE))
        })
    }, data_plot$r)
    data_plot$l <- exp(data_plot$ll - max(data_plot$ll))
    data_plot$l <- data_plot$l / sum(data_plot$l)
    brks <- seq(min(data_plot$r) - 0.02, max(data_plot$r) + 0.02, by = 0.02)
    plot(data_plot$r, data_plot$l,
        # breaks = brks,
        type = "l",
        col = "blue",
        xlab = "r",
        xaxt = "n",
        ylab = "likelihood"
    )
    xlabels <- seq(min(round(data_plot$r, 1)) - 0.1, max(round(data_plot$r, 1)) + 0.1, by = 0.1)
    axis(
        side = 1,
        at = xlabels,
        labels = sprintf("%.1f", xlabels)
    )
    abline(v = result_ml$par, col = "red", lwd = 1)
})
```

The likelihood function is approximately a normal distribution in this case. This is important in this context because the variance of the estimated r-value is derived from the curvature (the second derivative at the maximum) of the log-likelihood function.


## By Rate

This method leverages the idea that we can estimate how many meteors would have been observed if the limiting magnitude had been reduced by `1`. This yields the statistic

$$
\mathbb{E}\!\left[\frac{1}{r}\right] = \frac{1}{n}\sum \frac{
    f(m_{\mathrm{lim}} - m - 1)
}{
    f(m_{\mathrm{lim}} - m)
}
$$

where $f(\cdot)$ denotes the perception probability function.

The appeal of this statistic is that the r-value can be estimated by taking a mean of the ratio above.
Very bright meteors are inherently grouped into a single magnitude class whenever their
perception probability is approximately `1.0`, which suppresses outliers in this part of the magnitude range.
The resulting estimator is therefore inexpensive to compute and directly comparable to the likelihood-based approaches discussed earlier.

We estimate $r = 1/a$ from the sample mean of $a$. Since $r$ is a function
of $a$, we apply the delta method to correct for bias and to compute the variance of $r$.
The delta method uses a Taylor expansion around the mean of $a$ to approximate the
distribution of $r$.

```{r, echo=TRUE, results='hide'}
result_rate <- with(magnitudes, {
    N <- sum(Freq)
    a <- vmperception(lim_magn - magn - 1) / vmperception(lim_magn - magn)
    a_mean <- as.numeric(weighted.mean(a, w = Freq))
    a_var <- as.numeric(cov.wt(cbind(a), wt = Freq)$cov) / N
    # apply the delta method and return the result
    list(
        "mean" = 1 / a_mean - a_var / a_mean^3,
        "var" = a_var / a_mean^4
    )
})
```

This gives the expected value and the variance of the r-value:

```{r, echo=TRUE}
print(result_rate$mean) # mean of r
print(result_rate$var) # variance of r
```

Using the bootstrap method, it can be assessed whether the mean is normally distributed.

```{r, echo=TRUE, results='hide'}
# Bootstrapping Method
r_means <- with(magnitudes, {
    N <- sum(Freq)
    a <- vmperception(lim_magn - magn - 1) / vmperception(lim_magn - magn)
    replicate(50000, {
        s <- sample(a, size = N, replace = TRUE, prob = Freq)
        s_mean <- mean(s)
        s_var <- var(s) / N
        1 / s_mean - s_var / s_mean^3
    })
})
```

The graphical representation indicates that this is indeed approximately the case.

```{r, echo=TRUE, results='show'}
with(new.env(), {
    r_sd <- sqrt(result_rate$var)
    r_min <- result_rate$mean - 3 * r_sd
    r_max <- result_rate$mean + 3 * r_sd
    r <- subset(r_means, r_means > r_min & r_means < r_max)
    brks <- seq(min(r) - 0.02, max(r) + 0.02, by = 0.02)
    hist(r,
        breaks = brks,
        col = "skyblue",
        border = "black",
        main = "Histogram of mean r",
        xlab = "r",
        xaxt = "n",
        ylab = "count"
    )
    xlabels <- seq(min(round(r, 1)) - 0.1, max(round(r, 1)) + 0.1, by = 0.1)
    axis(
        side = 1,
        at = xlabels,
        labels = sprintf("%.1f", xlabels)
    )
    abline(v = result_rate$mean, col = "red", lwd = 1)
})
```


## Variance-Stabilizing Transformation as an Alternative Method

Estimation based on the maximum likelihood principle is computationally demanding.
As an alternative, a variance-stabilizing transformation can be applied.
This transformation maps meteor magnitudes onto a different scale,
yielding a distribution whose variance no longer depends on the parameter `r`.

The variance-stabilizing transformation has the following additional advantages:

- Since the variance does not depend on the r-value, the accuracy of estimators (e.g., maximum likelihood estimators) is easier to assess,
- The variance bound (Cramér–Rao lower bound) then often depends only on the sample size `n`, but not on the true r-value,
- Estimators such as the sample mean are homoscedastic, i.e., their dispersion is constant across the parameter space. As a result, many classical results of linear estimation theory (BLUE properties, Gauss–Markov theorem) hold without additional transformations,
- Because the variance is fixed, the calculation of standard errors is independent of the unknown r-value,
- Confidence intervals can be more easily standardized and are equally well calibrated across the entire parameter space,
- Test statistics (e.g., likelihood ratio or score tests) exhibit a more uniform distribution, since the variance does not need to be treated as an additional unknown. This improves test power and simplifies asymptotic approximations.

The resulting procedure is straightforward: it suffices to compute the mean of
the transformed meteor magnitudes, from which an estimate of the parameter `r` is obtained.

To convert the mean of the transformed values back to the original `r` scale, we apply
the delta method. This accounts for the nonlinearity of the back-transformation and
provides both a bias correction (second-order term).

```{r, echo=TRUE, results='show'}
result_vs <- with(magnitudes, {
    N <- sum(Freq)
    tm <- vmgeom_vst_from_magn(magn, lim_magn)
    tm_mean <- sum(Freq * tm) / N
    tm_var <- sum(Freq * (tm - tm_mean)^2) / (N - 1)
    tm_mean_var <- tm_var / N

    # Delta method: variance and variance of r
    r_hat <- vmgeom_vst_to_r(tm_mean)
    dr_dtm <- vmgeom_vst_to_r(tm_mean, deriv_degree = 1L)
    d2r_dtm2 <- vmgeom_vst_to_r(tm_mean, deriv_degree = 2L)
    r_hat <- r_hat - 0.5 * d2r_dtm2 * tm_mean_var
    r_var <- dr_dtm^2 * tm_mean_var + 0.5 * d2r_dtm2^2 * tm_mean_var^2

    list("mean" = r_hat, "var" = r_var)
})
```

Thus, one obtains the mean and the variance of the mean of `r``.

```{r, echo=TRUE, results='show'}
print(paste("r mean:", result_vs$mean))
print(paste("r var:", result_vs$var))
```

Using the bootstrap method, it can be assessed whether the mean is normally distributed.

```{r, echo=TRUE, results='hide'}
# Bootstrapping Method
r_means <- with(magnitudes, {
    N <- sum(Freq)
    tm <- vmgeom_vst_from_magn(magn, lim_magn)
    replicate(50000, {
        s <- sample(tm, size = N, replace = TRUE, prob = Freq)
        s_mean <- mean(s)
        s_var <- var(s) / N

        r_hat <- vmgeom_vst_to_r(s_mean)
        d2r_ds2 <- vmgeom_vst_to_r(s_mean, deriv_degree = 2L)
        r_hat - 0.5 * d2r_ds2 * s_var
    })
})
```

The graphical representation indicates that this is indeed approximately the case.

```{r, echo=TRUE, results='show'}
with(new.env(), {
    r_sd <- sqrt(result_vs$var)
    r_min <- as.vector(result_vs$mean - 3 * r_sd)
    r_max <- as.vector(result_vs$mean + 3 * r_sd)
    r <- subset(r_means, r_means > r_min & r_means < r_max)
    brks <- seq(min(r) - 0.02, max(r) + 0.02, by = 0.02)
    hist(r,
        breaks = brks,
        col = "skyblue",
        border = "black",
        main = "Histogram of mean r",
        xlab = "r",
        xaxt = "n",
        ylab = "count"
    )
    xlabels <- seq(min(round(r, 1)) - 0.1, max(round(r, 1)) + 0.1, by = 0.1)
    axis(
        side = 1,
        at = xlabels,
        labels = sprintf("%.1f", xlabels)
    )
    abline(v = result_vs$mean, col = "red", lwd = 1)
})
```

## Residual Analysis

So far, we have operated under the assumption that the real distribution of meteor magnitudes
is exponential and that the perception probabilities are accurate.
We now use the Chi-Square goodness-of-fit test to check whether the observed frequencies match
the expected frequencies. Then, using the estimated r-value, we retrieve the relative
frequencies `p` for each observation and add them to the data frame `magnitudes`:

```{r, echo=TRUE, results='hide'}
magnitudes$p <- with(magnitudes, dvmgeom(m = magn, lm = lim_magn, result_rate$mean))
```

We must also consider the probabilities for the magnitude class with the brightest meteors.

```{r, echo=TRUE, results='hide'}
magn_min <- min(magnitudes$magn)
```

The smallest magnitude class `magn_min` is `r magn_min`. In calculating the probabilities,
we assume that the magnitude class `r magn_min` contains meteors that are either brighter
or equally bright as `r magn_min` and thus use the function `pvmgeom()` to determine
their probability.

```{r, echo=TRUE, results='asis'}
idx <- magnitudes$magn == magn_min
magnitudes$p[idx] <- with(
    magnitudes[idx, ],
    pvmgeom(m = magn + 1L, lm = lim_magn, result_rate$mean, lower.tail = TRUE)
)
```

This ensures that the probability of observing a meteor of any given magnitude is 100%.
This is known as the normalization condition. Accordingly, the Chi-Square goodness-of-fit test
will fail if this condition is not met.

We now create the contingency table `magnitutes_observed` for the observed meteor magnitudes
and its margin table.

```{r, echo=TRUE}
magnitutes_observed <- xtabs(Freq ~ magn_id + magn, data = magnitudes)
magnitutes_observed_mt <- margin.table(magnitutes_observed, margin = 2)
print(magnitutes_observed_mt)
```

Next, we check which magnitude classes need to be aggregated so that each contains
at least 10 meteors, allowing us to perform a Chi-Square goodness-of-fit test.

The last output shows that meteors of magnitude class `0` or brighter must be combined into
a magnitude class `0-`. Meteors with a brightness less than `4` are grouped here in the
magnitude class `4+`, and a new contingency table magnitudes.observed is created:

```{r, echo=TRUE}
magnitudes$magn[magnitudes$magn <= 0] <- "0-"
magnitudes$magn[magnitudes$magn >= 4] <- "4+"
magnitutes_observed <- xtabs(Freq ~ magn_id + magn, data = magnitudes)
print(margin.table(magnitutes_observed, margin = 2))
```

We now need the corresponding expected relative frequencies

```{r, echo=TRUE}
magnitutes_expected <- xtabs(p ~ magn_id + magn, data = magnitudes)
magnitutes_row_freq <- margin.table(magnitutes_observed, margin = 1)
magnitutes_expected <- sweep(magnitutes_expected, 1, magnitutes_row_freq, `*`)
magnitutes_expected <- magnitutes_expected / sum(magnitutes_expected)
print(sum(magnitudes$Freq) * margin.table(magnitutes_expected, margin = 2))
```

and then carry out the Chi-Square goodness-of-fit test:

```{r, echo=TRUE, results='asis'}
chisq_test_result <- chisq.test(
    x = margin.table(magnitutes_observed, margin = 2),
    p = margin.table(magnitutes_expected, margin = 2)
)
```

As a result, we obtain the p-value:

```{r, echo=TRUE}
chi2_df <- chisq_test_result$parameter - 1
chi2_pval <- pchisq(chisq_test_result$statistic, df = chi2_df, lower.tail = FALSE)
print(chi2_pval)
```

If we set the level of significance at 5 percent, then it is clear that the p-value with
`r unname(chi2_pval)` is greater than 0.05. Thus, under the assumption that the
magnitude distribution follows an geometric meteor magnitude distribution and assuming that
the perception probabilities are correct (i.e., error-free or precisely known),
the assumptions cannot be rejected. However, the converse is not true; the assumptions
may not necessarily be correct. The total count of meteors here is too small for such
a conclusion.

To verify the p-value, we also graphically represent the Pearson residuals:

```{r, fig.show='hold'}
chisq_test_residuals <- with(new.env(), {
    chisq_test_residuals <- residuals(chisq_test_result)
    v <- as.vector(chisq_test_residuals)
    names(v) <- names(chisq_test_residuals)
    v
})
plot(
    chisq_test_residuals,
    main = "Residuals of the chi-square goodness-of-fit test",
    xlab = "m",
    ylab = "Residuals",
    ylim = c(-3, 3),
    xaxt = "n"
)
abline(h = 0.0, lwd = 2)
axis(1, at = seq_along(chisq_test_residuals), labels = names(chisq_test_residuals))
```

---
title: "vismeteor"
output:
    rmarkdown::html_vignette:
        toc: true
        fig_width: 6
        fig_height: 4
vignette: >
  %\VignetteIndexEntry{vismeteor}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

```{r setup, include = FALSE}
library(vismeteor)
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

## Introduction

`vismeteor` was developed to provide a comprehensive tool for analyzing visually observed
meteors. It takes into account human perception of different meteor magnitudes through
perception probabilities. Therefore, this package provides methods that incorporate
these perception probabilities in the evaluation of meteor magnitudes.

Perception probabilities are crucial for analyzing observed magnitude distributions,
as they give rise to specific distributions unique to visual meteor observations.
One such model is the Geometric Model of Visual Meteor Magnitudes.
It is based on the standard geometric distribution, adjusted by multiplying its probabilities with
the perception probabilities. The resulting distribution has a single parameter:
the population index r.
This package provides functions to work with this adjusted distribution.

Additionally, this package offers tools to facilitate the evaluation, recognizing that
visual meteor observations typically report only total counts over a specified period,
rather than individual meteor events.

For visual observations of meteor magnitudes, the observed counts can be specified in
fractional values (half meteors). This package includes a function for unbiased rounding
of these fractional values to use standard tools in R that only allow integer values.

Beyond magnitude distributions, the package also helps in analyzing how meteor activity
varies over time or solar longitude. It provides a routine for fitting smooth, parsimonious
profiles to noisy meteor count series, automatically choosing a small set of spline knot
positions from a user-supplied candidate grid under an information criterion or
cross-validation score.

## Knot Selection for Spline Models

The function `select_knots()` performs greedy stepwise selection of interior spline knots from
a user-supplied candidate grid. In each step it adds (forward) or removes (backward) the single
knot that most improves a user-defined score, such as BIC, AIC, or a cross-validation criterion.
The routine is model-agnostic: the scoring function is responsible for fitting the model — a
Poisson GLM, a natural cubic spline via `glm()`, or any other form — so `select_knots()` can
be reused across very different analyses.

A typical application is fitting the activity profile of a meteor shower over solar longitude
through noisy counts. A fine grid of candidate positions captures where the curve *could* bend;
`select_knots()` reduces it to a small subset where the data actually demand a bend, yielding
a smooth, parsimonious fit.

```{r, echo=TRUE}
set.seed(1)
x <- seq(0, 10, length.out = 200)
y <- rpois(length(x), lambda = 50 + 30 * sin(x))
dat <- data.frame(x = x, y = y)

fit <- \(d, knots) {
    f <- if (length(knots) == 0L) {
        y ~ splines::ns(x)
    } else {
        y ~ splines::ns(x, knots = knots)
    }
    glm(f, family = poisson(), data = d)
}
score_bic <- \(d, knots) fit(d, knots) |> BIC()

result <- select_knots(
    dat,
    knot_candidates = seq(1, 9, by = 0.5),
    score_fun = score_bic
)
final_knots <- sort(c(result$knots, result$fixed_knots))
final_fit <- fit(dat, final_knots)

op <- par(mar = c(7, 4, 4, 2) + 0.1)
plot(
    x, y,
    pch = 20, col = "gray",
    main = "Poisson spline fit with selected knots",
    xlab = "x", ylab = "count"
)
lines(x, 50 + 30 * sin(x), col = "darkgreen", lwd = 2, lty = 2)
lines(x, final_fit |> predict(type = "response"), col = "blue", lwd = 2)
abline(v = final_knots, col = "red", lty = 3)
legend(
    "bottom",
    inset = c(0, -0.35),
    legend = c("true mean", "spline fit", "selected knots"),
    col = c("darkgreen", "blue", "red"),
    lty = c(2, 1, 3),
    lwd = c(2, 2, 1),
    horiz = TRUE,
    bty = "n",
    xpd = TRUE
)
par(op)
```

See `?select_knots` and `vignette("select_knots")` for more information.

## Perception probabilities

Meteors appear randomly distributed across the sky. The perception of an observer can
vary significantly for each meteor, influenced largely by the region of the sky where
the meteor appears. Depending on its position, a meteor might be perceived or missed entirely,
leading to the concept of perception probabilities.

Perception probabilities quantify the likelihood of an observer detecting a meteor.
They provide a statistical measure reflecting the chance of seeing or not seeing a meteor
under various conditions. In statistical modeling of visual observations, these probabilities
are commonly expressed in relation to the difference between the meteor's magnitude and the
limiting magnitude of the observation.

This package includes the function `vmperception()` to return perception probabilities for
visual meteor magnitudes.

Custom perception probability functions can be passed to most functions of this package.
This allows users to tailor the behavior of the functions to their specific needs and
conduct their own studies on the appearance of observable meteor magnitudes.
This flexibility is particularly valuable for researchers wishing to incorporate
their own empirical data or models into the analysis.

```{r, echo=TRUE}
plot(
    vmperception,
    -0.5, 8,
    main = "Perception probability of visual meteor magnitudes",
    col = "blue",
    xlab = "limiting magnitude - meteor magnitude",
    ylab = "p"
)
```

See `?vmperception` for more information.

## Geometric Model of Visual Meteor Magnitudes

In visual meteor observations, magnitudes are typically recorded as integer values.  
Hence, the geometric model is discrete and its probability mass function is
$$
P[M = m] \sim
\begin{cases}
  f(m_{\mathrm{lim}} - m)\, r^m, & \text{if } m_{\mathrm{lim}} - m > -0.5,\\[5pt]
  0 & \text{otherwise,}
\end{cases}
$$

where $m_{\mathrm{lim}}$ denotes the limiting (non-integer) magnitude of the observation,
and $m$ the integer meteor magnitude.
The function $f(\cdot)$ denotes the perception probability function.

This distribution is the product of the perception probabilities and the
underlying geometric distribution of meteor magnitudes.
Therefore, the parameter $p$ of the geometric distribution is $p = 1 - 1/r$.

The main advantage of this model is its simplicity.
When the number of observed meteors is small, it can often be shown that the
distribution of meteor magnitudes is well described by this model.

```{r, echo=TRUE, results='hide'}
m <- seq(6, -4, -1)
limmag <- 6.5
r <- 2.0
p <- vismeteor::dvmgeom(m, limmag, r)
barplot(
    p,
    names.arg = m,
    main = paste0("Density (r = ", r, ", limmag = ", limmag, ")"),
    col = "blue",
    xlab = "m",
    ylab = "p",
    border = "blue",
    space = 0.5
)
axis(side = 2, at = pretty(p))
```

See `?vmgeom` for more information.

## Ideal Distribution of Visual Meteor Magnitudes

The ideal model of visual meteor magnitudes is an alternative to the geometric model.
It more accurately accounts for the finiteness of meteor magnitudes across the entire
magnitude spectrum, using the ideal gas from theoretical physics as a conceptual analogy.

This model is particularly useful when, compared to the geometric model,
fewer faint meteor magnitudes are observed.

The density of the ideal distribution is

$$
{\displaystyle \frac{\mathrm{d}p}{\mathrm{d}m} = \frac{3}{2} \, \log(r) \sqrt{\frac{r^{3 \, \psi + 2 \, m}}{(r^\psi + r^m)^5}}}
$$
where $m$ is the continuous (real-valued) meteor magnitude,
$r = 10^{0.4} \approx 2.51189 \dots$ is a constant and
$\psi$ is the only parameter of this magnitude distribution.

```{r, echo=TRUE, results='hide'}
m <- seq(6, -4, -1)
psi <- 5.0
limmag <- 6.5
p <- vismeteor::dvmideal(m, limmag, psi)
barplot(
    p,
    names.arg = m,
    main = paste0("Density (psi = ", psi, ", limmag = ", limmag, ")"),
    col = "blue",
    xlab = "m",
    ylab = "p",
    border = "blue",
    space = 0.5
)
axis(side = 2, at = pretty(p))
```

See `?mideal` and `?vmideal` for more information.

## Fractional Counting

In the statistical analysis of visual meteor observations, a method known as "count data"
in categorical time series is used. Observers record the number of meteors in each category
over specific time intervals.

In visual meteor observation, observers record numerical meteor magnitudes and sum them,
uniquely allowing for "half" counts when indecisive between two values. This fractional
counting reflects measurement uncertainty, splitting counts between adjacent magnitude categories.

While fractional counting accommodates measurement uncertainty with "half" counts, some tools
cannot process these fractional values and require integer rounding.
The function `vmtable()` addresses this by rounding the magnitudes in a contingency table to whole
numbers. It ensures that the rounding process only alters the marginal totals when necessary,
preserving the overall count integrity. This means both the grand total and the intermediate
sums of meteors observed remain consistent, ensuring accurate and usable data for
subsequent analysis.

Example:
```{r, echo=TRUE}
mt <- as.table(matrix(
    c(
        0.0, 0.0, 2.5, 0.5, 0.0, 1.0,
        0.0, 1.5, 2.0, 0.5, 0.0, 0.0,
        1.0, 0.0, 0.0, 3.0, 2.5, 0.5
    ),
    nrow = 3, ncol = 6, byrow = TRUE
))
colnames(mt) <- seq(6)
rownames(mt) <- c("A", "B", "C")
margin.table(mt, 1)
margin.table(mt, 2)

# contingency table with integer values
(mt_int <- vmtable(mt))
margin.table(mt_int, 1)
margin.table(mt_int, 2)
```

See `?vmtable` for more information.

## Quantile Analysis with Minimum Meteor Count

The function `freq_quantile()` is tailored for analyzing time series data in visual
meteor observation, where the count of meteors is recorded over specific time intervals.

Unlike traditional methods that sort quantiles based on time and percent, which often result
in some quantiles having fewer meteors than desired, `freq_quantile()` constructs quantiles
with a focus on ensuring a minimum number of meteors in each quantile.

This method addresses the challenge of varying meteor counts in each interval, including those
with zero meteors. By utilizing `freq_quantile()`, users can effectively divide the time series
into quantiles that are both time-based and density-based, enhancing the understanding of meteor
occurrence and distribution over the observed period while ensuring each quantile meets the
minimum count criterion.

This approach provides a more nuanced and reliable analysis, especially vital when dealing with
the inherent variability in meteor observations.

The following example represents a time-ordered list of observed meteor counts.
The objective is to group these counts into quantiles while maintaining their chronological order,
ensuring that each quantile contains at least 10 meteors.

```{r, echo=TRUE}
freq <- c(1, 8, 3, 3, 4, 9, 5, 0, 0, 2, 7, 8, 2, 6, 4)
f <- freq_quantile(freq, 10)
print(f)
print(tapply(freq, f, sum))
```

See `?freq_quantile` for more information.

## Interfacing VMDB Data with VMDB Loaders

The functions `load_vmdb_rates()` and `load_vmdb_magnitudes()` query an
[imo-vmdb](https://pypi.org/project/imo-vmdb/) web server through its REST API to retrieve
visual meteor observations from the Visual Meteor Database (VMDB). The imo-vmdb server ingests
and validates the original VMDB exports and serves them in a structured, relational form, so
callers receive enriched and consistent observation records rather than raw exports.

Both functions take a `base_url` pointing at the API endpoint and accept filters such as
`shower`, `period`, `sl` (solar longitude) and `lim_magn` (limiting magnitude). For rate
observations, `sun_alt_max` and `moon_alt_max` further constrain sky conditions. Optional
companion data — observing session metadata and per-observation magnitude distributions —
can be requested via `with_sessions` and `with_magnitudes`.

Within this package, `PER_2015_rates` and `PER_2015_magn` are provided as example data sets.
They are included for testing and training purposes, allowing users to
understand and utilize the full functionality of the entire package.

See `?load_vmdb`, `?PER_2015_rates` and `?PER_2015_magn` for more information.

---
title: "Knot Selection for Smooth Profiles of Visual Meteor Showers"
date: "`r Sys.Date()`"
output:
    rmarkdown::html_vignette:
        toc: true
        fig_width: 6
        fig_height: 4
vignette: >
  %\VignetteIndexEntry{Knot Selection for Smooth Profiles of Visual Meteor Showers}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

## Setup

Load the package before running any of the code below:

```{r setup}
library(vismeteor)
```

## Introduction

The activity of a visual meteor shower is a smooth, a priori unknown
function of solar longitude. The classical evaluation summarises this by
computing the Zenithal Hourly Rate (ZHR) per observation interval and
plotting it as a noisy stepped curve. Single-interval ZHRs have high
variance — Poisson noise plus observer-to-observer scatter — and the
underlying activity profile is barely visible.

A spline-based fit across solar longitude collapses this noise into a
smooth, bias-controlled curve, provided that the spline's complexity
(number and location of interior knots) is chosen wisely.
`select_knots()` automates that choice: given a candidate grid of
possible interior knots and a user-supplied score function (BIC, AIC, or
a cross-validation criterion), it returns a small subset that minimises
the score.

Conceptually, the supplied `knot_candidates` form a *grid of intervals*
over the solar longitude: between two adjacent candidates lies an
interval in which the spline can bend independently. `select_knots()`
decides which interval boundaries are actually required — i.e. which
positions carry enough curvature to justify a knot under the score. The
natural resolution of the input grid in visual meteor work is the
observation granularity itself: typical VMDB observation intervals are
10–15 minutes long, corresponding to roughly 0.007–0.010 degrees of
solar longitude. A fine candidate grid at that resolution is the
practice recommendation; `select_knots()` will then reduce it to a
handful of statistically significant knots.

## When to use `select_knots()`

The point of `select_knots()` is to keep the knot positions *explicit
and few*: a handful of values you can plot, label, and reason about,
instead of a dense anonymous knot grid hidden inside a penalised
smoother. The function is model-agnostic — your `score_fun` picks the
model family, offset, covariates, and the criterion (BIC, AIC, or a
cross-validation score) you optimise.

This function is the right tool when

- you want a smooth curve through noisy count data,
- knot positions should remain interpretable,
- the model (family, offset, covariates) is too specific for a generic
  GAM smoother.

Prefer alternatives when

- you want a continuous roughness penalty over a dense knot grid:
  `mgcv::s(bs = "ps")`;
- you want adaptive smoothing (smoothing parameter varying over the
  predictor): `mgcv::s(bs = "ad")`;
- you want a hinge-function basis with its own GCV-based pruning:
  `earth::earth`;
- you want a guaranteed global optimum: `select_knots()` is greedy and
  only finds a local optimum (see `?select_knots`).

**Important clarification.** `select_knots()` chooses knots for a *good
fit*, not for detecting extrema. Knots typically land *next to* maxima
or minima — where the curvature is high — and not *on* them. Local
extrema of the activity profile become visible from the *shape* of the
fitted curve, not from the list of selected knots.

The implementation is generic, but its motivation comes directly from
the visual-meteor use case: activity profiles are the natural setting
in which knot positions can be *physically interpreted* — main maximum,
sub-maxima, shower onset and end. The same interpretive value carries
over to other meteor-related smooth fits — for instance the apparent
magnitude distribution of a shower. Alternative packages (`mgcv`,
`earth`) replace explicit knots with continuous penalty mechanisms,
which gives up that interpretation.

## Data preparation

We use the package's example data set `PER_2015_rates` (visual rate
observations of the Perseids 2015; see `?PER_2015_rates` and
`?load_vmdb`). The raw filtering follows standard practice:

```{r, echo=TRUE}
data(PER_2015_rates)
obs <- PER_2015_rates$observations |> subset(
    rad_alt > 15 & moon_alt < 0 & sun_alt < -5 &
        sl_start >= 139.2 & sl_end <= 140.6
)

deg2rad <- \(x) x * pi / 180
obs$sl <- (obs$sl_start + obs$sl_end) / 2
obs$t <- with(obs, t_eff * sin(deg2rad(rad_alt)) / f)

rates <- data.frame(
    sl = obs$sl,
    t = obs$t,
    limmag = obs$lim_magn,
    freq = obs$freq
)
rates <- rates[order(rates$sl), ]
nrow(rates)
```

What each step does:

- **`rad_alt > 15`** — when the radiant is low above the horizon
  the geometric correction `sin(rad_alt)` becomes numerically
  fragile and atmospheric extinction dominates; observations below
  15 degrees are discarded by convention.
- **`moon_alt < 0`** — the Moon above the horizon brightens the sky and
  reduces the effective limiting magnitude in a way the simple `limmag`
  correction does not capture; these observations are excluded rather
  than corrected.
- **`sun_alt < -5`** — nautical or astronomical twilight; before that
  the sky is too bright.
- **`sl_start >= 139.2 & sl_end <= 140.6`** — *only for this vignette*:
  a tight window centred on the Perseid main peak, chosen narrow enough
  to keep the vignette build fast while still containing the peak
  (≈ 140.0 deg) and visible curvature on both sides. In a real analysis
  you would use the full activity range (roughly 107–155 deg for the
  Perseids).
- **`sl = (sl_start + sl_end) / 2`** — VMDB observations are intervals;
  we summarise each by its midpoint, the standard choice for the
  regressor.
- **`t = t_eff * sin(rad_alt) / f`** is the central correction. It
  rescales the raw observing time to what the observer would have
  spent under ideal sky conditions *at their own limiting magnitude*:
  radiant in the zenith (`sin(rad_alt)`) and a clear sky (`f`).
  `t_eff` is the effective observing time after deducting breaks; `f`
  is the cloud-cover correction factor (`>= 1`, larger when more sky
  is obstructed). The correction from each observer's own limiting
  magnitude to the ZHR reference 6.5 is **not** built into `t`; the
  GLM below picks it up via the `limmag` covariate.

We will call `t` simply the **observing time** from now on — not to be
confused with `t_eff`, the raw effective observing time after breaks;
`t` is `t_eff` after the zenith (`sin(rad_alt)`) and cloud-cover
(`/ f`) corrections.

## Model: Poisson GLM with a spline of solar longitude

```{r, echo=TRUE}
fit_glm <- \(rates, knots) {
    f <- if (length(knots) == 0L) {
        freq ~ splines::ns(sl) + limmag + offset(log(t))
    } else {
        freq ~ splines::ns(sl, knots = knots) +
            limmag + offset(log(t))
    }
    stats::glm(f, data = rates, family = stats::poisson())
}
score_bic <- \(rates, knots) stats::BIC(fit_glm(rates, knots))
```

`offset(log(t))` makes the linear predictor a log-rate per observing
hour at the observer's own limiting magnitude. The `limmag` term then
rescales that rate later to the ZHR reference (`limmag = 6.5`):
`exp(coef(limmag))` is the data-driven population index `r` of the
shower under this model. The spline term `splines::ns()` is a natural
cubic spline, well-behaved at the boundaries.

A practical note on the offset: `offset()` inside a formula is a
formula-special and must stay unqualified — both `stats::offset(...)`
in the formula and passing the term via `glm()`'s `offset =` argument
break `predict()` on `newdata`.

## Knot candidates

A regular grid is the simplest choice but ignores the actual observation
density. In practice the candidate grid is built from the data themselves
via `freq_quantile()` (see `?freq_quantile`): partition the observations,
ordered by solar longitude, into bins each containing at least a fixed
minimum number of meteors, then use the meteor-weighted solar-longitude
midpoint of each bin as a candidate. This produces a fine grid in
observation-rich regions and a coarse grid where data is sparse — exactly
the resolution `select_knots()` benefits from:

```{r, echo=TRUE}
rates$q <- vismeteor::freq_quantile(rates$freq, 12) # at least 12 meteors per bin
rates$qsl <- with(
    rates,
    ave(freq * sl, q, FUN = sum) / ave(freq, q, FUN = sum)
)
cand <- sort(unique(round(rates$qsl, 2)))
length(cand)
```

The `round(..., 2)` step collapses bins whose centres are practically
indistinguishable (resolution 0.01 deg). The minimum bin count (`12`)
controls how dense the candidate grid is; a smaller value gives a finer
grid (slower fit), a larger value a coarser one.

## Fixing knots

The `fixed_knots` argument pins knot positions that must be present in
every fitted model. In forward mode they are set from the start and the
search adds further knots on top; in backward mode they are never
proposed for removal, neither singly nor by bulk removal. `fixed_knots`
need not be a subset of `knot_candidates` — values supplied there are
forced into the working set regardless.

A fixed knot is a *constraint* on the search, not a finding from it: the
criterion no longer has the freedom to reject it. The earlier warning
that knots typically land *next to* extrema rather than on them still
applies — forcing a knot exactly on, say, a presumed maximum overrides
that pattern at the chosen position. Use it when there is a non-data
reason a position must be in the model (a documented event time, an
external anchor, a stable reference across analyses); see
`?select_knots` for the full semantics.

## Forward selection

The vignette runs serially (`n_cores = 1L`) because the search window
is narrow. For a full Perseids activity range (≈ 107–155 deg) the
per-round candidate scoring benefits from
`n_cores = parallel::detectCores() - 1L`; see `?select_knots` for the
reproducibility recipe with `mclapply()`.

```{r, echo=TRUE}
res_fwd <- vismeteor::select_knots(rates, cand, score_bic)
sort(c(res_fwd$knots, res_fwd$fixed_knots))
res_fwd$score
```

```{r, echo=TRUE, fig.alt="BIC trajectory of the forward search over the number of interior knots, with the score optimum at the endpoint"}
plot(res_fwd$history$n_knots, res_fwd$history$score,
    type = "b", pch = 19, col = "blue",
    xlab = "number of interior knots", ylab = "BIC",
    main = "Forward selection: BIC trajectory"
)
points(tail(res_fwd$history$n_knots, 1L), res_fwd$score,
    pch = 19, col = "red", cex = 1.5
)
abline(h = res_fwd$score, lty = 2, col = "grey50")
```

The red point marks the score minimum — and with the default
`n_steps = NULL` it is also the *endpoint* of the trajectory, because
the search stops as soon as no further addition improves the score. To
see what one more step would do, re-run the algorithm starting from the
optimum with a positive `n_steps`:

```{r, echo=TRUE}
res_fwd_plus1 <- vismeteor::select_knots(rates, cand, score_bic,
    fixed_knots = c(
        res_fwd$knots,
        res_fwd$fixed_knots
    ),
    n_steps = 1L
)
res_fwd_plus1$score - res_fwd$score # always >= 0: how much worse one step makes it
```

`n_steps` is an *exploration* knob, not a model-selection knob: it
forces the algorithm to take a fixed number of further iterations
regardless of whether they improve the score, so that the immediate
score landscape past the optimum becomes visible through `history`. A
shallow rise tells you the BIC optimum is soft and that a slightly
richer model would not cost much; a steep rise tells you the optimum
is sharp and the parsimony is justified. The same recipe — re-call
with `fixed_knots = c(prev$knots, prev$fixed_knots), n_steps = k` —
extends to as many follow-up steps as you like; see `?select_knots`.

## Backward elimination

```{r, echo=TRUE}
res_bwd <- vismeteor::select_knots(rates, cand, score_bic, backward = TRUE)
sort(c(res_bwd$knots, res_bwd$fixed_knots))
res_bwd$score
```

```{r, echo=TRUE, fig.alt="BIC trajectory of the backward search over the number of interior knots, with the score optimum at the endpoint"}
plot(res_bwd$history$n_knots, res_bwd$history$score,
    type = "b", pch = 19, col = "blue",
    xlab = "number of interior knots", ylab = "BIC",
    main = "Backward elimination: BIC trajectory"
)
points(tail(res_bwd$history$n_knots, 1L), res_bwd$score,
    pch = 19, col = "red", cex = 1.5
)
abline(h = res_bwd$score, lty = 2, col = "grey50")
```

The trajectory starts at the full candidate pool on the right and moves
left as knots are removed; gaps along the x-axis correspond to bulk
removal rounds (see `bulk_gap` in `?select_knots`). The red point again
marks the score minimum.

## Forward vs. backward: when and why they differ

Both searches walk the lattice of knot subsets greedily — each round
commits to the *locally* best add or drop — but they start at opposite
corners:

- **Forward** begins with no interior knots, an underfit model, and
  adds the knot that reduces BIC the most. It stops as soon as no
  single addition improves the score. Tends to be parsimonious; can
  miss combinations of knots that are only useful *jointly* because
  the first knot of the pair, on its own, did not look improving.
- **Backward** begins with *all* candidates as interior knots — an
  overfit, but flexible, model that already captures every local
  curvature — and removes the knot whose removal worsens BIC least.
  Tends to keep more knots; the danger of dropping a globally
  important knot just because its neighbours mask its individual
  contribution is real but less severe than forward's
  miss-by-combination problem.

That is *why* the two paths can disagree: at any intermediate state
the score of a candidate move depends on what is currently selected.
Adding `k_i` to the empty set is a genuinely different operation from
adding `k_i` to a set that already contains its neighbours, and the
two searches encounter `k_i` in incompatible contexts.

**Practical recommendation.** For activity-profile fitting we
recommend `backward = TRUE` for the final analysis, even though the
function's API default is `FALSE` (the function is generic and
defaults to the cheaper-per-fit direction). The reason is qualitative,
not performance: starting from the overfit end ensures that all real
curvature is already present in the initial model, so the search only
has to decide what is redundant — which is the safer mode against the
forward "miss-by-combination" failure. Use `backward = FALSE` (forward)
for a quick, sparse first sketch or when the full-candidate initial
fit is itself too expensive (very large candidate pool, complex model
family).

In practice neither direction is uniformly faster: forward iterates
many small fits one knot at a time, while backward starts with the
single most expensive fit and then collapses fast via the bulk-removal
mechanism (`bulk_gap = 4L` by default; see `?select_knots`).

In any case, running both directions and keeping the lower-BIC result
is a cheap robustness check; large disagreements (in either knot
count or score) are a warning that the chosen criterion is on the
flat part of its landscape and that the data alone may not justify a
unique parsimonious model. The `n_steps` argument lets either search
continue for a fixed number of iterations past its first local
minimum, which is useful for inspecting how flat or sharp that minimum
is.

## Final fit and ZHR curve

```{r, echo=TRUE}
best <- if (res_fwd$score <= res_bwd$score) {
    c(res_fwd$knots, res_fwd$fixed_knots)
} else {
    c(res_bwd$knots, res_bwd$fixed_knots)
}
fit <- fit_glm(rates, sort(best))

# Data-driven population index of the shower, with 95% Wald CI
# (log-scale CI on the limmag coefficient, exponentiated).
limmag_row <- summary(fit)$coefficients["limmag", ]
r_hat <- exp(limmag_row["Estimate"])
r_ci <- exp(limmag_row["Estimate"] +
    c(-1, 1) * 1.96 * limmag_row["Std. Error"])
c(
    estimate = unname(r_hat),
    lower = unname(r_ci[1]),
    upper = unname(r_ci[2])
)
```

The point estimate is the population index `r` implied by the
rate-vs-limmag slope of this dataset; the CI width tells you how
sharply that slope is pinned down by the observations (a tight band
means a well-determined `r`, a wide band means the data leave it
essentially free). It enters the ZHR prediction below.

```{r, echo=TRUE, fig.alt="Fitted ZHR curve over solar longitude with a 95% confidence band and selected knots marked on the axis"}
sl_grid <- seq(139.2, 140.6, length.out = 400)
pred_df <- data.frame(sl = sl_grid, limmag = 6.5, t = 1)
link <- predict(fit, newdata = pred_df, type = "link", se.fit = TRUE)
zhr <- exp(link$fit)
zhr_lo <- exp(link$fit - 1.96 * link$se.fit)
zhr_hi <- exp(link$fit + 1.96 * link$se.fit)

plot(sl_grid, zhr,
    type = "n",
    ylim = range(c(zhr_lo, zhr_hi)),
    xlab = "solar longitude (deg)", ylab = "ZHR",
    main = "PER 2015 - fitted activity profile (95% CI)"
)
polygon(c(sl_grid, rev(sl_grid)),
    c(zhr_lo, rev(zhr_hi)),
    col = adjustcolor("blue", alpha.f = 0.2), border = NA
)
lines(sl_grid, zhr, col = "blue", lwd = 2)
rug(best, col = "red")
```

The prediction is evaluated with `t = 1` (one ZHR-hour) and
`limmag = 6.5` (the reference limiting magnitude), so the response is
the ZHR per hour on the chosen reference scale. The red rug marks on
the x-axis show the selected interior knots; note that they sit *next
to* the activity peak rather than on top of it, as expected — that is
where the curvature is highest.

The shaded band is the pointwise 95 % confidence interval of the fit.
We obtain it on the **link** (log) scale — where the GLM's standard
errors are approximately Gaussian — and then transform back with
`exp()`. The band reflects parameter uncertainty of the spline + linear
predictor; it is *not* a prediction interval for individual
observations, which would also have to account for Poisson dispersion
of the counts themselves.

## Residual analysis with quantile bins

Raw residuals of individual Poisson observations with small expectations
are visually uninformative. A cleaner diagnostic groups the observations
into bins of fixed minimum meteor count using `freq_quantile()` and
compares the observed and expected totals per bin.

```{r, echo=TRUE, fig.alt="Binned observed-over-expected ratios with one-sigma error bars over solar longitude"}
# Use a coarser binning here than for the candidate grid above
# (>= 500 meteors per bin) so that bin uncertainties are small enough
# to read systematic deviations from the diagnostic plot.
rates$qres <- vismeteor::freq_quantile(rates$freq, 500)
rates$pred <- predict(fit, type = "response")

bins <- with(rates, {
    obs_n <- tapply(freq, qres, sum)
    pred_n <- tapply(pred, qres, sum)
    data.frame(
        sl    = tapply(sl * freq, qres, sum) / obs_n,
        ratio = obs_n / pred_n,
        se    = sqrt(obs_n) / pred_n
    )
})
nrow(bins)

plot(bins$sl, bins$ratio,
    pch = 19, col = "blue",
    ylim = range(c(
        bins$ratio - 2 * bins$se,
        bins$ratio + 2 * bins$se
    )),
    xlab = "solar longitude (deg)",
    ylab = "observed / expected",
    main = "PER 2015 - binned residual ratios (+/- 1 sigma)"
)
arrows(bins$sl, bins$ratio - bins$se,
    bins$sl, bins$ratio + bins$se,
    angle = 90, code = 3, length = 0.03, col = "blue"
)
abline(h = 1, lty = 2, col = "grey50")
```

A well-fitting model scatters the bin ratios randomly around 1, with
about two thirds of the one-sigma error bars crossing the dashed line.
Systematic arches over solar longitude would indicate too few knots in
a region; visible scatter with no trend is Poisson over-dispersion (not
pursued here).

The same `freq_quantile()` helper drives both the candidate grid above
and the diagnostic here, but with different minimum bin sizes: a small
minimum (12) for a fine candidate grid that `select_knots()` can prune
from, and a larger minimum (500) for a small number of stable bins on
which residual deviations are visually detectable. In the narrow vignette
window the larger minimum yields only a handful of bins (see the
`nrow(bins)` output); a full activity-range analysis would lower this
minimum to keep enough bins for a visible trend.

## See also

- `?select_knots` — the function reference.
- `?freq_quantile` — quantile bins used here for the residual diagnostic.
- `?PER_2015_rates` — the example data set.
- `?load_vmdb` — importing rate and magnitude observations from the
  VMDB API.

---
title: "Parameter Estimation of the Ideal Distribution of Visual Meteor Magnitudes"
date: "`r Sys.Date()`"
output:
    rmarkdown::html_vignette:
        toc: true
        fig_width: 6
        fig_height: 4
vignette: >
  %\VignetteIndexEntry{Parameter Estimation of the Ideal Distribution of Visual Meteor Magnitudes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
library(vismeteor)
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

## Introduction

The density of the ideal distribution of meteor magnitudes is
$$
    {\displaystyle f(m) = \frac{\mathrm{d}p}{\mathrm{d}m} = \frac{3}{2} \, \log(r) \sqrt{\frac{r^{3 \, \psi + 2 \, m}}{(r^\psi + r^m)^5}}}
$$
where $m$ denotes the continuous (real-valued) meteor magnitude,
$r = 10^{0.4} \approx 2.51189 \dots$ is a constant, and
$\psi$ is the only parameter of this magnitude distribution.

In visual meteor observations, magnitudes are usually estimated as integer values.
Hence, this distribution is discrete and its probability mass function is given by
$$
P[M = m] \sim
\begin{cases}
  g(m_{\mathrm{lim}} - m) \displaystyle \int\limits_{m-0.5}^{m+0.5} f(u) \, \mathrm{d}u, & \text{if } m_{\mathrm{lim}} - m > -0.5,\\[5pt]
  0 & \text{otherwise,}
\end{cases}
$$
where $m_{\mathrm{lim}}$ denotes the limiting (non-integer) magnitude of the observation,
and $m$ the integer meteor magnitude.
The function $f(\cdot)$ is the continuous density of the ideal magnitude distribution,
and $g(\cdot)$ denotes the perception probability function.

Here we demonstrate two methods to estimate the parameter $\psi$.

First, we obtain some magnitude observations from the example data set,
which also includes the limiting magnitude.

```{r, echo=TRUE, results='hide'}
observations <- with(PER_2015_magn$observations, {
    idx <- !is.na(lim_magn) & sl_start > 135.81 & sl_end < 135.87
    data.frame(
        magn_id = magn_id[idx],
        lim_magn = lim_magn[idx]
    )
})
head(observations, 5) # Example values
```

```{r, echo=FALSE, results='asis'}
knitr::kable(head(observations, 5))
```

Next, the observed meteor magnitudes are matched with the corresponding observations.
This is necessary as we need the limiting magnitudes of the observations to determine
the parameter.

Using

```{r, echo=TRUE, results='hide'}
magnitudes <- with(new.env(), {
    magnitudes <- merge(
        observations,
        as.data.frame(PER_2015_magn$magnitudes),
        by = "magn_id"
    )
    magnitudes$magn <- as.integer(as.character(magnitudes$magn))
    subset(magnitudes, (magnitudes$lim_magn - magnitudes$magn) > -0.5)
})
head(magnitudes[magnitudes$Freq > 0, ], 5) # Example values
```

we obtain a data frame with the absolute observed frequencies `Freq` for each
observation of a magnitude class. The expression
`subset(magnitudes, (magnitudes$lim_magn - magnitudes$magn) > -0.5`
ensures that meteors fainter than the limiting magnitude are not used if they exist.

```{r, echo=FALSE, results='asis'}
knitr::kable(head(magnitudes[magnitudes$Freq > 0, ], 5))
```

This data frame contains a total of `r sum(magnitudes$Freq)` meteors.
This is a sufficiently large number to estimate the parameter.


## Maximum Likelihood Method

The maximum likelihood method can be used to estimate the parameter in an asymptotically unbiased manner.
For this, the function `dvmideal()` is needed, which returns the probability density of
the observable meteor magnitudes when the parameter and the limiting magnitudes are known.

The following algorithm estimates the parameter by maximizing the likelihood with
the `optim()` function. The function `ll()` returns the negative log-likelihood,
as `optim()` identifies a minimum.

```{r, echo=TRUE, results='hide'}
# maximum likelihood estimation (MLE) of psi
result <- with(magnitudes, {
    # log likelihood function
    ll <- function(psi) -sum(Freq * dvmideal(magn, lim_magn, psi, log = TRUE))
    psi_start <- 6.0 # starting value
    psi_lower <- 4.0 # lowest expected value
    psi_upper <- 10.0 # highest expected value
    # find minimum
    optim(psi_start, ll, method = "Brent", lower = psi_lower, upper = psi_upper, hessian = TRUE)
})
```

This gives the expected value and the variance of the parameter:

```{r, echo=TRUE}
psi_mean <- result$par # mean of psi
print(psi_mean)
psi_var <- 1 / result$hessian[1][1] # variance of psi
print(psi_var)
```

We can additionally visualize the likelihood function here.

```{r, echo=TRUE, results='hide'}
with(new.env(), {
    data_plot <- data.frame(psi = seq(4.0, 11, 0.1))
    data_plot$ll <- mapply(function(psi) {
        with(magnitudes, {
            # log likelihood function
            sum(Freq * dvmideal(magn, lim_magn, psi, log = TRUE))
        })
    }, data_plot$psi)
    data_plot$l <- exp(data_plot$ll - max(data_plot$ll))
    data_plot$l <- data_plot$l / sum(data_plot$l)
    plot(data_plot$psi, data_plot$l,
        type = "l",
        col = "blue",
        xlab = "psi",
        ylab = "likelihood"
    )
    abline(v = result$par, col = "red", lwd = 1)
})
```

It is clearly visible that the likelihood function is not normally distributed.
This distribution even belongs to the class of heavy-tailed distributions.
While its maximum is indeed an asymptotically unbiased estimator, this does not hold for the variance.
This is important in this context because the variance of the estimated $\psi$-value is derived from the curvature (the second derivative at the maximum) of the log-likelihood function.
Therefore, the estimator for the variance of $\psi$ is far too small.

## Variance-Stabilizing Transformation as an Alternative Method

Estimation based on the maximum likelihood principle is computationally demanding.
As an alternative, a variance-stabilizing transformation can be applied.
This transformation maps meteor magnitudes onto a different scale,
yielding a distribution whose variance no longer depends on the parameter $\psi$.

The variance-stabilizing transformation has the following additional advantages:

- Since the variance does not depend on $\psi$, the accuracy of estimators (e.g., maximum likelihood estimators) is easier to assess,
- The variance bound (Cramér–Rao lower bound) then often depends only on the sample size `n`, but not on the true value of $\psi$,
- Estimators such as the sample mean are homoscedastic, i.e., their dispersion is constant across the parameter space. As a result, many classical results of linear estimation theory (BLUE properties, Gauss–Markov theorem) hold without additional transformations,
- Because the variance is fixed, the calculation of standard errors is independent of the unknown value of $\psi$,
- Confidence intervals can be more easily standardized and are equally well calibrated across the entire parameter space,
- Test statistics (e.g., likelihood ratio or score tests) exhibit a more uniform distribution, since the variance does not need to be treated as an additional unknown. This improves test power and simplifies asymptotic approximations.

The resulting procedure is straightforward: it suffices to compute the mean of
the transformed meteor magnitudes, from which an estimate of the parameter $\psi$ is obtained.

Note that the variance-stabilizing transformation yields a mean that can be directly analyzed,
but converting it back to the $\psi$ parameter requires applying the delta method.
This accounts for the nonlinearity of the transformation and provides appropriate uncertainty
estimates for $\psi$.

```{r, echo=TRUE, results='show'}
tm_mean <- with(magnitudes, {
    N <- sum(Freq)
    tm <- vmideal_vst_from_magn(magn, lim_magn)
    tm_mean <- sum(Freq * tm) / N
    tm_var <- sum(Freq * (tm - tm_mean)^2) / (N - 1)
    tm_mean_var <- tm_var / N
    list("val" = tm_mean, "var" = tm_mean_var, "sd" = sqrt(tm_mean_var))
})
```

Thus, one obtains the mean and the variance of the mean of `tm`.

```{r, echo=TRUE, results='show'}
print(paste("tm mean:", tm_mean$val))
print(paste("tm var:", tm_mean$var))
```

Using the bootstrap method, it can be assessed whether the mean is normally distributed.



```{r, echo=TRUE, results='show'}
tm_means <- with(magnitudes, {
    N <- sum(Freq)
    tm <- vmideal_vst_from_magn(magn, lim_magn)
    replicate(50000, {
        mean(sample(tm, size = N, replace = TRUE, prob = Freq))
    })
})
```

The graphical representation indicates that this is indeed approximately the case.

```{r, echo=TRUE, results='show'}
with(new.env(), {
    tm_min <- tm_mean$val - 3 * tm_mean$sd
    tm_max <- tm_mean$val + 3 * tm_mean$sd
    tm_means <- subset(tm_means, tm_means > tm_min & tm_means < tm_max)
    brks <- seq(min(tm_means) - 0.02, max(tm_means) + 0.02, by = 0.02)
    hist(tm_means,
        breaks = brks,
        col = "skyblue",
        border = "black",
        main = "Histogram of mean tm",
        xlab = "tm",
        ylab = "count",
        xaxt = "n"
    )
    axis(1, at = seq(round(min(brks), 1), round(max(brks), 1) + 0.1, by = 0.1))
    abline(v = 0, col = "red", lwd = 1)
})
```

A mean value of `0.0` implies that $\psi$ lies at infinity.
Negative values can be interpreted as a kind of “beyond infinity”, which is not meaningful.
There are two possible explanations:

1. The distribution is not ideal, i.e., the observations do not fit the model as described above.
2. Random variation led to this result.

Although the expected value is clearly above `0.0` and can therefore be used, it is more appropriate
in this context to interpret the value as the median.

```{r, echo=TRUE}
lim_magn_mean <- with(magnitudes, {
    N <- sum(Freq)
    sum(Freq * lim_magn) / N
})
print(paste("lim_magn_mean:", lim_magn_mean))
print(paste("mean psi:", vmideal_vst_to_psi(tm_mean$val, lim_magn_mean)))
```

The mean limiting magnitude `lim_magn_mean` is required to estimate the location parameter $\psi$.
In visual meteor observations, different limiting magnitudes usually occur.
In practice, the observed quantity is the difference between $\psi$ and the limiting magnitude
for each observation. Since $\psi$ is unknown, one must assume that the mean `tm` is correlated
with the limiting magnitude.

If this correlation can be described by a simple linear regression, the mean limiting magnitude serves as
a reliable reference point for estimating $\psi$. This is typically the case when the limiting magnitudes
of the individual observations differ only slightly, or $\psi$ is smaller than or equal to
the limiting magnitudes.

By contrast, if $\psi$ is significantly larger than the limiting magnitude, estimation becomes problematic:
$\psi$ effectively tends to infinity, and the observable magnitude distribution approaches
the geometric model of visual meteor magnitudes with a population index of $r \approx 2.5$.

Nevertheless, if all observations share the same limiting magnitude, the procedure yields an exact estimate.

In practice, however, it is preferable to use a confidence interval estimate.
For example, one can estimate that $\psi$ is, with 10 percent probability, not smaller than:

```{r, echo=TRUE}
print(vmideal_vst_to_psi(qnorm(0.90, tm_mean$val, tm_mean$sd), lim_magn_mean))
```

## Residual Analysis

So far, we have operated under the assumption that the real distribution of meteor magnitudes
is exponential and that the perception probabilities are accurate.
We now use the Chi-Square goodness-of-fit test to check whether the observed frequencies match
the expected frequencies. Then, using the estimated parameter, we retrieve the relative
frequencies `p` for each observation and add them to the data frame `magnitudes`:

```{r, echo=TRUE, results='asis'}
psi_mean <- vmideal_vst_to_psi(tm_mean$val, lim_magn_mean)
magnitudes$p <- with(magnitudes, dvmideal(m = magn, lm = lim_magn, psi_mean))
```

We must also consider the probabilities for the magnitude class with the brightest meteors.

```{r, echo=TRUE, results='hide'}
magn_min <- min(magnitudes$magn)
```

The smallest magnitude class `magn_min` is `r magn_min`. In calculating the probabilities,
we assume that the magnitude class `r magn_min` contains meteors that are either brighter
or equally bright as `r magn_min` and thus use the function `pvmideal()` to determine
their probability.

```{r, echo=TRUE, results='asis'}
idx <- magnitudes$magn == magn_min
magnitudes$p[idx] <- with(
    magnitudes[idx, ],
    pvmideal(m = magn + 1L, lm = lim_magn, psi_mean, lower.tail = TRUE)
)
```

This ensures that the probability of observing a meteor of any given magnitude is 100%.
This is known as the normalization condition. Accordingly, the Chi-Square goodness-of-fit test
will fail if this condition is not met.

We now create the contingency table `magnitutes_observed` for the observed meteor magnitudes
and its margin table.

```{r, echo=TRUE}
magnitutes_observed <- xtabs(Freq ~ magn_id + magn, data = magnitudes)
magnitutes_observed_mt <- margin.table(magnitutes_observed, margin = 2)
print(magnitutes_observed_mt)
```

Next, we check which magnitude classes need to be aggregated so that each contains
at least 10 meteors, allowing us to perform a Chi-Square goodness-of-fit test.

The last output shows that meteors of magnitude class `0` or brighter must be combined into
a magnitude class `0-`. Meteors with a brightness less than `4` are grouped here in the
magnitude class `4+`, and a new contingency table magnitudes.observed is created:

```{r, echo=TRUE}
magnitudes$magn[magnitudes$magn <= 0] <- "0-"
magnitudes$magn[magnitudes$magn >= 4] <- "4+"
magnitutes_observed <- xtabs(Freq ~ magn_id + magn, data = magnitudes)
print(margin.table(magnitutes_observed, margin = 2))
```

We now need the corresponding expected relative frequencies

```{r, echo=TRUE}
magnitutes_expected <- xtabs(p ~ magn_id + magn, data = magnitudes)
magnitutes_row_freq <- margin.table(magnitutes_observed, margin = 1)
magnitutes_expected <- sweep(magnitutes_expected, 1, magnitutes_row_freq, `*`)
magnitutes_expected <- magnitutes_expected / sum(magnitutes_expected)
print(sum(magnitudes$Freq) * margin.table(magnitutes_expected, margin = 2))
```

and then carry out the Chi-Square goodness-of-fit test:

```{r, echo=TRUE, results='asis'}
chisq_test_result <- chisq.test(
    x = margin.table(magnitutes_observed, margin = 2),
    p = margin.table(magnitutes_expected, margin = 2)
)
```

As a result, we obtain the p-value:

```{r, echo=TRUE}
chi2_df <- chisq_test_result$parameter - 1
chi2_pval <- pchisq(chisq_test_result$statistic, df = chi2_df, lower.tail = FALSE)
print(chi2_pval)
```

If we set the level of significance at 5 percent, then it is clear that the p-value with
`r unname(chi2_pval)` is greater than 0.05. Thus, under the assumption that the
magnitude distribution follows the ideal meteor magnitude distribution and that
the perception probabilities are correct (i.e., error-free or precisely known),
these assumptions cannot be rejected. However, the converse is not true; the assumptions
may not necessarily be correct. The total count of meteors here is too small for such
a conclusion.

To verify the p-value, we also graphically represent the Pearson residuals:

```{r, fig.show='hold'}
chisq_test_residuals <- with(new.env(), {
    chisq_test_residuals <- residuals(chisq_test_result)
    v <- as.vector(chisq_test_residuals)
    names(v) <- names(chisq_test_residuals)
    v
})

plot(
    chisq_test_residuals,
    main = "Residuals of the chi-square goodness-of-fit test",
    xlab = "m",
    ylab = "Residuals",
    ylim = c(-3, 3),
    xaxt = "n"
)
abline(h = 0.0, lwd = 2)
axis(1, at = seq_along(chisq_test_residuals), labels = names(chisq_test_residuals))
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splines.R
\name{select_knots}
\alias{select_knots}
\title{Greedy stepwise knot selection for a regression spline}
\usage{
select_knots(
  data,
  knot_candidates,
  score_fun,
  backward = FALSE,
  n_steps = NULL,
  bulk_gap = 4L,
  fixed_knots = numeric(0),
  verbose = FALSE,
  n_cores = 1L
)
}
\arguments{
\item{data}{Data passed unchanged to \code{score_fun}; typically a
\code{data.frame}. The function itself does not inspect or mutate it.}

\item{knot_candidates}{Numeric vector of candidate interior-knot positions.
Duplicates and unsorted input are tolerated (sorted/uniq'd internally).}

\item{score_fun}{\code{\\(data, knots) -> numeric scalar}. Lower is
better. The caller is responsible for fitting whatever model they want
and returning a single criterion value (BIC, AIC, cross-validation
error, ...).}

\item{backward}{Logical. \code{FALSE} (default) = forward selection,
\code{TRUE} = backward elimination.}

\item{n_steps}{\code{NULL} (default) or a positive integer.
\code{NULL} runs the greedy search until the next move would not
improve the score and stops there --- so the end state of the search
IS the score optimum (under the current direction and constraints).
A positive integer \code{N} runs exactly \code{N} iterations regardless
of whether each one improves the score; this is a deliberate
exploration mode, intended for inspecting the score landscape just
past a previously-found optimum. The canonical recipe is to first
call with \code{n_steps = NULL} to find the optimum, then re-call with
\code{fixed_knots = prev$knots} and
\code{n_steps = 1L} (or 2L, 3L) to take one or more controlled steps
onward and see how rapidly the score deteriorates. The value
\code{0L} is not allowed (use \code{NULL} for "no steps past the
optimum"); non-integer or negative values error out.}

\item{bulk_gap}{Integer (>= 0). Minimum index gap between knots removed in
the same round of backward elimination. Ignored when
\code{backward = FALSE}. \code{0L} disables bulk removal. The default
\code{4L} matches the support of a cubic spline basis
(\code{degree + 1} for cubic B-splines; the same value also fits
\code{splines::ns}, which is cubic by construction); for higher-degree
B-splines pick \code{degree + 1} accordingly. Note that
\code{bulk_gap = 1L} imposes no real separation constraint and therefore
removes \emph{every} improving knot in one round -- a very aggressive
setting that defeats the per-round verify-fit's role as a safety net.}

\item{fixed_knots}{Numeric vector of knot positions that must be present
in every fitted model during the search. In forward mode they are
set from the start, and further knots may be added on top; in
backward mode they are never proposed for removal, neither singly nor
in bulk. Need not be a subset of \code{knot_candidates} -- included
regardless. Duplicates and unsorted input are tolerated. Default
\code{numeric(0)} (no fixed knots).}

\item{verbose}{Logical. If \code{TRUE}, prints per-round progress
(\code{cat()} to stdout). Default \code{FALSE}.}

\item{n_cores}{Integer (>= 1). Performance-only knob: the per-round
candidate scoring runs in parallel across \code{n_cores} workers. Default
\code{1L} (serial, no extra dependency). When \code{n_cores > 1L} the
base-R package \pkg{parallel} is loaded and \code{parallel::mclapply()}
is used (fork-based on macOS/Linux; falls back to serial on Windows).
Quick recipes:
\itemize{
\item \code{n_cores = 1L} -- safe default, no extra package loaded.
\item \code{n_cores = max(1L, parallel::detectCores() - 1L)} -- use
all cores except one; good default for interactive multi-tasking.
\item \code{n_cores = parallel::detectCores()} -- use every core;
fastest but resource-hungry (the whole machine is busy).
}
The function errors out with a clear message if \code{n_cores > 1L} but
\pkg{parallel} is not installed.

Reproducibility deserves a note when \code{n_cores > 1L}:
\code{mclapply()} is fork-based and inherits the parent RNG state. If
your \code{score_fun} uses randomness (e.g. cross-validation splits),
set \code{RNGkind("L'Ecuyer-CMRG")} and seed via \code{set.seed()}
\emph{before} calling \code{select_knots()} to get reproducible
per-worker streams; otherwise results can differ run to run and across
\code{n_cores}.}
}
\value{
A list with elements (in this order):
\describe{
\item{\code{backward}}{Logical — the direction used.}
\item{\code{knots}}{Sorted numeric vector — the interior knots the
algorithm itself selected, with \code{fixed_knots} \emph{excluded}.
Disjoint from \code{fixed_knots} by construction. The full knot
vector to fit a model on is \code{c(knots, fixed_knots)} (or its
sorted form). With the default \code{n_steps = NULL} this is the
score-optimal selection; with \code{n_steps > 0L} it can be a state
past the optimum.}
\item{\code{fixed_knots}}{Sorted numeric vector — the user-supplied
fixed knots echoed back (deduplicated and sorted). Empty vector if
the caller did not supply any.}
\item{\code{score}}{Numeric — the score at the end state, i.e. at
\code{c(knots, fixed_knots)}.}
\item{\code{n_steps}}{The \code{n_steps} value used (\code{NULL} or a
positive integer).}
\item{\code{history}}{\code{data.frame} of per-round records:
\code{step}, \code{n_knots} (total interior knot count including
\code{fixed_knots}), \code{changed_knot} (the knot added or removed
in that round; \code{NA} for the initial state and for bulk-removal
rounds), \code{score}, \code{extra} (boolean: \code{TRUE} when that
step worsened the score). For \code{n_steps = NULL} runs no
worsening step is taken, so \code{extra} is always \code{FALSE}.}
}
}
\description{
Selects a parsimonious subset of \code{knot_candidates} as interior knots
for a regression spline by greedy forward addition or backward elimination,
scored by a user-supplied criterion (BIC, AIC, or any cross-validation
function). The function is model-agnostic: it only chooses which
knot positions to include and passes that vector to your \code{score_fun};
your \code{score_fun} decides which basis (\code{splines::ns},
\code{splines::bs} of any degree, …) and which model family to fit.
Typical use: fitting a smooth, parsimonious trend through noisy data — for
example an activity profile of a meteor shower across solar longitude —
where the number and position of knots should remain small and inspectable.
}
\details{
The typical use case --- meteor shower rates over the solar longitude or
time, for example --- is a smooth, slowly varying signal whose shape is
not known in advance and whose curvature changes locally: too rigid for
a low-order polynomial, too noisy for a histogram. A regression spline
(typically cubic, e.g. \code{splines::ns} or
\code{splines::bs(degree = 3)}) with a modest number of well-placed
interior knots gives a smooth, locally flexible fit with interpretable
degrees of freedom. Picking that number is a bias/variance trade-off:
too many knots overfit (ringing, unstable derivatives, slow fits), too
few introduce bias. \code{select_knots()} automates the trade-off by
greedy forward addition or backward elimination, scored by a
user-defined criterion; you control which knots are even allowed
(\code{knot_candidates}) and what counts as "better" (\code{score_fun}).

Your \code{score_fun} must take \code{(data, knots)} and return a single
\code{numeric} value; \emph{lower is better}. Typically it fits a model
with these interior knots (\code{length(knots) == 0L} means "no interior
knots") and reports an information criterion (\code{stats::BIC},
\code{stats::AIC}) or a held-out score. \code{knots} arrives sorted, so
the fit code does not need to sort it again. A divergent or failed fit
should ideally return \code{Inf}; the function additionally wraps each
call in \code{tryCatch()} and treats errors as \code{Inf} so the search
continues robustly. See \code{Examples} for a runnable template.

Backward elimination can drop multiple knots per round in "bulk" mode.
When \code{backward = TRUE} and \code{bulk_gap >= 1L}, each round removes
several well-separated knots at once (minimum index gap \code{bulk_gap}
in the current sorted knot list). For a B-spline basis of degree
\code{d}, each basis function has support over \code{d + 1} consecutive
knot intervals, so removing knots that are at least \code{d + 1}
positions apart has nearly additive effect on the score --- giving
roughly a \code{bulk_gap}-fold speed-up at the cost of a small
approximation (mitigated by a verify-fit after each bulk round). The
default \code{4L} matches \code{d + 1} for cubic bases such as
\code{splines::ns} or \code{splines::bs(degree = 3)}; for other bases
pick \code{bulk_gap = degree + 1}, or set \code{bulk_gap = 0L} for
strict one-knot-per-round behaviour.

The result splits the final knot set into two disjoint vectors:
\code{knots} (positions the algorithm itself selected) and
\code{fixed_knots} (the user-supplied anchors, echoed back, deduplicated
and sorted). The full vector to fit a model on is
\code{c(knots, fixed_knots)} (or its sorted form). With the default
\code{n_steps = NULL} the loop stops as soon as no further move improves
the score, so the end state \emph{is} the score-optimum; with
\code{n_steps > 0L} the loop runs that many iterations regardless of
improvement, and the score-best point along the trajectory is recoverable
from \code{history} via the row at \code{which.min(history$score)}.

If the starting state already is (locally) optimal --- typical when
\code{fixed_knots} alone overfits in forward mode, or when
\code{knot_candidates} is so small that the full pool already minimises
the score in backward mode --- the \code{n_steps = NULL} run terminates
immediately with \code{history} containing only the initial row and
\code{score} equal to the starting score; an explicit \code{n_steps = N}
run will instead take \code{N} worsening steps so the post-optimum
landscape is still observable.

Knots sit next to extrema, not on them. \code{select_knots()} chooses
knots for a \emph{good fit}, not for detecting extrema of the response.
Knots typically land next to peaks or troughs --- where the curvature is
highest --- and not on them. Read local extrema off the \emph{shape} of
the fitted curve, not from \code{knots}. A knot supplied
through \code{fixed_knots} is the exception: it sits wherever the user
puts it, since it is a constraint imposed on the search rather than a
finding produced by it.
}
\section{When to use this}{

\code{select_knots()} is for situations where knot \emph{positions} are
meaningful and you want a small, inspectable set of interior knots scored
under a criterion you choose. If you instead want a continuous roughness
penalty over a dense knot grid, the \pkg{mgcv} package's penalised
B-splines (\code{bs = "ps"}) or adaptive smoothers (\code{bs = "ad"}) are
usually a better fit; for greedy stepwise selection on a hinge-function
basis, see the \pkg{earth} package.

Like every greedy stepwise procedure, \code{select_knots()} performs a
\emph{local} search: in each round it commits to the locally best
add/drop. It therefore returns a local optimum of the score, which is
usually but not provably the global optimum --- a knot \emph{combination}
that beats every individually-best move can be unreachable once an
earlier, locally-attractive knot has been picked. The only way to
guarantee globality would be exhaustive enumeration over all
\eqn{2^{|knot\_candidates|}} subsets, which is exponential and
impractical for any realistic candidate pool. In practice the greedy
optimum is close; repeating the search in the opposite direction
(\code{backward = TRUE}) is a cheap robustness check.
}

\examples{
\dontrun{
# Greedy knot selection on a simple synthetic signal.
set.seed(1)
n <- 200
x <- seq(0, 10, length.out = n)
y <- stats::rpois(n, lambda = 50 + 30 * sin(x))
dat <- data.frame(x = x, y = y)

# score_fun: fit a Poisson GLM with a natural cubic spline, return BIC.
# Lower is better.
fit <- \(d, knots) {
    f <- if (length(knots) == 0L) {
        y ~ splines::ns(x)
    } else {
        y ~ splines::ns(x, knots = knots)
    }
    stats::glm(f, family = stats::poisson(), data = d)
}
score_bic <- \(d, knots) stats::BIC(fit(d, knots))

cand <- seq(1, 9, by = 0.5)
res <- select_knots(dat, cand, score_bic, verbose = TRUE)
# Full knot vector to fit the final model on:
final_knots <- sort(c(res$knots, res$fixed_knots))
}

}
\seealso{
\code{\link[splines]{ns}}, \code{\link[splines]{bs}},
\code{\link[stats]{BIC}}, \code{\link[stats]{AIC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mideal.R
\name{mideal}
\alias{mideal}
\alias{dmideal}
\alias{pmideal}
\alias{qmideal}
\alias{rmideal}
\title{Ideal Distribution of Meteor Magnitudes}
\usage{
dmideal(m, psi = 0, log = FALSE)

pmideal(m, psi = 0, lower.tail = TRUE, log = FALSE)

qmideal(p, psi = 0, lower.tail = TRUE)

rmideal(n, psi = 0)
}
\arguments{
\item{m}{numeric; meteor magnitude.}

\item{psi}{numeric; the location parameter of a probability distribution.
It is the only parameter of the distribution.}

\item{log}{logical; if \code{TRUE}, probabilities p are given as \code{log(p)}.}

\item{lower.tail}{logical; if \code{TRUE} (default) probabilities are
\eqn{P[M \le m]}, otherwise, \eqn{P[M > m]}.}

\item{p}{numeric; probability.}

\item{n}{numeric; count of meteor magnitudes.}
}
\value{
\code{dmideal} gives the density, \code{pmideal} gives the distribution function,
\code{qmideal} gives the quantile function and \code{rmideal} generates random deviates.

The length of the result is determined by \code{n} for \code{rmideal}, and is the maximum
of the lengths of the numerical vector arguments for the other functions.

\code{qmideal} can return \code{NaN} value with a warning.
}
\description{
Density, distribution function, quantile function and random generation
for the ideal distribution of meteor magnitudes.
}
\details{
The density of the ideal distribution of meteor magnitudes is
\deqn{
    {\displaystyle \frac{\mathrm{d}p}{\mathrm{d}m} =
     \frac{3}{2} \, \log(r) \sqrt{\frac{r^{3 \, \psi + 2 \, m}}{(r^\psi + r^m)^5}}}
}
where \eqn{m} is the continuous (real-valued) meteor magnitude,
\eqn{r = 10^{0.4} \approx 2.51189 \dots} is a constant and
\eqn{\psi} is the only parameter of this magnitude distribution.
}
\examples{
old_par <- par(mfrow = c(2, 2))
psi <- 5.0
plot(
    \(m) dmideal(m, psi, log = FALSE),
    -5, 10,
    main = paste0("Density of the Ideal Meteor Magnitude\nDistribution (psi = ", psi, ")"),
    col = "blue",
    xlab = "m",
    ylab = "dp/dm"
)
abline(v = psi, col = "red")

plot(
    \(m) dmideal(m, psi, log = TRUE),
    -5, 10,
    main = paste0("Density of the Ideal Meteor Magnitude\nDistribution (psi = ", psi, ")"),
    col = "blue",
    xlab = "m",
    ylab = "log( dp/dm )"
)
abline(v = psi, col = "red")

plot(
    \(m) pmideal(m, psi),
    -5, 10,
    main = paste0("Probability of the Ideal Meteor Magnitude\nDistribution (psi = ", psi, ")"),
    col = "blue",
    xlab = "m",
    ylab = "p"
)
abline(v = psi, col = "red")

plot(
    \(p) qmideal(p, psi),
    0.01, 0.99,
    main = paste("Quantile of the Ideal Meteor Magnitude\nDistribution (psi = ", psi, ")"),
    col = "blue",
    xlab = "p",
    ylab = "m"
)
abline(h = psi, col = "red")

# generate random meteor magnitudes
m <- rmideal(1000, psi)

# log likelihood function
llr <- function(psi) {
    -sum(dmideal(m, psi, log = TRUE))
}

# maximum likelihood estimation (MLE) of psi
est <- optim(2, llr, method = "Brent", lower = 0, upper = 8, hessian = TRUE)

# estimations
est$par # mean of psi
sqrt(1 / est$hessian[1][1]) # standard deviation of psi

par(old_par)
}
\references{
Richter, J. (2018) \emph{About the mass and magnitude distributions of meteor showers}.
WGN, Journal of the International Meteor Organization, vol. 46, no. 1, p. 34-38
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/freq.R
\name{freq_quantile}
\alias{freq_quantile}
\title{Quantiles with a minimum frequency}
\usage{
freq_quantile(freq, min)
}
\arguments{
\item{freq}{integer; A vector of frequencies.}

\item{min}{integer; Minimum total frequency per quantile.}
}
\value{
A factor of indices is returned.
The index references the corresponding passed frequency \code{freq}.
}
\description{
This function generates quantiles with a minimum frequency.
These quantiles are formed from a vector \code{freq} of frequencies.
Each quantile then has the minimum total frequency \code{min}.
}
\details{
The frequencies \code{freq} are grouped in the order in which they
are passed as a vector.
The minimum \code{min} must be greater than \code{0}.
}
\examples{
freq <- c(1, 2, 3, 4, 5, 6, 7, 8, 9)
cumsum(freq)
(f <- freq_quantile(freq, 10))
tapply(freq, f, sum)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vmgeom_vst.R
\name{vmgeom_vst}
\alias{vmgeom_vst}
\alias{vmgeom_vst_from_magn}
\alias{vmgeom_vst_to_r}
\title{Variance-Stabilizing Transformation for Geometric Visual Meteor Magnitudes}
\usage{
vmgeom_vst_from_magn(m, lm)

vmgeom_vst_to_r(tm, log = FALSE, deriv_degree = 0L)
}
\arguments{
\item{m}{integer; meteor magnitude.}

\item{lm}{numeric; limiting magnitude.}

\item{tm}{numeric; transformed magnitude.}

\item{log}{logical; if \code{TRUE}, the logarithm of the population index \code{r} is returned.}

\item{deriv_degree}{integer; the order of the derivative at \code{tm} to return
instead of \code{r} or \code{log(r)}. Must be \code{0}, \code{1}, or \code{2}.}
}
\value{
\itemize{
\item \code{vmgeom_vst_from_magn}: numeric value, the transformed meteor magnitude.
\item \code{vmgeom_vst_to_r}: numeric value of the population index \code{r}, derived from
the mean of \code{tm}.
}

The argument \code{deriv_degree} can be used to apply the delta method.
If \code{log = TRUE}, the logarithm of \code{r} is returned.
}
\description{
Applies a variance-stabilizing transformation to visual meteor magnitudes
under the geometric model.
}
\details{
Many linear models require the variance of visual meteor magnitudes to be
homoscedastic. The function \code{vmgeom_vst_from_magn} applies a transformation
that produces homoscedastic distributions of visual meteor magnitudes if the
underlying distribution follows a geometric model.

The geometric model of visual meteor magnitudes
depends on the \link[=vmgeom]{population index} \code{r} and the limiting magnitude \code{lm},
resulting in a two-parameter distribution. Without detection probabilities,
the magnitude distribution is purely geometric, and for integer limiting
magnitudes the variance depends only on the population index \code{r}. Since the
limiting magnitude \code{lm} is a fixed parameter and never estimated
statistically, the magnitudes can be transformed such that, for example,
the mean of the transformed magnitudes directly provides an estimate of \code{r}
using the function \code{vmgeom_vst_to_r}.

A key advantage of this transformation is that the limiting magnitude \code{lm}
is already incorporated into subsequent analyses. In this sense, the
transformation acts as a normalization of meteor magnitudes and yields a
variance close to \code{1.0}.

This transformation is valid for \eqn{1.4 \le r \le 3.5}.
The numerical form of the transformation is version-specific and may change
substantially in future releases. Do not rely on equality of transformed
values across package versions.
}
\note{
The internal approximations used here are derived from the perception
probabilities produced by \link{vmperception}.
For details on the derivation, see the script \code{inst/derivation/vmgeom_vst.R} in the
package's source code.
}
\examples{
N <- 100
r <- 2.0
limmag <- 6.3

# Simulate magnitudes
m <- rvmgeom(N, limmag, r)

# Variance-stabilizing transformation
tm <- vmgeom_vst_from_magn(m, limmag)
tm_mean <- mean(tm)
tm_var <- var(tm)

# Estimator for r from the transformed mean
r_hat <- vmgeom_vst_to_r(tm_mean)

# Derivative dr/d(tm) at tm_mean (needed for the delta method)
dr_dtm <- vmgeom_vst_to_r(tm_mean, deriv_degree = 1L)

# Variance of the sample mean of tm
var_tm.mean <- tm_var / N

# Delta method: variance and standard error of r_hat
var_r.hat <- (dr_dtm^2) * var_tm.mean
se_r.hat <- sqrt(var_r.hat)

# Results
print(r_hat)
print(se_r.hat)
}
\seealso{
\link{vmgeom} \link{vmperception}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vmideal_vst.R
\name{vmidealVst}
\alias{vmidealVst}
\alias{vmidealVstFromMagn}
\alias{vmidealVstToPsi}
\title{Variance-stabilizing Transformation for the Ideal Distribution of Visual Meteor Magnitudes}
\usage{
vmidealVstFromMagn(m, lm)

vmidealVstToPsi(tm, lm, deriv.degree = 0L)
}
\arguments{
\item{m}{integer; the meteor magnitude.}

\item{lm}{numeric; limiting magnitude.}

\item{tm}{numeric; transformed magnitude.}

\item{deriv.degree}{integer; the degree of the derivative at \code{tm} to return
instead of \code{psi}. Must be \code{0}, \code{1} or \code{2}.}
}
\value{
\itemize{
\item \code{vmidealVstFromMagn}: a numeric value, the transformed meteor magnitude.
\item \code{vmidealVstToPsi}: a numeric value of the parameter \code{psi}, derived from
the mean of \code{tm}.
The argument \code{deriv.degree} can be used to apply the delta method.
}
}
\description{
Applies a variance-stabilizing transformation to meteor magnitudes
under the assumption of the ideal magnitude distribution.
}
\details{
Many linear models require the variance of visual meteor magnitudes to be
homoscedastic. The function \code{vmidealVstFromMagn} applies a transformation
that produces homoscedastic distributions of visual meteor magnitudes if the
underlying magnitudes follow the ideal magnitude distribution.
In this sense, the transformation acts as a normalization of meteor magnitudes
and yields a variance close to \code{1.0}.

The ideal distribution of visual meteor magnitudes
depends on the \link[=vmideal]{parameter psi} and the limiting magnitude \code{lm},
resulting in a two-parameter distribution. Without detection probabilities,
the magnitude distribution reduces to a pure \link[=mideal]{ideal magnitude distribution},
which depends only on the parameter \code{psi}. Since the
limiting magnitude \code{lm} is a fixed parameter and never estimated
statistically, the magnitudes can be transformed such that, for example,
the mean of the transformed magnitudes directly provides an estimate of \code{psi}
using the function \code{vmidealVstToPsi}.

This transformation is valid for \eqn{-10 \le \texttt{psi} \le 9}.
The numerical form of the transformation is version-specific and may change
substantially in future releases. Do not rely on equality of transformed
values across package versions.
}
\note{
The internal approximations used here are derived from the perception
probabilities produced by \link{vmperception}.
For details on the derivation, see the script \code{inst/derivation/vmideal_vst.R} in the
package's source code.
}
\examples{
N <- 100
psi <- 5.0
limmag <- 6.3

# Simulate magnitudes
m <- rvmideal(N, limmag, psi)

# Variance-stabilizing transformation
tm <- vmidealVstFromMagn(m, limmag)
tm.mean <- mean(tm)
tm.var  <- var(tm)

# Estimator for psi from the transformed mean
psi.hat  <- vmidealVstToPsi(tm.mean, limmag)

# Derivative d(psi)/d(tm) at tm.mean (needed for the delta method)
dpsi_dtm <- vmidealVstToPsi(tm.mean, limmag, deriv.degree = 1L)

# Variance of the sample mean of tm
var_tm.mean <- tm.var / N

# Delta method: variance and standard error of psi.hat
var_psi.hat <- (dpsi_dtm^2) * var_tm.mean
se_psi.hat  <- sqrt(var_psi.hat)

# Results
print(psi.hat)
print(se_psi.hat)
}
\seealso{
\link{vmideal} \link{mideal} \link{vmperception}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{PER_2015_magn}
\alias{PER_2015_magn}
\title{Visual magnitude observations of Perseids from 2015}
\format{
A list with the same structure as returned by \link{load_vmdb_magnitudes}.
}
\description{
Visual magnitude observations of the Perseid shower from 2015.
}
\details{
\code{PER_2015_magn} are magnitude observations loaded with
\link{load_vmdb_magnitudes}.
}
\seealso{
\link{load_vmdb}
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/freq.R
\name{vmtable}
\alias{vmtable}
\title{Rounds a contingency table of meteor magnitude frequencies}
\usage{
vmtable(mt)
}
\arguments{
\item{mt}{table; A two-dimensional contingency table of meteor magnitude frequencies.}
}
\value{
A rounded contingency table of meteor magnitudes is returned.
}
\description{
The meteor magnitude contingency table of VMDB contains half meteor counts (e.g. \code{3.5}).
This function converts these frequencies to integer values.
}
\details{
The contingency table of meteor magnitudes \code{mt} must be two-dimensional.
The row names refer to the magnitude observations.
Column names must be integer meteor magnitude values.
Also, the columns must be sorted in ascending or descending order of meteor magnitude.

A sum-preserving algorithm is used for rounding.
It ensures that the total frequency of meteors per observation is preserved.
The marginal frequencies of the magnitudes are also preserved with
the restriction that the deviation is at most \eqn{\pm 0.5}.
If the total sum of a meteor magnitude is integer,
then the deviation is \eqn{\pm 0}.

The algorithm is unbiased: for a fixed observation order it preserves the original
totals without introducing systematic drift, even though each run follows the
deterministic sequence dictated by the observed counts and their ordering.
}
\note{
Internally the counts are doubled to half-meteor units, leftover halves are
alternated between rows so column margins stay within \eqn{\pm 0.5}, and when the
grand total is odd the matrix is temporarily mirrored so the unavoidable surplus
meteor originates from the opposite end of the magnitude scale rather than always
favouring the faintest bin. The mirroring is only the initial condition; the loop
then processes the table cell by cell so the rounding direction alternates between
bright and faint magnitudes depending on the current row and column state.
}
\examples{
# For example, create a contingency table of meteor magnitudes
mt <- as.table(matrix(
    c(
        0.0, 0.0, 2.5, 0.5, 0.0, 1.0,
        0.0, 1.5, 2.0, 0.5, 0.0, 0.0,
        1.0, 0.0, 0.0, 3.0, 2.5, 0.5
    ),
    nrow = 3, ncol = 6, byrow = TRUE
))
colnames(mt) <- seq(6)
rownames(mt) <- c("A", "B", "C")
mt
margin.table(mt, 1)
margin.table(mt, 2)

# contingency table with integer values
(mt_int <- vmtable(mt))
margin.table(mt_int, 1)
margin.table(mt_int, 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_data.R
\name{load_vmdb}
\alias{load_vmdb}
\alias{load_vmdb_rates}
\alias{load_vmdb_magnitudes}
\title{Loading visual meteor observations via the imo-vmdb REST API}
\usage{
load_vmdb_rates(
  base_url,
  shower = NULL,
  period = NULL,
  sl = NULL,
  lim_magn = NULL,
  sun_alt_max = NULL,
  moon_alt_max = NULL,
  session_id = NULL,
  rate_id = NULL,
  with_sessions = FALSE,
  with_magnitudes = FALSE
)

load_vmdb_magnitudes(
  base_url,
  shower = NULL,
  period = NULL,
  sl = NULL,
  lim_magn = NULL,
  session_id = NULL,
  magn_id = NULL,
  with_sessions = FALSE,
  with_magnitudes = TRUE
)
}
\arguments{
\item{base_url}{character; base URL of the imo-vmdb API, e.g.
\code{"http://localhost:8000/api/v1"}.}

\item{shower}{character; selects by meteor shower codes.
\code{NA} loads sporadic meteors.}

\item{period}{selects an observation time range by minimum/maximum.
Accepts \code{POSIXct}, \code{Date}, or character.  Character
elements may be \code{"YYYY-MM-DDTHH:MM:SS"},
\code{"YYYY-MM-DD HH:MM:SS"}, or date-only \code{"YYYY-MM-DD"}.
Date-only inputs are expanded to midnight (lower bound) and
23:59:59 (upper bound) of the given day, in UTC.  IMO data is UTC
by convention; the timezone marker is omitted on the wire.}

\item{sl}{numeric; selects a range of solar longitudes by minimum/maximum.}

\item{lim_magn}{numeric; selects a range of limiting magnitudes by
minimum/maximum.}

\item{sun_alt_max}{numeric; selects the maximum altitude of the sun
(rates only).}

\item{moon_alt_max}{numeric; selects the maximum altitude of the moon
(rates only).}

\item{session_id}{integer; selects by session ids.}

\item{rate_id}{integer; selects rate observations by ids.}

\item{with_sessions}{logical; if \code{TRUE}, also load the corresponding
session data.}

\item{with_magnitudes}{logical; if \code{TRUE}, also load the corresponding
magnitude observations.}

\item{magn_id}{integer; selects magnitude observations by ids.}
}
\value{
Both functions return a list, with

\tabular{ll}{
\code{observations} \tab data frame, rate or magnitude observations,\cr
\code{sessions} \tab data frame; session data of observations,\cr
\code{magnitudes} \tab table; contingency table of meteor magnitude
frequencies.
}

\code{observations} depends on the function call.
\code{load_vmdb_rates} returns a data frame with columns:

\tabular{ll}{
\code{rate_id} \tab unique identifier of the rate observation,\cr
\code{shower} \tab IAU code of the shower. \code{NA} for sporadic.\cr
\code{period_start} \tab \code{POSIXct} (UTC); start of observation,\cr
\code{period_end} \tab \code{POSIXct} (UTC); end of observation,\cr
\code{sl_start} \tab solar longitude at start,\cr
\code{sl_end} \tab solar longitude at end,\cr
\code{session_id} \tab reference to the session,\cr
\code{freq} \tab count of observed meteors,\cr
\code{lim_magn} \tab limiting magnitude,\cr
\code{t_eff} \tab net observed time in hours,\cr
\code{f} \tab correction factor of cloud cover,\cr
\code{sidereal_time} \tab sidereal time,\cr
\code{sun_alt} \tab altitude of the sun,\cr
\code{sun_az} \tab azimuth of the sun,\cr
\code{moon_alt} \tab altitude of the moon,\cr
\code{moon_az} \tab azimuth of the moon,\cr
\code{moon_illum} \tab illumination of the moon (\code{0.0 .. 1.0}),\cr
\code{field_alt} \tab altitude of the field of view (optional),\cr
\code{field_az} \tab azimuth of the field of view (optional),\cr
\code{rad_alt} \tab altitude of the radiant (optional),\cr
\code{rad_az} \tab azimuth of the radiant (optional),\cr
\code{magn_id} \tab reference to the magnitude observations (optional),\cr
\code{magn_solo} \tab \code{TRUE} if this rate is the sole contributor
to its linked magnitude observation; \code{FALSE} if the magnitude
aggregates this rate with others; \code{NA} when \code{magn_id} is
\code{NA} (optional).
}

\code{load_vmdb_magnitudes} returns an \code{observations} data frame with:

\tabular{ll}{
\code{magn_id} \tab unique identifier of the magnitude observation,\cr
\code{shower} \tab IAU code of the shower. \code{NA} for sporadic.\cr
\code{period_start} \tab \code{POSIXct} (UTC); start of observation,\cr
\code{period_end} \tab \code{POSIXct} (UTC); end of observation,\cr
\code{sl_start} \tab solar longitude at start,\cr
\code{sl_end} \tab solar longitude at end,\cr
\code{session_id} \tab reference to the session,\cr
\code{freq} \tab count of observed meteors,\cr
\code{magn_mean} \tab mean magnitude,\cr
\code{lim_magn} \tab limiting magnitude (optional).
}

The \code{sessions} data frame contains

\tabular{ll}{
\code{session_id} \tab unique identifier of the session,\cr
\code{longitude} \tab location's longitude,\cr
\code{latitude} \tab location's latitude,\cr
\code{elevation} \tab height above mean sea level in km,\cr
\code{country} \tab country name,\cr
\code{location_name} \tab location name,\cr
\code{observer_id} \tab observer id (optional),\cr
\code{observer_name} \tab observer name (optional).
}

\code{magnitudes} is a contingency table of meteor magnitude frequencies.
Row names are magnitude observation IDs; column names are magnitude classes.
}
\description{
Loads visual meteor observations from an
\href{https://pypi.org/project/imo-vmdb/}{imo-vmdb} web server via its
REST API.
}
\details{
\code{sl}, \code{period} and \code{lim_magn} expect a vector with
successive minimum and maximum values.
\code{sun_alt_max} and \code{moon_alt_max} are expected to be scalar values.

\strong{Note:} Unlike the previous DBI-based version, only a single range
per filter parameter is supported.  If you previously passed a matrix with
multiple rows to \code{period}, \code{sl}, or \code{lim_magn}, flatten
them to a single min/max pair or issue multiple calls and combine with
\code{rbind()}.
}
\note{
Angle values are expected and returned in degrees.
}
\examples{
\dontrun{
# Load rate observations including session and magnitude data
data <- load_vmdb_rates(
    base_url       = "http://localhost:8000/api/v1",
    shower         = "PER",
    sl             = c(135.5, 145.5),
    period         = c("2015-08-01", "2015-08-31"),
    lim_magn       = c(5.3, 6.7),
    with_magnitudes = TRUE,
    with_sessions   = TRUE
)

# Load magnitude observations
data <- load_vmdb_magnitudes(
    base_url     = "http://localhost:8000/api/v1",
    shower       = "PER",
    sl           = c(135.5, 145.5),
    period       = c("2015-08-01", "2015-08-31"),
    lim_magn     = c(5.3, 6.7),
    with_sessions = TRUE
)

# Period can also be given as POSIXct or as full ISO 8601 datetime
# strings — useful for narrowing to a specific time window within a day.
data <- load_vmdb_rates(
    base_url = "http://localhost:8000/api/v1",
    shower   = "PER",
    period   = c("2015-08-12T20:00:00", "2015-08-13T04:00:00")
)
}
}
\references{
\url{https://pypi.org/project/imo-vmdb/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vmideal_vst.R
\name{vmideal_vst}
\alias{vmideal_vst}
\alias{vmideal_vst_from_magn}
\alias{vmideal_vst_to_psi}
\title{Variance-stabilizing Transformation for the Ideal Distribution of Visual Meteor Magnitudes}
\usage{
vmideal_vst_from_magn(m, lm)

vmideal_vst_to_psi(tm, lm, deriv_degree = 0L)
}
\arguments{
\item{m}{integer; the meteor magnitude.}

\item{lm}{numeric; limiting magnitude.}

\item{tm}{numeric; transformed magnitude.}

\item{deriv_degree}{integer; the degree of the derivative at \code{tm} to return
instead of \code{psi}. Must be \code{0}, \code{1} or \code{2}.}
}
\value{
\itemize{
\item \code{vmideal_vst_from_magn}: a numeric value, the transformed meteor magnitude.
\item \code{vmideal_vst_to_psi}: a numeric value of the parameter \code{psi}, derived from
the mean of \code{tm}.
The argument \code{deriv_degree} can be used to apply the delta method.
}
}
\description{
Applies a variance-stabilizing transformation to meteor magnitudes
under the assumption of the ideal magnitude distribution.
}
\details{
Many linear models require the variance of visual meteor magnitudes to be
homoscedastic. The function \code{vmideal_vst_from_magn} applies a transformation
that produces homoscedastic distributions of visual meteor magnitudes if the
underlying magnitudes follow the ideal magnitude distribution.
In this sense, the transformation acts as a normalization of meteor magnitudes
and yields a variance close to \code{1.0}.

The ideal distribution of visual meteor magnitudes
depends on the \link[=vmideal]{parameter psi} and the limiting magnitude \code{lm},
resulting in a two-parameter distribution. Without detection probabilities,
the magnitude distribution reduces to a pure \link[=mideal]{ideal magnitude distribution},
which depends only on the parameter \code{psi}. Since the
limiting magnitude \code{lm} is a fixed parameter and never estimated
statistically, the magnitudes can be transformed such that, for example,
the mean of the transformed magnitudes directly provides an estimate of \code{psi}
using the function \code{vmideal_vst_to_psi}.

This transformation is valid for \eqn{-10 \le \texttt{psi} \le 9}.
The numerical form of the transformation is version-specific and may change
substantially in future releases. Do not rely on equality of transformed
values across package versions.
}
\note{
The internal approximations used here are derived from the perception
probabilities produced by \link{vmperception}.
For details on the derivation, see the script \code{inst/derivation/vmideal_vst.R} in the
package's source code.
}
\examples{
N <- 100
psi <- 5.0
limmag <- 6.3

# Simulate magnitudes
m <- rvmideal(N, limmag, psi)

# Variance-stabilizing transformation
tm <- vmideal_vst_from_magn(m, limmag)
tm_mean <- mean(tm)
tm_var <- var(tm)

# Estimator for psi from the transformed mean
psi_hat <- vmideal_vst_to_psi(tm_mean, limmag)

# Derivative d(psi)/d(tm) at tm_mean (needed for the delta method)
dpsi_dtm <- vmideal_vst_to_psi(tm_mean, limmag, deriv_degree = 1L)

# Variance of the sample mean of tm
var_tm.mean <- tm_var / N

# Delta method: variance and standard error of psi_hat
var_psi.hat <- (dpsi_dtm^2) * var_tm.mean
se_psi.hat <- sqrt(var_psi.hat)

# Results
print(psi_hat)
print(se_psi.hat)
}
\seealso{
\link{vmideal} \link{mideal} \link{vmperception}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vmideal.R
\name{vmideal}
\alias{vmideal}
\alias{dvmideal}
\alias{pvmideal}
\alias{qvmideal}
\alias{rvmideal}
\alias{cvmideal}
\title{Ideal Distribution of Visual Meteor Magnitudes}
\usage{
dvmideal(m, lm, psi, log = FALSE, perception_fun = vmperception)

pvmideal(
  m,
  lm,
  psi,
  lower.tail = TRUE,
  log = FALSE,
  perception_fun = vmperception
)

qvmideal(p, lm, psi, lower.tail = TRUE, perception_fun = vmperception)

rvmideal(n, lm, psi, perception_fun = vmperception)

cvmideal(lm, psi, log = FALSE, perception_fun = vmperception)
}
\arguments{
\item{m}{integer; visual meteor magnitude.}

\item{lm}{numeric; limiting magnitude.}

\item{psi}{numeric; the location parameter of the probability distribution.}

\item{log}{logical; if \code{TRUE}, probabilities are returned as \code{log(p)}.}

\item{perception_fun}{function; optional perception probability function.
The default is \link{vmperception}.}

\item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
\eqn{P[M < m]}; otherwise, \eqn{P[M \ge m]}.}

\item{p}{numeric; probability.}

\item{n}{numeric; count of meteor magnitudes.}
}
\value{
\itemize{
\item \code{dvmideal}: density
\item \code{pvmideal}: distribution function
\item \code{qvmideal}: quantile function
\item \code{rvmideal}: random generation
\item \code{cvmideal}: partial convolution of the ideal distribution of meteor magnitudes
with the perception probabilities.
}

The length of the result is determined by \code{n} for \code{rvmideal}, and is the maximum
of the lengths of the numeric vector arguments for the other functions.
All arguments are vectorized; standard R recycling rules apply.

Since the distribution is discrete, \code{qvmideal} and \code{rvmideal} always return integer values.
\code{qvmideal} may return \code{NaN} with a warning.
}
\description{
Density, distribution function, quantile function, and random generation
for the ideal distribution of visual meteor magnitudes.
}
\details{
The density of the \link[=mideal]{ideal distribution of meteor magnitudes} is
\deqn{
    {\displaystyle f(m) = \frac{\mathrm{d}p}{\mathrm{d}m} =
     \frac{3}{2} \, \log(r) \sqrt{\frac{r^{3 \, \psi + 2 \, m}}{(r^\psi + r^m)^5}}}
}
where \eqn{m} denotes the continuous (real-valued) meteor magnitude,
\eqn{r = 10^{0.4} \approx 2.51189 \dots} is a constant, and
\eqn{\psi} is the only parameter of this magnitude distribution.

In visual meteor observations, magnitudes are usually estimated as integer values.
Hence, this distribution is discrete and its probability mass function is given by
\deqn{
P[M = m] \sim
\begin{cases}
  g(m_{\mathrm{lim}} - m) \displaystyle \int\limits_{m-0.5}^{m+0.5} f(u) \, \mathrm{d}u,
  & \text{if } m_{\mathrm{lim}} - m > -0.5,\\[5pt]
  0 & \text{otherwise,}
\end{cases}
}
where \eqn{m_{\mathrm{lim}}} denotes the limiting (non-integer) magnitude of the observation,
and \eqn{m} the integer meteor magnitude.
The function \eqn{f(\cdot)} is the continuous density of the ideal magnitude distribution,
and \eqn{g(\cdot)} denotes the \link[=vmperception]{perception probability function}.

If a different perception probability function \code{perception_fun} is supplied,
it must have the signature \verb{function(x)} and return the perception probabilities of
the difference \code{x} between the limiting magnitude and the meteor magnitude.
If \code{x >= 15.0}, the \code{perception_fun} function should return a perception probability of \code{1.0}.
The argument \code{perception_fun} is resolved using \link{match.fun}.
}
\examples{
N <- 100
psi <- 5.0
limmag <- 6.5
(m <- seq(6, -4))

# discrete density of `N` meteor magnitudes
(freq <- round(N * dvmideal(m, limmag, psi)))

# log likelihood function
lld <- function(psi) {
    -sum(freq * dvmideal(m, limmag, psi, log = TRUE))
}

# maximum likelihood estimation (MLE) of psi
est <- optim(2, lld, method = "Brent", lower = 0, upper = 8, hessian = TRUE)

# estimations
est$par # mean of psi

# generate random meteor magnitudes
m <- rvmideal(N, limmag, psi)

# log likelihood function
llr <- function(psi) {
    -sum(dvmideal(m, limmag, psi, log = TRUE))
}

# maximum likelihood estimation (MLE) of psi
est <- optim(2, llr, method = "Brent", lower = 0, upper = 8, hessian = TRUE)

# estimations
est$par # mean of psi
sqrt(1 / est$hessian[1][1]) # standard deviation of psi

m <- seq(6, -4, -1)
p <- vismeteor::dvmideal(m, limmag, psi)
barplot(
    p,
    names.arg = m,
    main = paste0("Density (psi = ", psi, ", limmag = ", limmag, ")"),
    col = "blue",
    xlab = "m",
    ylab = "p",
    border = "blue",
    space = 0.5
)
axis(side = 2, at = pretty(p))

plot(
    \(lm) vismeteor::cvmideal(lm, psi, log = TRUE),
    -5, 10,
    main = paste0(
        "Partial convolution of the ideal meteor magnitude distribution\n",
        "with the perception probabilities (psi = ", psi, ")"
    ),
    col = "blue",
    xlab = "lm",
    ylab = "log(rate)"
)
}
\references{
Richter, J. (2018) \emph{About the mass and magnitude distributions of meteor showers}.
WGN, Journal of the International Meteor Organization, vol. 46, no. 1, p. 34-38
}
\seealso{
\link{mideal}
\link{vmperception}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vmgeom.R
\name{vmgeom}
\alias{vmgeom}
\alias{dvmgeom}
\alias{pvmgeom}
\alias{qvmgeom}
\alias{rvmgeom}
\title{Geometric Model of Visual Meteor Magnitudes}
\usage{
dvmgeom(m, lm, r, log = FALSE, perception_fun = vmperception)

pvmgeom(
  m,
  lm,
  r,
  lower.tail = TRUE,
  log = FALSE,
  perception_fun = vmperception
)

qvmgeom(p, lm, r, lower.tail = TRUE, perception_fun = vmperception)

rvmgeom(n, lm, r, perception_fun = vmperception)
}
\arguments{
\item{m}{integer; the meteor magnitude.}

\item{lm}{numeric; limiting magnitude.}

\item{r}{numeric; the population index.}

\item{log}{logical; if \code{TRUE}, probabilities p are given as \code{log(p)}.}

\item{perception_fun}{function; perception probability function (optional).
Default is \link{vmperception}.}

\item{lower.tail}{logical; if \code{TRUE} (default) probabilities are
\eqn{P[M < m]}, otherwise, \eqn{P[M \ge m]}.}

\item{p}{numeric; probability.}

\item{n}{numeric; count of meteor magnitudes.}
}
\value{
\itemize{
\item \code{dvmgeom}: density
\item \code{pvmgeom}: distribution function
\item \code{qvmgeom}: quantile function
\item \code{rvmgeom}: random generation
}

The length of the result is determined by \code{n} for \code{rvmgeom}, and by the maximum
of the lengths of the numeric vector arguments for the other functions.
All arguments are vectorized; standard R recycling rules apply.

Since the distribution is discrete, \code{qvmgeom} and \code{rvmgeom} always return integer values.
\code{qvmgeom} may return \code{NaN} with a warning.
}
\description{
Density, distribution function, quantile function, and random generation
for the geometric model of visual meteor magnitudes.
}
\details{
In visual meteor observations, magnitudes are estimated as integer values.
Consequently, the distribution of observed magnitudes is discrete, and its
probability mass function is given by

\deqn{
P[M = m] \sim
\begin{cases}
  f(m_{\mathrm{lim}} - m)\, r^m, & \text{if } m_{\mathrm{lim}} - m > -0.5,\\[5pt]
  0 & \text{otherwise,}
\end{cases}
}

where \eqn{m_{\mathrm{lim}}} denotes the limiting (non-integer) magnitude of the observation,
and \eqn{m} the integer meteor magnitude.
The function \eqn{f(\cdot)} denotes the \link[=vmperception]{perception probability function}.

Thus, the distribution is the product of the perception probabilities and the
underlying \link[stats:Geometric]{geometric distribution} of meteor magnitudes.
Therefore, the parameter \code{p} of the geometric distribution is given by \code{p = 1 - 1/r}.

The parameter \code{lm} specifies the limiting magnitude for the meteor magnitude \code{m}.
\code{m} must be an integer meteor magnitude.
The length of the vector \code{lm} must either equal the length of the vector \code{m},
or \code{lm} must be a scalar value.
In the case of \code{rvmgeom}, the length of the vector \code{lm} must equal \code{n},
or \code{lm} must be a scalar value.

If a different perception probability function \code{perception_fun} is provided,
it must have the signature \verb{function(x)} and return the perception probability of
the difference \code{x} between the limiting magnitude and the meteor magnitude.
If \code{x >= 15.0}, the function \code{perception_fun} should return a perception probability of \code{1.0}.
The argument \code{perception_fun} is resolved using \link{match.fun}.
}
\examples{
N <- 100
r <- 2.0
limmag <- 6.5
(m <- seq(6, -7))

# discrete density of `N` meteor magnitudes
(freq <- round(N * dvmgeom(m, limmag, r)))

# log likelihood function
lld <- function(r) {
    -sum(freq * dvmgeom(m, limmag, r, log = TRUE))
}

# maximum likelihood estimation (MLE) of r
est <- optim(2, lld, method = "Brent", lower = 1.1, upper = 4)

# estimations
est$par # mean of r

# generate random meteor magnitudes
m <- rvmgeom(N, r, lm = limmag)

# log likelihood function
llr <- function(r) {
    -sum(dvmgeom(m, limmag, r, log = TRUE))
}

# maximum likelihood estimation (MLE) of r
est <- optim(2, llr, method = "Brent", lower = 1.1, upper = 4, hessian = TRUE)

# estimations
est$par # mean of r
sqrt(1 / est$hessian[1][1]) # standard deviation of r

m <- seq(6, -4, -1)
p <- vismeteor::dvmgeom(m, limmag, r)
barplot(
    p,
    names.arg = m,
    main = paste0("Density (r = ", r, ", limmag = ", limmag, ")"),
    col = "blue",
    xlab = "m",
    ylab = "p",
    border = "blue",
    space = 0.5
)
axis(side = 2, at = pretty(p))
}
\seealso{
\link{vmperception}
\link[stats:Geometric]{stats::Geometric}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/perception.R
\name{vmperception}
\alias{vmperception}
\title{Perception Probabilities of Visual Meteor Magnitudes}
\usage{
vmperception(dm)
}
\arguments{
\item{dm}{numeric; difference between the limiting magnitude and the meteor magnitude.}
}
\value{
This function returns the visual perception probabilities.
}
\description{
Provides the perception probability of visual meteor magnitudes.
}
\details{
The perception probabilities of \emph{Koschack R., Rendtel J., 1990b}
are estimated with the formula
\deqn{
    p(dm) = \begin{cases}
        1.0 - \exp\left(-z(dm + 0.5)\right)\  & \text{ if } dm > -0.5,\\
        0.0 \  & \text{ otherwise,}
    \end{cases}
}
where
\deqn{
z(x) = 0.0037 \, x + 0.0019 \, x^2 + 0.00271 \, x^3 + 0.0009 \, x^4
}
and \code{dm} is the difference between the limiting magnitude and the meteor magnitude.
}
\examples{
# Perception probability of visually estimated meteor of magnitude 3.0
# with a limiting magnitude of 5.6.
vmperception(5.6 - 3.0)

# plot
old_par <- par(mfrow = c(1, 1))
plot(
    vmperception,
    -0.5, 8,
    main = paste(
        "perception probability of",
        "visual meteor magnitudes"
    ),
    col = "blue",
    xlab = "dm",
    ylab = "p"
)

par(old_par)
}
\references{
Koschack R., Rendtel J., 1990b
\emph{Determination of spatial number density and mass index from visual meteor observations (II).}
WGN 18, 119–140.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vmgeom_vst.R
\name{vmgeomVst}
\alias{vmgeomVst}
\alias{vmgeomVstFromMagn}
\alias{vmgeomVstToR}
\title{Variance-Stabilizing Transformation for Geometric Visual Meteor Magnitudes}
\usage{
vmgeomVstFromMagn(m, lm)

vmgeomVstToR(tm, log = FALSE, deriv.degree = 0L)
}
\arguments{
\item{m}{integer; meteor magnitude.}

\item{lm}{numeric; limiting magnitude.}

\item{tm}{numeric; transformed magnitude.}

\item{log}{logical; if \code{TRUE}, the logarithm of the population index \code{r} is returned.}

\item{deriv.degree}{integer; the order of the derivative at \code{tm} to return
instead of \code{r} or \code{log(r)}. Must be \code{0}, \code{1}, or \code{2}.}
}
\value{
\itemize{
\item \code{vmgeomVstFromMagn}: numeric value, the transformed meteor magnitude.
\item \code{vmgeomVstToR}: numeric value of the population index \code{r}, derived from
the mean of \code{tm}.
}

The argument \code{deriv.degree} can be used to apply the delta method.
If \code{log = TRUE}, the logarithm of \code{r} is returned.
}
\description{
Applies a variance-stabilizing transformation to visual meteor magnitudes
under the geometric model.
}
\details{
Many linear models require the variance of visual meteor magnitudes to be
homoscedastic. The function \code{vmgeomVstFromMagn} applies a transformation
that produces homoscedastic distributions of visual meteor magnitudes if the
underlying distribution follows a geometric model.

The geometric model of visual meteor magnitudes
depends on the \link[=vmgeom]{population index} \code{r} and the limiting magnitude \code{lm},
resulting in a two-parameter distribution. Without detection probabilities,
the magnitude distribution is purely geometric, and for integer limiting
magnitudes the variance depends only on the population index \code{r}. Since the
limiting magnitude \code{lm} is a fixed parameter and never estimated
statistically, the magnitudes can be transformed such that, for example,
the mean of the transformed magnitudes directly provides an estimate of \code{r}
using the function \code{vmgeomVstToR}.

A key advantage of this transformation is that the limiting magnitude \code{lm}
is already incorporated into subsequent analyses. In this sense, the
transformation acts as a normalization of meteor magnitudes and yields a
variance close to \code{1.0}.

This transformation is valid for \eqn{1.4 \le r \le 3.5}.
The numerical form of the transformation is version-specific and may change
substantially in future releases. Do not rely on equality of transformed
values across package versions.
}
\note{
The internal approximations used here are derived from the perception
probabilities produced by \link{vmperception}.
For details on the derivation, see the script \code{inst/derivation/vmgeom_vst.R} in the
package's source code.
}
\examples{
N <- 100
r <- 2.0
limmag <- 6.3

# Simulate magnitudes
m <- rvmgeom(N, limmag, r)

# Variance-stabilizing transformation
tm <- vmgeomVstFromMagn(m, limmag)
tm.mean <- mean(tm)
tm.var  <- var(tm)

# Estimator for r from the transformed mean
r.hat  <- vmgeomVstToR(tm.mean)

# Derivative dr/d(tm) at tm.mean (needed for the delta method)
dr_dtm <- vmgeomVstToR(tm.mean, deriv.degree = 1L)

# Variance of the sample mean of tm
var_tm.mean <- tm.var / N

# Delta method: variance and standard error of r.hat
var_r.hat <- (dr_dtm^2) * var_tm.mean
se_r.hat  <- sqrt(var_r.hat)

# Results
print(r.hat)
print(se_r.hat)
}
\seealso{
\link{vmgeom} \link{vmperception}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{vismeteor-package}
\alias{vismeteor}
\alias{vismeteor-package}
\title{vismeteor: Analysis of Visual Meteor Data}
\description{
Provides a suite of analytical functionalities to process and analyze visual meteor observations from the Visual Meteor Database of the International Meteor Organization \url{https://www.imo.net/}.
}
\details{
The data used in this package can created and provided by
\href{https://pypi.org/project/imo-vmdb/}{imo-vmdb}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/jankorichter/vismeteor}
  \item Report bugs at \url{https://github.com/jankorichter/vismeteor/issues}
}

}
\author{
\strong{Maintainer}: Janko Richter \email{janko@richtej.de}

Authors:
\itemize{
  \item Janko Richter \email{janko@richtej.de}
}

}
